static const char *const k1IntroStringsDOSCDHebrew[51] = {
	"This is a text test - 1",
	"This is a text test - 2",
	"\xE1\xF8\xEB\xE5\xFA, \xE7\xE9\xEC\xE9\xF7...",
	"...\xEE\xF0\xE4\xE9\xE2\xED \xF8\xE1 \xE4\xF2\xE5\xF6\xEE\xE4 \xF9\xEC \xE4\xEE\xE9\xF1\xE8\xE9\xF7\xF0\xE9\xED \xE4\xEE\xEC\xEB\xE5\xFA\xE9\xE9\xED!",
	"\xE4\xE1\xE4\xEC\xFA\xE9 \xE0\xE5\xFA\xEA?",
	"\xEE\xEC\xF7\xE5\xEC\xED!",
	"\xF9\xEE\xF2\xFA\xE9 \xE0\xFA\xEE\xE5\xEC \xF2\xEC \xE0\xE5\xE3\xE5\xFA \xE4\xE1\xF8\xE9\xE7\xE4 \xF9\xEC\xEA.",
	"\xE5\xF6\xE9\xF4\xE9\xFA\xE9 \xF9\xFA\xE1\xE5\xE0 \xEE\xE5\xF7\xE3\xED \xE9\xE5\xFA\xF8.",
	"\xEC\xEE\xE4 \xEC\xE9 \xEC\xEE\xE4\xF8?",
	"\xE4\xE0\xF8\xF5 \xE4\xE6\xE5 \xF0\xFA\xE5\xF0\xE4 \xEC\xEE\xF8\xE5\xFA\xE9!",
	"\xE4\xF7\xEC\xEC\xE4 \xE4\xF4\xF2\xE5\xE8\xE4 \xF9\xEC\xEA \xE0\xE9\xF0\xE4 \xE9\xEB\xE5\xEC\xE4 \xEC\xF2\xF6\xE5\xF8 \xE0\xE5\xFA\xE9.",
	"\xF7\xE3\xE9\xEE\xE4, \xF7\xE8\xE5\xEC \xE0\xE5\xFA\xE9 \xF2\xEB\xF9\xE9\xE5!",
	"\xF0\xE5\xFA\xF8 \xEC\xE9 \xF8\xF7 \xEE\xF2\xE8 \xF7\xF1\xED.",
	"\xE0\xF0\xE9 \xEE\xFA\xEB\xE5\xE5\xEF \xEC\xF4\xE2\xE5\xF2 \xE1\xEA, \xE1\xE6\xE4 \xE0\xE9\xEF \xF1\xF4\xF7...",
	"...\xF2\xEB\xF9\xE9\xE5 \xE0\xFA\xE4 \xE0\xE1\xEF...",
	"...\xE0\xE1\xEC \xE0\xFA \xF2\xE9\xF0\xE9\xEA \xE0\xF0\xE9 \xEE\xF9\xE0\xE9\xF8 \xEB\xF4\xE9 \xF9\xE4\xEF.",
	"\xE0\xE9\xF0\xF0\xE9 \xEE\xE6\xE9\xEC \xE3\xEE\xF2\xE4 \xF2\xEC \xF7\xE9\xF8\xF0\xE3\xE9\xE4...",
	"...\xE0\xE1\xEC \xEC\xE0 \xE0\xE5\xEB\xEC \xEC\xF9\xEC\xE5\xEC \xEE\xEE\xEA \xE0\xFA \xF9\xEC\xEA.",
	"Malcolm the Jester has broken free!",
	"He now controls the Kyragem...",
	"...source of all magic in Kyrandia!",
	"",
	"\xE0\xFA\xE4 \xEE\xF2\xE6 \xEC\xE4\xEB\xF9\xE9\xEC \xE0\xE5\xFA\xE9!",
	"\xE0\xEA \xEC\xE0, \xE9\xF9 \xEC\xE9 \xEE\xF9\xE4\xE5 \xEE\xE9\xE5\xE7\xE3 \xE1\xF9\xE1\xE9\xEC\xEA...",
	"\xEB\xE9\xF9\xE5\xF3 \xEE\xE1\xE3\xE7!",
	"\xE0\xE1\xEC \xE0\xFA\xEF \xEC\xEA \xE0\xE6\xE4\xF8\xE4 \xE4\xE5\xE2\xF0\xFA.",
	"\xE0\xEC \xFA\xF7\xF4\xE5\xF5 \xF2\xEC \xE4\xF2\xF5 \xE4\xE6\xE4!",
	"\xE0\xEC \xFA\xE8\xF4\xF1 \xF2\xEC \xE4\xF2\xF5 \xE4\xE6\xE4 \xF1\xF0\xE0\xE9 \xF7\xE8\xEF!",
	"\xE1\xE5!",
	"\xE6\xE4 \xEE\xF6\xE7\xE9\xF7 \xE9\xE5\xFA\xF8 \xEB\xEB\xE4...",
	"...\xEC\xE0 \xEB\xEA?",
	"\xE7\xE4 \xE7\xE4 \xE7\xE4...",
	"\xE4\xE0\xE9\xEF \xE0\xF6\xE1\xF2\xE5\xFA \xE3\xE1\xF8 \xF0\xF4\xEC\xE0?",
	"\xE8\xE5\xE1 \xEC\xE4\xE9\xE5\xFA \xE1\xE7\xE9\xE9\xED!",
	"\xE1\xF8\xF0\xE3\xE5\xEF!",
	"\xF1\xE1\xE0! \xF2\xF9\xE9\xF0\xE5 \xE6\xE0\xFA!",
	"\xEC\xE0! \xE0\xFA\xE4 \xF2\xF9\xE9\xFA \xE6\xE0\xFA!",
	"\xE9\xE7\xE9 \xE4\xEE\xEC\xEA \xE1\xF8\xF0\xE3\xE5\xEF!",
	"\xEC\xE7\xE9\xE9 \xE4\xEE\xE9\xF1\xE8\xE9\xF7\xF0\xE9\xED \xE4\xEE\xEC\xEB\xE5\xFA\xE9\xE9\xED!",
	"\xE5\xE1\xEB\xEF, \xE4\xE1\xE4 \xF0\xF6\xE0 \xEC\xF9\xF7\xED \xE0\xFA \xF7\xE9\xF8\xF0\xE3\xE9\xE4!",
	"\xF8\xF2\xE9\xE5\xEF \xEE\xF6\xE5\xE9\xEF!",
	"\xEC\xE4\xEB\xF8\xE6\xE4 \xE4\xEE\xEC\xEB\xE5\xFA\xE9\xFA \xE4\xF8\xE0\xF9\xE5\xF0\xE4 \xF9\xEC\xE9...",
	"\xE0\xF0\xE9 \xEE\xEE\xF0\xE4 \xE0\xFA \xE4\xF1\xF0\xE3\xEC\xE9\xED \xEC\xF0\xF2\xEC\xE9\xE9\xED \xE4\xF8\xF9\xEE\xE9\xE5\xFA \xF9\xEC \xF7\xE9\xF8\xF0\xE3\xE9\xE4!",
	"\xEB\xEC \xE4\xEB\xE1\xE5\xE3, \xE1\xF8\xF0\xE3\xE5\xEF!",
	"\xE4\xE0\xF8\xF5 \xE4\xF4\xF1\xE9\xF7\xE4 \xEC\xE4\xFA\xE9\xE9\xF4\xE7...",
	"\xE5\xE7\xE1\xF8\xE9\xEA \xF9\xE1\xE5 \xEC\xF2\xF6\xEE\xED!",
	"\xE1\xF8\xE9\xEF \xE4\xE9\xF7\xF8\xE4...",
	"\xEE\xEC\xF7\xE5\xEC\xED \xE4\xF6\xEC\xE9\xE7 \xEC\xE4\xF9\xFA\xE7\xF8\xF8!",
	"\xE1\xF7\xF8\xE5\xE1 \xE4\xE5\xE0 \xE9\xE2\xE9\xF2 \xE0\xEC\xE9\xE9.",
	"\xF0\xE0 \xF2\xE6\xF8\xE9 \xEC\xE1\xF8\xF0\xE3\xE5\xEF...",
	""
};

static const StringListProvider k1IntroStringsDOSCDHebrewProvider = { ARRAYSIZE(k1IntroStringsDOSCDHebrew), k1IntroStringsDOSCDHebrew };

static const char *const k1ItemNamesDOSCDHebrew[107] = {
	"\xF0\xE5\xF4\xEA",
	"\xE0\xE7\xEC\xEE\xE4",
	"\xFA\xF8\xF9\xE9\xF9",
	"\xE9\xE4\xEC\xE5\xED",
	"\xE1\xF8\xF7\xFA",
	"\xF4\xF0\xE9\xF0\xE4",
	"\xE0\xE5\xE3\xED",
	"\xF4\xE8\xE3\xE4",
	"\xF1\xF4\xE9\xF8",
	"\xEC\xF9\xED",
	"\xE8\xE5\xF4\xE6",
	"\xF9\xE4\xED",
	"\xE0\xE1\xEF \xF9\xEE\xF9",
	"\xE0\xE1\xEF \xE9\xF8\xE7",
	"\xE0\xE1\xEF \xF7\xF9\xFA",
	"\xE0\xE1\xEF \xF9\xE5\xE0\xE1\xFA",
	"\xF9\xE5\xF9\xF0\xE4",
	"\xF6\xE1\xF2\xE5\xF0\xE9",
	"\xF1\xE7\xEC\xE1",
	"\xF9\xE5\xF9\xF0\xFA \xEB\xF1\xF3",
	"\xF4\xF1\xEC\xE5\xEF \xEB\xF1\xF3",
	"\xEE\xE8\xE1\xF2 \xEB\xF1\xF3",
	"\xEE\xE8\xE1\xF2 \xE6\xE4\xE1",
	"\xE8\xE1\xF2\xFA \xE6\xE4\xE1",
	"\xE2\xE1\xE9\xF2 \xEE\xEC\xEB\xE5\xFA\xE9",
	"\xE0\xF6\xE8\xF8\xE5\xE1\xEC",
	"\xE1\xEC\xE5\xE8",
	"\xE0\xE2\xE5\xE6",
	"\xE2\xF8\xE2\xF8\xE9 \xE0\xF9 \xF7\xE5\xF8\xF0\xE9-\xF2\xE3",
	"\xE2\xF8\xE2\xF8\xE9 \xE0\xF9",
	"\xE2\xF8\xE2\xF8\xE9 \xE0\xF9",
	"\xE2\xF8\xE2\xF8\xE9 \xE0\xF9",
	"\xE2\xF8\xE2\xF8\xE9 \xE0\xF9",
	"\xE2\xF8\xE2\xF8\xE9 \xE0\xF9",
	"\xE3\xE2",
	"\xE0\xE3\xF8\xE4",
	"\xF9\xE5\xF7 \xE8\xEC\xE4",
	"\xF2\xF6\xED",
	"\xFA\xF4\xE5\xE7",
	"\xEC\xE9\xE1\xFA \xFA\xF4\xE5\xE7",
	"\xE0\xE5\xEB\xEE\xF0\xE9\xE5\xFA",
	"\xF4\xE8\xF8\xE9\xE4",
	"\xF4\xFA\xF7",
	"\xE2\xE5\xEC\xE4",
	"\xEE\xF1\xE5\xF8",
	"'\xF2\xF0\xE7",
	"\xF0\xE5\xF6\xE4",
	"\xE1\xE9\xF6\xE4",
	"\xF2\xEC\xE4",
	"\xFA\xEC\xFA\xEF",
	"\xEB\xE5\xEB\xE1 \xF0\xE5\xF4\xEC",
	"\xEB\xE3\xE5\xF8 \xE1\xE3\xE5\xEC\xE7",
	"\xE3\xEE\xF2\xE4",
	"\xF8\xE0\xE9",
	"\xF4\xFA\xE9\xFA \xF7\xF8\xE7",
	"\xE0\xE1\xE5\xE1",
	"\xF9\xF2\xE5\xEF \xE7\xE5\xEC",
	"\xEE\xF4\xFA\xE7 \xE1\xF8\xE6\xEC",
	"\xEE\xF4\xFA\xE7 \xE9\xF8\xF7\xEF",
	"\xEE\xF4\xFA\xE7 \xE1\xE6\xEC\xFA",
	"\xF9\xE9\xF7\xE5\xE9 \xE0\xE3\xE5\xED",
	"\xF9\xE9\xF7\xE5\xE9 \xE0\xE3\xE5\xED",
	"\xF9\xE9\xF7\xE5\xE9 \xEB\xE7\xE5\xEC",
	"\xF9\xE9\xF7\xE5\xE9 \xEB\xE7\xE5\xEC",
	"\xF9\xE9\xF7\xE5\xE9 \xF6\xE4\xE5\xE1",
	"\xF9\xE9\xF7\xE5\xE9 \xF6\xE4\xE5\xE1",
	"\xF9\xE9\xF7\xE5\xE9 \xE9\xF8\xE5\xF7",
	"\xF9\xE9\xF7\xE5\xE9 \xEB\xFA\xE5\xED",
	"\xF9\xE9\xF7\xE5\xE9 \xF1\xE2\xE5\xEC",
	"\xF9\xE9\xF7\xE5\xE9 \xE4\xF7\xF9\xFA \xE1\xF2\xF0\xEF",
	"\xEE\xE9\xED \xEE\xFA\xE5\xF7\xE9\xED",
	"\xEE\xE9\xED \xEE\xFA\xE5\xF7\xE9\xED",
	"\xEE\xE9\xED \xEE\xEC\xE5\xE7\xE9\xED",
	"\xEE\xE9\xED \xEE\xEC\xE5\xE7\xE9\xED",
	"\xEE\xE9\xED \xEE\xE9\xF0\xF8\xEC\xE9\xE9\xED",
	"\xEE\xE9\xED \xEE\xE9\xF0\xF8\xEC\xE9\xE9\xED",
	"\xEE\xE9\xED \xF7\xF1\xE5\xEE\xE9\xED",
	"\xEE\xE9\xED \xF7\xF1\xE5\xEE\xE9\xED",
	"\xE1\xF7\xE1\xE5\xF7\xE5\xEF \xF8\xE9\xF7",
	"\xE1\xF7\xE1\xE5\xF7\xE5\xEF \xF8\xE9\xF7",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xEE\xE2\xE9\xEC\xE4",
	"\xF4\xE9\xF1\xFA \xF7\xEC\xF3",
	"\xF4\xE9\xF1\xFA \xF7\xEC\xF3",
	"\xF4\xE9\xF1\xFA \xF7\xEC\xF3",
	"\xF4\xE9\xF1\xFA \xF7\xEC\xF3",
	"\xF4\xE9\xF1\xFA \xF7\xEC\xF3",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xE0\xE3\xE5\xEE\xE4",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xEB\xFA\xE5\xEE\xE4",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xF6\xE4\xE5\xE1\xE4",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xE9\xF8\xE5\xF7\xE4",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xEB\xE7\xE5\xEC\xE4-\xE9\xF8\xE5\xF7\xE4",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xEB\xE7\xE5\xEC\xE4",
	"\xE0\xE1\xEF \xEB\xF9\xF3 \xF1\xE2\xE5\xEC\xE4",
	"\xF1\xEC\xF2 \xEB\xE1\xE3",
	"\xEB\xFA\xF8 \xEE\xEC\xEB\xE5\xFA",
	"\xF9\xF8\xE1\xE9\xE8 \xEE\xEC\xEB\xE5\xFA",
	"\xEE\xF4\xFA\xE7 \xE6\xE4\xE1",
	"\xF4\xF8\xE9\xE8 \xEC\xE0 \xEE\xE5\xEB\xF8"
};

static const StringListProvider k1ItemNamesDOSCDHebrewProvider = { ARRAYSIZE(k1ItemNamesDOSCDHebrew), k1ItemNamesDOSCDHebrew };

static const char *const k1TakenStringsDOSCDHebrew[2] = {
	"\xEC\xF7\xE7\xFA\xED ",
	"\xEC\xF7\xE7\xFA\xED "
};

static const StringListProvider k1TakenStringsDOSCDHebrewProvider = { ARRAYSIZE(k1TakenStringsDOSCDHebrew), k1TakenStringsDOSCDHebrew };

static const char *const k1PlacedStringsDOSCDHebrew[1] = {
	"\xE4\xF0\xE7\xFA\xED "
};

static const StringListProvider k1PlacedStringsDOSCDHebrewProvider = { ARRAYSIZE(k1PlacedStringsDOSCDHebrew), k1PlacedStringsDOSCDHebrew };

static const char *const k1DroppedStringsDOSCDHebrew[1] = {
	"\xF9\xEE\xE8\xFA\xED "
};

static const StringListProvider k1DroppedStringsDOSCDHebrewProvider = { ARRAYSIZE(k1DroppedStringsDOSCDHebrew), k1DroppedStringsDOSCDHebrew };

static const char *const k1NoDropStringsDOSCDHebrew[2] = {
	"\xEC\xE0 \xF0\xE9\xFA\xEF \xEC\xF9\xEE\xE5\xE8 \xF2\xE5\xE3 \xE7\xF4\xF6\xE9\xED \xE1\xE6\xE9\xF8\xE4 \xE4\xE6\xE5.",
	"\xEC\xE0 \xF0\xE9\xFA\xEF \xEC\xF9\xE9\xED \xE0\xFA \xE6\xE4 \xEB\xE0\xEF."
};

static const StringListProvider k1NoDropStringsDOSCDHebrewProvider = { ARRAYSIZE(k1NoDropStringsDOSCDHebrew), k1NoDropStringsDOSCDHebrew };

static const char *const k1PutDownStringDOSCDHebrew[1] = {
	"\xE0\xE5\xEC\xE9 \xEE\xE5\xE8\xE1 \xFA\xE7\xE9\xEC\xE4 \xF9\xE0\xF0\xE9\xE7 \xE0\xFA \xE6\xE4."
};

static const StringListProvider k1PutDownStringDOSCDHebrewProvider = { ARRAYSIZE(k1PutDownStringDOSCDHebrew), k1PutDownStringDOSCDHebrew };

static const char *const k1WaitAmuletStringDOSCDHebrew[1] = {
	"\xE0\xF0\xE9 \xEE\xF0\xE9\xE7 \xF9\xF2\xEC\xE9\xE9 \xEC\xE4\xEE\xFA\xE9\xEF \xF9\xEB\xE5\xE7\xE5\xFA\xE9\xE5 \xF9\xEC \xE4\xF7\xEE\xF2 \xF9\xEC\xE9 \xE9\xE7\xE6\xF8\xE5."
};

static const StringListProvider k1WaitAmuletStringDOSCDHebrewProvider = { ARRAYSIZE(k1WaitAmuletStringDOSCDHebrew), k1WaitAmuletStringDOSCDHebrew };

static const char *const k1BlackJewelStringDOSCDHebrew[1] = {
	"\xE6\xE5 \xE0\xE1\xEF \xE7\xEF, \xE0\xE1\xEC \xEE\xE3\xE5\xF2 \xE4\xE9\xE0 \xF9\xE7\xE5\xF8\xE4?"
};

static const StringListProvider k1BlackJewelStringDOSCDHebrewProvider = { ARRAYSIZE(k1BlackJewelStringDOSCDHebrew), k1BlackJewelStringDOSCDHebrew };

static const char *const k1HealingTipStringDOSCDHebrew[1] = {
	"\xE1\xE7\xE9\xE9, \xE6\xE4 \xF2\xF9\xE5\xE9 \xEC\xE4\xE9\xE5\xFA \xF9\xE9\xEE\xE5\xF9\xE9 \xE0\xED \xE7\xEC\xE9\xEC\xE4 \xE0\xF4\xE2\xF2."
};

static const StringListProvider k1HealingTipStringDOSCDHebrewProvider = { ARRAYSIZE(k1HealingTipStringDOSCDHebrew), k1HealingTipStringDOSCDHebrew };

static const char *const k1PoisonGoneStringDOSCDHebrew[2] = {
	"\xEE\xE3\xE4\xE9\xED!",
	"\xE4\xF9\xF4\xF2\xE5\xFA \xE4\xF8\xF2\xEC \xF0\xF2\xEC\xEE\xE5 \xEB\xEC\xE0 \xE4\xE9\xE5!"
};

static const StringListProvider k1PoisonGoneStringDOSCDHebrewProvider = { ARRAYSIZE(k1PoisonGoneStringDOSCDHebrew), k1PoisonGoneStringDOSCDHebrew };

static const char *const k1ThePoisonStringsDOSCDHebrew[4] = {
	"\xE4\xE0\xF8\xF1...",
	"\xEC\xE0 \xEE\xF6\xEC\xE9\xE7 \xEC\xF0\xF9\xE5\xED...",
	"\xE0\xF0\xE9 \xEC\xE0 \xEE\xF8\xE2\xE9\xF9 \xE8\xE5\xE1 \xEB\xEC \xEB\xEA...",
	"\xE4\xF0\xE7\xF9 \xE4\xE6\xE4\r\xE1\xE5\xE3\xE0\xE9 \xE0\xF8\xF1\xE9!"
};

static const StringListProvider k1ThePoisonStringsDOSCDHebrewProvider = { ARRAYSIZE(k1ThePoisonStringsDOSCDHebrew), k1ThePoisonStringsDOSCDHebrew };

static const char *const k1FluteStringsDOSCDHebrew[2] = {
	"\xEC\xE0 \xF0\xF9\xEE\xF2 \xEE\xF8\xF9\xE9\xED \xE1\xEE\xE9\xE5\xE7\xE3.",
	"\xE4\xFA\xE5 \xE4\xE0\xE7\xF8\xE5\xEF \xE1\xE4\xE7\xEC\xE8 \xE4\xE9\xE4 \xE2\xE1\xE5\xE4!"
};

static const StringListProvider k1FluteStringsDOSCDHebrewProvider = { ARRAYSIZE(k1FluteStringsDOSCDHebrew), k1FluteStringsDOSCDHebrew };

static const char *const k1WispJewelStringsDOSCDHebrew[2] = {
	"\xEC\xE0 \xEE\xEE\xF9 \xEE\xFA\xE7\xF9\xF7 \xEC\xE9 \xEB\xF8\xE2\xF2.",
	"\xE0\xE5\xEC\xE9 \xEE\xE5\xE8\xE1 \xFA\xE7\xE9\xEC\xE4 \xF9\xE0\xF0\xE9\xE7 \xE0\xFA \xE6\xE4."
};

static const StringListProvider k1WispJewelStringsDOSCDHebrewProvider = { ARRAYSIZE(k1WispJewelStringsDOSCDHebrew), k1WispJewelStringsDOSCDHebrew };

static const char *const k1MagicJewelStringsDOSCDHebrew[1] = {
	"\xE6\xE5 \xE4\xE9\xFA\xE4 \xE4\xF8\xE2\xF9\xE4 \xEE\xE5\xE6\xF8\xE4."
};

static const StringListProvider k1MagicJewelStringsDOSCDHebrewProvider = { ARRAYSIZE(k1MagicJewelStringsDOSCDHebrew), k1MagicJewelStringsDOSCDHebrew };

static const char *const k1FlaskFullStringDOSCDHebrew[1] = {
	"\xE4\xE1\xF7\xE1\xE5\xF7\xE5\xEF \xEB\xE1\xF8 \xEE\xEC\xE0."
};

static const StringListProvider k1FlaskFullStringDOSCDHebrewProvider = { ARRAYSIZE(k1FlaskFullStringDOSCDHebrew), k1FlaskFullStringDOSCDHebrew };

static const char *const k1FullFlaskStringDOSCDHebrew[4] = {
	"\xE4\xE1\xF7\xE1\xE5\xF7\xE5\xEF \xEE\xEC\xE0 \xF2\xEB\xF9\xE9\xE5\r\xE1\xEE\xE9\xED \xEE\xFA\xE5\xF7\xE9\xED \xE5\xEE\xE1\xF2\xE1\xF2\xE9\xED.",
	"\xE4\xE1\xF7\xE1\xE5\xF7\xE5\xEF \xEE\xEC\xE0 \xF2\xEB\xF9\xE9\xE5\r\xE1\xEE\xE9\xED \xEE\xEC\xE5\xE7\xE9\xED.",
	"\xE4\xE1\xF7\xE1\xE5\xF7\xE5\xEF \xEE\xEC\xE0 \xF2\xEB\xF9\xE9\xE5\r\xE1\xEE\xE9\xED \xEE\xE9\xF0\xF8\xEC\xE9\xE9\xED.",
	"\xEE\xE9\xED \xF7\xF1\xE5\xEE\xE9\xED."
};

static const StringListProvider k1FullFlaskStringDOSCDHebrewProvider = { ARRAYSIZE(k1FullFlaskStringDOSCDHebrew), k1FullFlaskStringDOSCDHebrew };

static const char *const k1OutroHomeStringDOSCDHebrew[1] = {
	"\xE1\xE9\xFA"
};

static const StringListProvider k1OutroHomeStringDOSCDHebrewProvider = { ARRAYSIZE(k1OutroHomeStringDOSCDHebrew), k1OutroHomeStringDOSCDHebrew };

static const char *const k1VeryCleverStringDOSCDHebrew[1] = {
	"\xF9\xF0\xE5\xEF \xE1\xE9\xE5\xFA\xF8!  \xE0\xE1\xEC \xE4\xEE\xE0\xEE\xF6\xE9\xED \xE4\xE0\xE5\xEE\xEC\xEC\xE9\xED \xE4\xEC\xEC\xE5 \xEC\xE7\xE9\xF0\xED."
};

static const StringListProvider k1VeryCleverStringDOSCDHebrewProvider = { ARRAYSIZE(k1VeryCleverStringDOSCDHebrew), k1VeryCleverStringDOSCDHebrew };

static const char *const k1GUIStringsDOSCDHebrew[81] = {
	"\xE3\xE1\xF8\xE9 \xE9\xEE\xE9 \xF7\xE9\xF8\xF0\xE3\xE9\xE4",
	"\xE8\xF2\xE9\xF0\xFA \xEE\xF9\xE7\xF7",
	"\xF9\xEE\xE9\xF8\xFA \xE4\xEE\xF9\xE7\xF7",
	"\xE1\xF7\xF8\xE9 \xEE\xF9\xE7\xF7",
	"\xE9\xF6\xE9\xE0\xE4 \xEE\xE4\xEE\xF9\xE7\xF7",
	"\xE7\xE6\xF8\xE4 \xEC\xEE\xF9\xE7\xF7",
	"\xE1\xF7\xF8\xE9 \xEE\xF9\xE7\xF7",
	"\xE0\xE9\xE6\xE4 \xEE\xF9\xE7\xF7 \xEC\xE8\xF2\xE5\xEF?",
	"\xE1\xE7\xF8\xE5 \xE4\xE9\xEB\xEF \xEC\xF9\xEE\xE5\xF8:",
	"] \xEE\xF9\xE1\xF6\xFA \xF8\xE9\xF7\xE4 [",
	"\xE1\xE9\xE8\xE5\xEC",
	"\xE4\xE6\xE9\xF0\xE5 \xFA\xE9\xE0\xE5\xF8 \xF9\xEC \xE4\xEE\xF9\xE7\xF7 \xE4\xF9\xEE\xE5\xF8:",
	"\xF9\xEE\xE9\xF8\xE4",
	"\xF0\xE5\xE7 \xF2\xEC \xEE\xF9\xEB\xE1\xEA \xE1\xF9\xEC\xE5\xED, \xE1\xF8\xF0\xE3\xE5\xEF.",
	"\xE4\xE0\xED \xE0\xFA\xED \xE1\xE8\xE5\xE7\xE9\xED \xF9\xE1\xF8\xF6\xE5\xF0\xEB\xED \xEC\xF6\xE0\xFA?",
	"XXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXXXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXXXXXXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"\xFA\xF4\xF8\xE9\xE8 \xF8\xE0\xF9\xE9",
	"\xF4\xE5\xF2\xEC",
	"\xEB\xE1\xE5\xE9",
	"\xEB\xEF",
	"\xEC\xE0",
	"\xEE\xE4\xE9\xF8\xE5\xFA \xE4\xEC\xE9\xEB\xE4 ",
	"\xEE\xE4\xE9\xF8\xE5\xFA \xE8\xF7\xF1\xE8 ",
	"\xEE\xE5\xE6\xE9\xF7\xE4 ",
	"\xF6\xEC\xE9\xEC\xE9\xED ",
	"\xF7\xE5\xEC / \xE8\xF7\xF1\xE8 "
};

static const StringListProvider k1GUIStringsDOSCDHebrewProvider = { ARRAYSIZE(k1GUIStringsDOSCDHebrew), k1GUIStringsDOSCDHebrew };

static const char *const k1NewGameStringDOSCDHebrew[1] = {
	"] \xE4\xFA\xE7\xEC\xFA \xEE\xF9\xE7\xF7 \xE7\xE3\xF9 ["
};

static const StringListProvider k1NewGameStringDOSCDHebrewProvider = { ARRAYSIZE(k1NewGameStringDOSCDHebrew), k1NewGameStringDOSCDHebrew };

static const char *const k1ConfigStringsDOSCDHebrew[12] = {
	"\xE0\xE8\xE9\xFA \xE1\xE9\xE5\xFA\xF8",
	"\xE0\xE8\xE9\xFA",
	"\xF8\xE2\xE9\xEC\xE4",
	"\xEE\xE4\xE9\xF8\xE4",
	"\xEE\xE4\xE9\xF8\xE4 \xE1\xE9\xE5\xFA\xF8",
	"\xE8\xF7\xF1\xE8 \xE1\xEC\xE1\xE3",
	"\xF7\xE5\xEC \xE1\xEC\xE1\xE3",
	"\xF7\xE5\xEC \xE5\xE8\xF7\xF1\xE8",
	"\xE0\xE8\xE9\xFA \xE1\xE9\xE5\xFA\xF8",
	"\xF8\xE2\xE9\xEC\xE4",
	"\xEE\xE4\xE9\xF8\xE4",
	"\xEC\xE7\xE9\xF6\xE4"
};

static const StringListProvider k1ConfigStringsDOSCDHebrewProvider = { ARRAYSIZE(k1ConfigStringsDOSCDHebrew), k1ConfigStringsDOSCDHebrew };

