return {
PASS="Pass!\n",
YOU_SEE="you see %s",
ON="On - ",
TO="To - ",
NOTHING="nothing!\n",
NO_EFFECT="\nNo effect\n",
AARGH="Aargh!\n",
BLOCKED="Blocked!\n",
HOW_MANY="How many? ",
YOU_ARE_CARRYING_TOO_MUCH_ALREADY="You are carrying too much already.\n",
YOU_PRIED_THE_NAILS_OUT_OF_THE_CRATE="You pried the nails out of the crate.\n",
YOU_NAIL_THE_CRATE_SHUT="You nail the crate shut.\n",
IT_HAS_NO_EFFECT="It has no effect.\n",
IT_IS_NAILED_SHUT="It is nailed shut.\n",
IT_IS_LOCKED="It is locked.\n",
IS_NOT_STRONG_ENOUGH="%s is not strong enough.\n",
NOT_WHILE_UNDERGROUND="Not while underground.\n",
YOU_ARE_SOMEWHERE_NEAR="You are somewhere near",
YOU_CANT_DIG_ANY_DEEPER="You can't dig any deeper.\n",
IT_IS_STUCK="It is stuck.\n",
THE_CONTROL_PANEL_OPERATES_THE_DREAM_MACHINE="The control panel operates the dream machine.\n",
YOU_NEED_TO_USE_PLIERS_TO_ATTACH_THE_CABLE="You need to use pliers to attach the cable.",
YOU_WILL_CHEW_IT_INSTINCTIVELY="You will chew it instinctively.\n",
THE_PLATE_IN_THE_CAMERA_HAS_ALREADY_BEEN_EXPOSED="The plate in the camera has already been exposed.\n",
YOU_MUST_USE_A_SHOVEL_TO_MOVE_THAT="You must use a shovel to move that.\n",
YOU_CANT_DETERMINE_HOW_TO_READ_THE_TIME="You can't determine how to read the time.\n",
THE_BERRIES_ARE_NOT_EDIBLE="The berries are not edible.\n",
YOU_DONT_KNOW_HOW_IT_WORKS="You don't know how it works.\n",
THE_GROUND_IS_NOT_CLEAR_FOR_DIGGING="The ground is not clear for digging.\n",
YOU_CANT_DIG_HERE="You can't dig here.\n",
YOU_FILLED_IN_THE_HOLE="You filled in the hole.\n",
YOU_MAY_USE_THE_RUBY_SLIPPERS_TO_GO_HOME="You may use the ruby slippers to go home. If you choose to, you may now view the Grand Finale.\n\n*Afterwards, you will exit to DOS. You will be able to resume play from your last saved game.\n\n*Would you like to see the Grand Finale now? If so, press Y. To resume play now, hit N: (Y N)\n",
CLICK="Click.\n",
NOT_WHILE_IN_SOLO_MODE="Not while in solo mode!\n",
SOMETHING="something",
SEARCHING_HERE_YOU_FIND="Searching here, you find %s",
SEARCHING_HERE_YOU_FIND_NOTHING="Searching here, you find nothing.\n",
SEARCH_NEXT_OBJ=", %s",
SEARCH_LAST_OBJ=" and %s",
YOU_CANT_SLEEP_IN_A_DREAM="You can't sleep in a dream.\n",
TENT_OBJ_IN_THE_WAY="The %s is in the way.\n",
THE_GROUND_IS_NOT_FLAT_ENOUGH="The ground is not flat enough.\n",
REST="Rest",
NOT_WHILE_IN_COMBAT_MODE="Not while in combat mode!\n",
NOT_WHILE_FOES_ARE_NEAR=" - Not while foes are near!\n",
IS_TOO_NEAR_TO_SETUP_CAMP=" - %s is too near to set up camp.\n",
IS_POISONED="\n%s is poisoned!",
DO_YOU_REALLY_WANT_TO_SLEEP="\nDo you really want to sleep anyway? ",
REST_UNTIL="\nRest until %s?",
SUNRISE="sunrise",
SUNSET="sunset",
HOW_MANY_HOURS="\nHow many hours? ",
HAS_A_DREAM="\n%s has a dream...\n\n",
SEES_THREE_STONE_OBELISKS="%s sees three stone obelisks, one inscribed with a book, one a heart, and one a sword.\n\n*",
FEELS_DRAWN_TO_ONE_OF_THE_OBELISKS="%s feels drawn to one of the obelisks.\n\n",
DOES_TOUCH_THE_OBELISK="Does %s touch the obelisk with the @book, the @heart, or the @sword?\n",
WHICH_BHS="Which (B H S):\n",
FEELS_BETTER="%s feels better.\n",
THE_TONGS_WILL_NOW_PROTECT_YOUR_HANDS="The tongs will now protect your hands.\n",
THEY_WONT_COME_OFF="They won't come off!\n",
OUCH_IT_IS_VERY_HOT="OUCH. It is very hot.\n",
COMPLAINS_OF_TOO_MUCH_LIGHT_AND_INANIMATE="\n%s complains of too much light and inanimate things keep talking.*\n",
COMPLAINS_OF_INANIMATE_THINGS_TALKING="\n%s complains of inanimate things talking.*\n",
COMPLAINS_OF_TOO_MUCH_LIGHT="\n%s complains of too much light.*\n",
YOU_CANT_SLEEP="\nYou can't sleep.\n",
NOBODY_CAN_SLEEP="\nNobody can sleep.\n",
A_MECHANICAL_PERSON_CANT_EAT_BERRIES="A mechanical person can't eat berries!\n",
YOU_EAT_A_MARTIAN_BERRY_YOU_FEEL_AN_INCREASE_IN_THE_STRANGE="You eat a Martian berry. You feel an increase in the strange ",
YOU_EAT_A_MARTIAN_BERRY_YOU_FEEL_A_STRANGE="You eat a Martian berry. You feel a strange ",
RELATIONSHIP_WITH_THINGS_AROUND_YOU="relationship with things around you.\n",
SENSITIVITY_TO_THE_FEELINGS_OF_OBJECTS_AROUND_YOU="sensitivity to the feelings of objects around you.\n",
SENSE_OF_YOUR_SPATIAL_LOCATION="sense of your spatial location.\n",
SUDDEN_FLUSH_DIZZINESS_AND_NAUSEA="sudden flush, dizziness, and nausea.\n",
BREATHES_EASIER="%s breathes easier!\n",
GASPS_FOR_AIR="%s gasps for air!\n",
THE_IS_OUT_OF_FUEL="The %s is out of fuel.\n",
WENT_OUT="%s went out.\n",
FEELS_WARMER="%s feels warmer!\n",
IS_FREEZING="%s is freezing!\n",
A_PSYCHIC_POWER_FADES="A psychic power fades...\n",
SOME_ICE_HAS_MELTED="Some ice has melted.\n",
THE_ICE_IS_MELTING="The ice is melting.\n",
THAT_WOULD_BE_A_WASTE_OUTSIDE_OF_COMBAT="That would be a waste outside of combat.\n",
ENTERS_A_BATTLE_FRENZY="%s enters a battle frenzy!\n",
YOU_SPIT_INTO_THE_SPITTOON="You spit into the spittoon.\n",
YOUR_MOUTH_IS_TOO_DRY="Your mouth is too dry.\n",
CHING="Ching!\n",
SHING="Shing!\n",
WHEEEEZE="Wheeeeze\n",
THUMP_THUMP="Thump! Thump!\n",
GONG="GONG!!!!\n",
YOU_PLAY_THE="You play the %s.\n",
THE_TIME_IS="\n\nThe time is ",
TIME_AM="%d:%02d A.M.\n",
TIME_PM="%d:%02d P.M.\n",
THE_CANNON_WILL_FIRE_STEEL_CANNON_BALLS="The cannon will fire steel cannon balls.\n",
IT_IS_EMPTY="It is empty.\n",
DIRT="dirt",
ROCK="rock",
IRON_ORE="iron ore",
COAL="coal",
IT_HAS_1_LOAD_OF="It has 1 load of %s.\n",
IT_HAS_N_LOADS_OF="It has %d loads of %s.\n",
IT_HAS_A_HEARTSTONE="It has a heartstone.\n",
IT_IS_GROWING="It is growing.\n",
IT_IS_RIPE="It is ripe.\n",
IT_IS_NOT_GROWING="It is not growing.\n",
IT_IS_READY_TO_USE="It is ready to use.\n",
THE_RUBBER_IN_THE_SPRAYER_SYSTEM_IS_NOT_FLESH_COLORED="The rubber in the sprayer system is not flesh-colored.\n",
THERE_IS_NO_RUBBER_IN_THE_SPRAYER_SYSTEM="There is no rubber in the sprayer system.\n",
IT_IS_ALSO_NOT_FLESH_COLORED="It is also not flesh-colored.\n",
CABINET_1="It appears to control a mechanical system.\n",
CABINET_2="It appears to control a valve system.\n",
CABINET_3="It appears to control an electrical system.\n",
CABINET_4="It appears to control this dream machine.\n",
CABINET_5="It appears to control this spray system.\n",
CABINET_6="It appears to control the ruby lens system.\n",
CABINET_7="It appears to control the cistern valves.\n",
YOU_CANNOT_DECIPHER_WHAT_IT_CONTROLS="You cannot decipher what it controls.\n",
IT_APPEARS_TO_BE_LOOSE="It appears to be loose.\n",
IT_APPEARS_TO_BE_BROKEN="It appears to be broken.\n",
THESE_APPEAR_TO_BE_MECHANICAL_CONTROLS="These appear to be mechanical controls.\n",
THESE_APPEAR_TO_BE_VALVE_CONTROLS="These appear to be valve controls.\n",
THESE_APPEAR_TO_BE_ELECTRICAL_CONTROLS="These appear to be electrical controls.\n",
IT_APPEARS_TO_CONTROL_A_DREAM_MACHINE="It appears to control a dream machine.\n",
IT_APPEARS_TO_CONTROL_A_SPRAY_SYSTEM="It appears to control a spray system.\n",
IT_APPEARS_TO_CONTROL_THE_RUBY_LENS_SYSTEM="It appears to control the ruby lens system.\n",
IT_APPEARS_TO_CONTROL_THE_CISTERN_VALVES="It appears to control the cistern valves.\n",
YOU_CANNOT_DECIPHER_ITS_PURPOSE="You cannot decipher its purpose.\n",
THE_PANEL_IS_LOOSE="The panel is loose.\n",
THE_PANEL_IS_BROKEN="The panel is broken.\n",
THE_PANEL_IS_NOT_INSTALLED="The panel is not installed.\n",
IT_IS_LOADED_WITH="It's loaded with %s",
AND_HAS_1_CHARGE_LEFT=", and has 1 charge left.\n",
AND_HAS_N_CHARGES_LEFT=", and has %d charges left.\n",
COMBINATION="Combination",
RIFLE="Rifle",
SHOTGUN="Shotgun",
IT_IS_SET_TO="It is set to '%s'",
IT_SEEMS_TO_BE_LOOSE="\nIt seems to be loose.\n",
IT_IS_NOT_INSTALLED="\nIt is not installed.\n",
PLENTY_OF="\nPlenty of ",
IT_HAS_N_MORE_MINUTES_OF="\nIt has %d more minutes of ",
ITS_ALMOST_OUT_OF="\nIt's almost out of \n",
NO="\nNo ",
FUEL="fuel.\n",
WICK="wick.\n",
IT_IS_APPARENTLY_RUSTED_SHUT="It is apparently rusted shut.\n",
IT_IS_APPARENTLY_LOCKED="It is apparently locked.\n",
THE_SWITCH_IS_FASTENED="The switch is fastened.\n",
THIS_SWITCH_CANNOT_BE_FIXED="This switch cannot be fixed.\n",
THE_SWITCH_IS_LOOSE="The switch is loose.\n",
THE_DRILL_MUST_BE_INSTALLED_ONTO_A_DRILL_CART="The drill must be installed onto a drill cart.\n",
THE_DRILLS_POWER_IS_CONNECTED="The drill's power is connected.\n",
IT_MUST_BE_REPAIRED_FIRST="It must be repaired first.\n",
PANELS_ARE_ONLY_INSTALLED_ONTO_CABINETS="Panels are only installed onto cabinets.\n",
THIS_CABINET_REQUIRES_A_DIFFERENT_TYPE_OF_PANEL="This cabinet requires a different type of panel.\n",
THE_PANEL_IS_FASTENED="The panel is fastened.\n",
YOU_GET_TWO_HANDFULS_OF_OXIUM_FROM_THE_BIN="You get two handfuls of oxium from the bin.\n",
WHICH_SECTION_OF_RAIL_NEEDS_FIXING="Which section of rail needs fixing -\n",
THE_WORK_IS_TO_PRECISE_TO_PERFORM_TELEKINETICALLY="The work is too precise to perform telekinetically.\n",
THE_CABLE_DOES_NOT_NEED_REPLACEMENT="The cable does not need replacement.\n",
OUT_OF_AMMUNITION="Out of ammunition!\n",
OUT_OF_ARROWS="Out of arrows!\n",
OVERCOME_BY_YOUR_WOUNDS_YOU_FALL_UNCONSCIOUS="\nOvercome by your wounds, you fall unconscious..... \n",
YOU_AWAKEN_BACK_AT_FEELING_RESTORED="\n....You awaken back at %s, feeling restored.\n",
THE_SPACE_CAPSULE="the space capsule",
THE_OLYMPUS_MINE="the Olympus mine",
IT_WONT_GO_IN_THAT_DIRECTION="It won't go in that direction!\n",
IT_TURNS_LOOSELY="It turns loosely.\n",
THERE_IS_NOTHING_TO_DRILL_INTO="There is nothing to drill into.\n",
THERE_IS_NO_ROOM_LEFT_FOR_THE_ORE="There is no room left for the ore.\n",
YOU_PUT_THE_ORE_INTO_THE_WHEELBARROW="You put the %s into the wheelbarrow.\n",
YOU_PUT_THE_ORE_INTO_THE_RAIL_CAR="You put the %s into the rail car.\n",
THERE_IS_NO_MORE_ROOM="There is no more room.\n",
YOU_CANT_MIX_DIFFERENT_THINGS_IN_THE_SAME_LOAD="You can't mix different things in the same load!\n",
THERE_IS_NOTHING_TO_UNLOAD="There is nothing to unload.\n",
THE_COAL_BEGINS_MOVING_DOWN_THE_BELT="The coal begins moving down the belt.\n",
OOOPS_THESE_ROLLERS_CAN_NEVER_BE_FIXED="Ooops. These rollers can never be fixed!\n",
STOKERS_PATH_IS_BLOCKED="Stoker's path is blocked!\n",
THE_CONVEYOR_BELT_STOPS="The conveyor belt stops.\n",
ACTOR_DIVIDES="%s divides!\n",
ACTOR_CRITICAL="`%s critical!\n",
ACTOR_HEAVILY_WOUNDED="`%s heavily wounded.\n",
MAN="man",
WOMAN="woman",
HOW_DARE_YOU_YOUNG_PERSON="How dare you, young %s!\n",
ACTOR_GRAZED="`%s grazed.\n",
ACTOR_ATTACKS="`%s attacks!\n",
ACTOR_PARALYZED="`%s paralyzed!\n",
ACTOR_CHARMED="%s charmed!\n",
ACTOR_DODGES="%s dodges!\n",
BECOMES_VISIBLE="`%s becomes visible!\n",
HEAVILY="heavily",
LIGHTLY="lightly",
BARELY="barely",
DAMAGED="damaged",
WOUNDED="wounded",
ACTOR_HIT_MESSAGE="`%s %s %s.\n",
OUT_OF_RANGE="Out of range!\n",
YOUR_HANDS_ARE_FULL="\nYour hands are full.\n",
WONT_BUDGE="Won't budge !\n",
THE_TOTAL_IS_TOO_HEAVY="\nThe total is too heavy.\n",
THAT_IS_NOT_POSSIBLE="That is not possible.\n",
AFFIDAVIT="--Affidavit--\n\nThe undersigned do hereby vouch for the trust-\nworthiness of %s, who bears this chit, commending same to the members of the Olympus Mons outpost.\n\n*Signed:\n%s\n*",
YOU_CANT_READ_IT="You can't read it.\n",
YOU_NEED_A_POOL_CUE="You need a cue stick in your hands.\n",
GOOD_SHOT_OLD_BEAN="Good shot, old bean!\n",
THE_ROBOT_ALREADY_HAS_A_HEARTSTONE="The robot already has a heartstone.\n",
THE_HEARTSTONE_IS_INSTALLED="The heartstone is installed.\n",
THE_ALIEN_DEVICE_DOESNT_SEEM_TO_BE_FUNCTIONING="The alien device doesn't seem to be functioning.\n",
NOTHING_APPEARS_TO_HAPPEN="Nothing appears to happen.\n",
YOU_MUST_BE_IN_PARTY_MODE_TO_ENTER="You must be in party mode to enter.\n",
YOU_MUST_BE_IN_PARTY_MODE_TO_LEAVE="You must be in party mode to leave.\n",
THIS_DOOR_IS_NOT_RUSTED="This door is not rusted.\n",
THE_DOOR_IS_UNSTUCK="The door is unstuck.\n",
YOU_SEE_YOURSELF="You see yourself.\n",
YOU_RECEIVE_A_MOMENTARY_IMPRESSION="You receive a momentary impression of a house falling on you.\n",
THIS_WAS_WORN_LAST_BY_A_DYING_MAN="This was worn last by a dying man.\n",
THIS_WAS_USED_BY_A_MARTIAN_MEASURING_TIME="This was used by a Martian whose entire life focussed on measuring time.\n",
THIS_JEWELRY_WAS_WORN_BY="This jewelry was worn by a distinguished Martian, a mayor or \"Agrarian&.\n",
THESE_FOOTBAGS_WERE_WORN_BY="These footbags were worn by an elderly, distinguished Martian whose job, \"Gatherer&, has no human analog - a combination of undertaker, dustman, and treasurer.\n",
THIS_SAW_WAS_USED_BY="This saw was used by a Martian doctor, or \"Arborist&, for surgery.\n",
THIS_DEVICE_WAS_USED_BY="This device was used by a Martian doctor, or \"Arborist&, on patients.\n",
THIS_SCULPTURE_HAD_GREAT_RITUAL_OR_RELIGIOUS_SIGNIFICANCE="This sculpture had great ritual or religious significance to a city, or \"Grove&, of Martians.  The image is of \"very proud soil&.",
YOU_RECIEVE_NO_PSYCHIC_IMPRESSIONS="You receive no psychic impressions.\n",
THE_RADIUM_HAS_BEEN_INSTALLED="The radium has been installed.\n",
RADIUM_HAS_ALREADY_BEEN_INSTALLED="Radium has already been installed.\n",
THE_OBJ_IS_FULLY_CHARGED="The %s is fully charged.\n",
THE_OBJ_ONLY_NEEDED_N_RADIUM_BLOCKS="The %s only needed %d radium blocks.\n",
THE_HEADGEAR_IS_INSTALLED="The headgear is installed.\n",
YOU_ARE_COMPLETELY_UNSURE_WHAT_YOU_JUST_DID="You are completely unsure what you just did.\n",
LIGHTS_FLASH_AND_CHANGE_COLOR_BUT_NOTHING_ELSE_HAPPENS="Lights flash and change color but nothing else happens.\n",
STATUS_LIGHTS_CHANGE_BUT_YOU_SEE_NO_OTHER_EFFECT="Status lights change but you see no other effect.\n",
YOU_ACTUATE_THE_MECHANISM_TO_NO_APPARENT_EFFECT="You actuate the mechanism to no apparent effect.\n",
THE_DREAM_MACHINES_SEEM_TO_HAVE_CEASED_FUNCTIONING="The dream machines seem to have ceased functioning.\n",
THE_MACHINE_DOES_NOT_WORK="The machine does not work.\n",
THERE_IS_NOBODY_SITTING_IN_THE_MACHINE="There is nobody sitting in the machine.\n",
ACTOR_DREAMS_UNTIL_DAWN_WHILE_THE_PARTY_WAITS="%s dreams until dawn while the party waits.\n",
YOU_TRY_TO_WAKE_YOURSELF_UP="You try to wake yourself up.\n",
YOU_WAKE_UP="You wake up.\n",
YOU_ARENT_YET_THAT_INSANE="You aren't yet that insane.\n",
NO_BACKTALK_FROM_YOU_YOUNG_MAN="No backtalk from you, young man!\n",
NO_BACKTALK_FROM_YOU_YOUNG_WOMAN="No backtalk from you, young man!\n",
YOU_FALL_OFF_THE_CLIFF="You fall off the cliff!\n",
YOU_SHAKE_YOURSELF_AWAKE_FROM_THE_NIGHTMARE="You shake yourself awake from the nightmare.\n",
YOU_FEEL_YOUR_DREAM_CONSCIOUSNESS_RETURNING="You feel your dream consciousness returning.\n",
WHAT_AN_ODD_LEVER="What an odd lever!\n",
STRANGELY_NOTHING_HAPPENS="Strangely, nothing happens.\n",
STRANGELY_IT_DOESNT_MOVE="Strangely, it doesn't move.\n",
THE_POD_SPLITS_OPEN="The pod splits open!\n",
THE_ICE_THAWS="The ice thaws.\n",
THE_WATER_FREEZES="The water freezes.\n",
ATTACK_WITH_BARE_HANDS="Attack with bare hands-",
ATTACK_WITH_WEAPON="Attack with %s-",
FASTER="...faster!\n",
YOU_IMAGINE="You imagine ",
SPRAY_GUN_GREEN_PAINT="The spray gun now contains 20 charges of green paint.\n",
SPRAY_GUN_WEED_KILLER="The spray gun now contains 10 charges of weed killer.\n",
SPRAY_GUN_10_MORE_CHARGES="The spray gun has 10 more charges.\n",
THERE_IS_NOTHING_IN_THE_GUN="There is nothing in the gun.\n",
YOU_KILLED_THE_PLANT="You killed the plant.\n",
}
