static const char *const k1IntroStringsDOSDemoEnglish[32] = {
	"This is a text test - 1",
	"This is a text test - 2",
	"Greetings, Kallak,",
	"Powerful leader of the Royal Mystics!",
	"Did I frighten you?",
	"Malcom!",
	"I heard yesterday of your escape.",
	"And have expected you before now.",
	"Why would I rush?",
	"I rule this land.",
	"Your puny curse restrains me naught.",
	"So, slay me now.",
	"I have little magic left.",
	"I mean to harm you, have no doubt,",
	"Now you are stone,",
	"Yet I leave your eyes.",
	"I shed no tears for Kyrandia,",
	"But cannot deny you yours.",
	"Malcom the jester has broken free.",
	"He now controls the Kyragem",
	"source of all magic in Kyrandia.",
	"",
	"How dare you trip ME!",
	"But no,...something special for you.",
	"A humorous hex!",
	"But, I am fair!",
	"And I shall give fair warning!",
	"Don't climb up THAT tree little squirrel!",
	"Boo!",
	"Tis funnier that way,",
	"Is it not?",
	"Ha ha ha..."
};

static const StringListProvider k1IntroStringsDOSDemoEnglishProvider = { ARRAYSIZE(k1IntroStringsDOSDemoEnglish), k1IntroStringsDOSDemoEnglish };

