static const char *const k1IntroStringsMacEnglish[51] = {
	"This is a text test - 1",
	"This is a text test - 2",
	"Greetings, Kallak...",
	"...Powerful leader of the Royal Mystics!",
	"Did I frighten you?",
	"Malcolm!",
	"I heard yesterday of your escape.",
	"And have expected you before now.",
	"Why would I rush?",
	"I rule this land!",
	"Your puny curse restrains me naught.",
	"So, slay me now!",
	"I have little magic left.",
	"I mean to harm you, have no doubt...",
	"...Now you are stone...",
	"...Yet I leave your eyes.",
	"I shed no tears for Kyrandia...",
	"...But cannot deny you yours.",
	"Malcolm the Jester has broken free!",
	"He now controls the Kyragem...",
	"...source of all magic in Kyrandia!",
	"",
	"How dare you trip ME!",
	"But no, something special for you...",
	"A humorous hex!",
	"But I shall give fair warning.",
	"Don't jump on THAT tree!",
	"Don't climb up THAT tree little squirrel!",
	"Boo!",
	"Tis funnier that way...",
	"...Is it not?",
	"Ha ha ha...",
	"Aren't fingers wonderful?",
	"It's great to be alive!",
	"Brandon!",
	"Grandfather! We did it!",
	"No! YOU did it!",
	"Long live King Brandon!",
	"Long live the Royal Mystics!",
	"Now, let's go put Kyrandia back together!",
	"Good idea!",
	"As my first Royal Proclamation...",
	"I name sandals as the official footwear of Kyrandia!",
	"Well done, Brandon!",
	"The Land no longer weeps...",
	"And your friends have been restored!",
	"Dear Brynn...",
	"Malcolm has broken free!",
	"Soon he will come for me.",
	"Please help Brandon...",
	""
};

static const StringListProvider k1IntroStringsMacEnglishProvider = { ARRAYSIZE(k1IntroStringsMacEnglish), k1IntroStringsMacEnglish };

static const char *const k1ItemNamesMacEnglish[107] = {
	"Garnet",
	"Amethyst",
	"Aquamarine",
	"Diamond",
	"Emerald",
	"Pearl",
	"Ruby",
	"Peridot",
	"Sapphire",
	"Opal",
	"Topaz",
	"Onyx",
	"Sunstone",
	"Moonstone",
	"Rainbowstone",
	"Lodestone",
	"Rose",
	"Tulip",
	"Orchid",
	"Silver Rose",
	"Silver Statuette",
	"Silver Coin",
	"Gold Coin",
	"Gold Ring",
	"Royal Chalice",
	"Pinecone",
	"Acorn",
	"Walnut",
	"Everglowing Fireberries",
	"Fireberries",
	"Fireberries",
	"Fireberries",
	"Fireberries",
	"Fireberries",
	"Fish",
	"Fish Bone",
	"Mutton Leg",
	"Bone",
	"Apple",
	"Apple Core",
	"Blueberries",
	"Mushroom",
	"Note",
	"Marble",
	"Saw",
	"Ankh",
	"Feather",
	"Egg",
	"Leaf",
	"Shamrock",
	"Fallen Star",
	"Crystal Ball",
	"Teardrop",
	"Mirror",
	"Ice Shard",
	"Flute",
	"Hourglass",
	"Iron Key",
	"Jade Key",
	"Obsidion Key",
	"Red Potion",
	"Red Potion",
	"Blue Potion",
	"Blue Potion",
	"Yellow Potion",
	"Yellow Potion",
	"Green Potion",
	"Orange Potion",
	"Purple Potion",
	"Potion of Rainbows",
	"Fresh Water",
	"Fresh Water",
	"Salt Water",
	"Salt Water",
	"Mineral Water",
	"Mineral Water",
	"Magical Water",
	"Magical Water",
	"Empty Flask",
	"Empty Flask",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Parchment scrap",
	"Parchment scrap",
	"Parchment scrap",
	"Parchment scrap",
	"Parchment scrap",
	"Red Magestone",
	"Orange Magestone",
	"Yellow Magestone",
	"Green Magestone",
	"Blue-Green Magestone",
	"Blue Magestone",
	"Purple Magestone",
	"Heavy rock",
	"Royal Crown",
	"Royal Sceptre",
	"Gold key",
	"Unknown item"
};

static const StringListProvider k1ItemNamesMacEnglishProvider = { ARRAYSIZE(k1ItemNamesMacEnglish), k1ItemNamesMacEnglish };

static const char *const k1TakenStringsMacEnglish[2] = {
	" taken.",
	" taken."
};

static const StringListProvider k1TakenStringsMacEnglishProvider = { ARRAYSIZE(k1TakenStringsMacEnglish), k1TakenStringsMacEnglish };

static const char *const k1PlacedStringsMacEnglish[1] = {
	" placed."
};

static const StringListProvider k1PlacedStringsMacEnglishProvider = { ARRAYSIZE(k1PlacedStringsMacEnglish), k1PlacedStringsMacEnglish };

static const char *const k1DroppedStringsMacEnglish[1] = {
	" dropped."
};

static const StringListProvider k1DroppedStringsMacEnglishProvider = { ARRAYSIZE(k1DroppedStringsMacEnglish), k1DroppedStringsMacEnglish };

static const char *const k1NoDropStringsMacEnglish[2] = {
	"You can't drop any more items in this scene.",
	"You can't put that there."
};

static const StringListProvider k1NoDropStringsMacEnglishProvider = { ARRAYSIZE(k1NoDropStringsMacEnglish), k1NoDropStringsMacEnglish };

static const char *const k1PutDownStringMacEnglish[1] = {
	"Maybe I'd better put this down first."
};

static const StringListProvider k1PutDownStringMacEnglishProvider = { ARRAYSIZE(k1PutDownStringMacEnglish), k1PutDownStringMacEnglish };

static const char *const k1WaitAmuletStringMacEnglish[1] = {
	"I guess I'll have to wait for my Amulet to regain its power."
};

static const StringListProvider k1WaitAmuletStringMacEnglishProvider = { ARRAYSIZE(k1WaitAmuletStringMacEnglish), k1WaitAmuletStringMacEnglish };

static const char *const k1BlackJewelStringMacEnglish[1] = {
	"It's a jewel, but why is it black?"
};

static const StringListProvider k1BlackJewelStringMacEnglishProvider = { ARRAYSIZE(k1BlackJewelStringMacEnglish), k1BlackJewelStringMacEnglish };

static const char *const k1HealingTipStringMacEnglish[1] = {
	"Gee, this could come in handy if I get hurt."
};

static const StringListProvider k1HealingTipStringMacEnglishProvider = { ARRAYSIZE(k1HealingTipStringMacEnglish), k1HealingTipStringMacEnglish };

static const char *const k1PoisonGoneStringMacEnglish[2] = {
	"Amazing!",
	"The effects of the poison are gone!"
};

static const StringListProvider k1PoisonGoneStringMacEnglishProvider = { ARRAYSIZE(k1PoisonGoneStringMacEnglish), k1PoisonGoneStringMacEnglish };

static const char *const k1ThePoisonStringsMacEnglish[4] = {
	"The poison...",
	"I can't breathe...",
	"I don't feel so good...",
	"That snake must\nbe poisonous!"
};

static const StringListProvider k1ThePoisonStringsMacEnglishProvider = { ARRAYSIZE(k1ThePoisonStringsMacEnglish), k1ThePoisonStringsMacEnglish };

static const char *const k1FluteStringsMacEnglish[2] = {
	"Not very impressive sounding.",
	"That last note sure was high!"
};

static const StringListProvider k1FluteStringsMacEnglishProvider = { ARRAYSIZE(k1FluteStringsMacEnglish), k1FluteStringsMacEnglish };

static const char *const k1WispJewelStringsMacEnglish[3] = {
	"I'm not really feeling up to it right now.",
	"I probably should put down my ",
	", first."
};

static const StringListProvider k1WispJewelStringsMacEnglishProvider = { ARRAYSIZE(k1WispJewelStringsMacEnglish), k1WispJewelStringsMacEnglish };

static const char *const k1MagicJewelStringsMacEnglish[1] = {
	"That felt strange."
};

static const StringListProvider k1MagicJewelStringsMacEnglishProvider = { ARRAYSIZE(k1MagicJewelStringsMacEnglish), k1MagicJewelStringsMacEnglish };

static const char *const k1FlaskFullStringMacEnglish[1] = {
	"This flask is already full."
};

static const StringListProvider k1FlaskFullStringMacEnglishProvider = { ARRAYSIZE(k1FlaskFullStringMacEnglish), k1FlaskFullStringMacEnglish };

static const char *const k1FullFlaskStringMacEnglish[4] = {
	"The flask is now filled\nwith fresh, sparkling water.",
	"The flask is now filled\nwith salt water.",
	"The flask is now filled\nwith mineral water.",
	"Magic water."
};

static const StringListProvider k1FullFlaskStringMacEnglishProvider = { ARRAYSIZE(k1FullFlaskStringMacEnglish), k1FullFlaskStringMacEnglish };

static const char *const k1OutroHomeStringMacEnglish[1] = {
	"Home"
};

static const StringListProvider k1OutroHomeStringMacEnglishProvider = { ARRAYSIZE(k1OutroHomeStringMacEnglish), k1OutroHomeStringMacEnglish };

static const char *const k1VeryCleverStringMacEnglish[1] = {
	"Very clever!  But your feeble efforts are futile."
};

static const StringListProvider k1VeryCleverStringMacEnglishProvider = { ARRAYSIZE(k1VeryCleverStringMacEnglish), k1VeryCleverStringMacEnglish };

static const char *const k1GUIStringsMacEnglish[28] = {
	"The Legend of Kyrandia",
	"Load a game",
	"Save this game",
	"Game controls",
	"Quit playing",
	"Resume game",
	"Game Controls",
	"Which game would you like to reload?",
	"Select a position to save to:",
	"[ EMPTY SLOT ]",
	"Cancel",
	"Enter a description of your saved game:",
	"Save",
	"Rest in peace, Brandon.",
	"Are you sure you want to quit playing?",
	"XXX",
	"XXX",
	"XXXXXXX",
	"XXXXXXXXX",
	"Main menu",
	"on",
	"off",
	"Yes",
	"No",
	"Walk speed ",
	"Text speed ",
	"Music is ",
	"Sounds are "
};

static const StringListProvider k1GUIStringsMacEnglishProvider = { ARRAYSIZE(k1GUIStringsMacEnglish), k1GUIStringsMacEnglish };

static const char *const k1NewGameStringMacEnglish[1] = {
	"[ START A NEW GAME ]"
};

static const StringListProvider k1NewGameStringMacEnglishProvider = { ARRAYSIZE(k1NewGameStringMacEnglish), k1NewGameStringMacEnglish };

static const char *const k1ConfigStringsMacEnglish[9] = {
	"slowest",
	"slow",
	"normal",
	"fast",
	"fastest",
	"slowest",
	"normal",
	"fast",
	"clickable"
};

static const StringListProvider k1ConfigStringsMacEnglishProvider = { ARRAYSIZE(k1ConfigStringsMacEnglish), k1ConfigStringsMacEnglish };

static const char *const k1ConfigStrings2MacEnglish[3] = {
	"None",
	"High Quality",
	"Low Impact"
};

static const StringListProvider k1ConfigStrings2MacEnglishProvider = { ARRAYSIZE(k1ConfigStrings2MacEnglish), k1ConfigStrings2MacEnglish };

static const uint8 k1CreditsStringsMacEnglish[] = {
	0x0a, 0x02, 0x54, 0x48, 0x45, 0x20, 0x45, 0x4e,
	0x44, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x0a, 0x0a, 0x02, 0x46, 0x61, 0x62, 0x6c, 0x65,
	0x73, 0x20, 0x26, 0x20, 0x46, 0x69, 0x65, 0x6e,
	0x64, 0x73, 0x0a, 0x02, 0x42, 0x6f, 0x6f, 0x6b,
	0x20, 0x49, 0x0a, 0x02, 0x54, 0x48, 0x45, 0x20,
	0x4c, 0x45, 0x47, 0x45, 0x4e, 0x44, 0x20, 0x4f,
	0x46, 0x20, 0x4b, 0x59, 0x52, 0x41, 0x4e, 0x44,
	0x49, 0x41, 0x0a, 0x0a, 0x50, 0x75, 0x62, 0x6c,
	0x69, 0x73, 0x68, 0x65, 0x64, 0x20, 0x62, 0x79,
	0x0a, 0x02, 0x57, 0x65, 0x73, 0x74, 0x77, 0x6f,
	0x6f, 0x64, 0x20, 0x53, 0x74, 0x75, 0x64, 0x69,
	0x6f, 0x73, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x03, 0x44, 0x69, 0x72, 0x65, 0x63, 0x74, 0x6f,
	0x72, 0x3a, 0x05, 0x04, 0x42, 0x72, 0x65, 0x74,
	0x74, 0x20, 0x57, 0x2e, 0x20, 0x53, 0x70, 0x65,
	0x72, 0x72, 0x79, 0x0a, 0x0a, 0x03, 0x44, 0x65,
	0x73, 0x69, 0x67, 0x6e, 0x3a, 0x05, 0x04, 0x4d,
	0x69, 0x63, 0x68, 0x61, 0x65, 0x6c, 0x20, 0x4c,
	0x65, 0x67, 0x67, 0x0a, 0x04, 0x52, 0x69, 0x63,
	0x6b, 0x20, 0x50, 0x61, 0x72, 0x6b, 0x73, 0x0a,
	0x04, 0x50, 0x61, 0x75, 0x6c, 0x20, 0x53, 0x2e,
	0x20, 0x4d, 0x75, 0x64, 0x72, 0x61, 0x0a, 0x0a,
	0x03, 0x4c, 0x65, 0x61, 0x64, 0x20, 0x50, 0x72,
	0x6f, 0x67, 0x72, 0x61, 0x6d, 0x6d, 0x65, 0x72,
	0x3a, 0x05, 0x04, 0x4d, 0x69, 0x63, 0x68, 0x61,
	0x65, 0x6c, 0x20, 0x4c, 0x65, 0x67, 0x67, 0x0a,
	0x0a, 0x03, 0x53, 0x63, 0x72, 0x69, 0x70, 0x74,
	0x20, 0x50, 0x72, 0x6f, 0x67, 0x72, 0x61, 0x6d,
	0x6d, 0x65, 0x72, 0x3a, 0x05, 0x04, 0x53, 0x63,
	0x6f, 0x74, 0x74, 0x20, 0x42, 0x6f, 0x77, 0x65,
	0x6e, 0x0a, 0x0a, 0x03, 0x41, 0x6d, 0x69, 0x67,
	0x61, 0x20, 0x50, 0x72, 0x6f, 0x67, 0x72, 0x61,
	0x6d, 0x6d, 0x65, 0x72, 0x3a, 0x05, 0x04, 0x42,
	0x69, 0x6c, 0x6c, 0x20, 0x53, 0x74, 0x6f, 0x6b,
	0x65, 0x73, 0x0a, 0x0a, 0x03, 0x49, 0x6e, 0x74,
	0x72, 0x6f, 0x20, 0x26, 0x20, 0x46, 0x69, 0x6e,
	0x61, 0x6c, 0x65, 0x3a, 0x05, 0x04, 0x43, 0x68,
	0x72, 0x69, 0x73, 0x74, 0x6f, 0x70, 0x68, 0x65,
	0x72, 0x20, 0x59, 0x61, 0x74, 0x65, 0x73, 0x0a,
	0x0a, 0x03, 0x41, 0x72, 0x74, 0x20, 0x26, 0x20,
	0x47, 0x72, 0x61, 0x70, 0x68, 0x69, 0x63, 0x73,
	0x3a, 0x05, 0x04, 0x52, 0x69, 0x63, 0x6b, 0x20,
	0x50, 0x61, 0x72, 0x6b, 0x73, 0x0a, 0x04, 0x52,
	0x65, 0x6e, 0x20, 0x4f, 0x6c, 0x73, 0x65, 0x6e,
	0x0a, 0x04, 0x4c, 0x6f, 0x75, 0x69, 0x73, 0x65,
	0x20, 0x53, 0x61, 0x6e, 0x64, 0x6f, 0x76, 0x61,
	0x6c, 0x0a, 0x04, 0x4a, 0x6f, 0x73, 0x65, 0x70,
	0x68, 0x20, 0x48, 0x65, 0x77, 0x69, 0x74, 0x74,
	0x20, 0x49, 0x56, 0x0a, 0x04, 0x4a, 0x75, 0x64,
	0x69, 0x74, 0x68, 0x20, 0x50, 0x65, 0x74, 0x65,
	0x72, 0x73, 0x6f, 0x6e, 0x0a, 0x04, 0x41, 0x61,
	0x72, 0x6f, 0x6e, 0x20, 0x50, 0x6f, 0x77, 0x65,
	0x6c, 0x6c, 0x0a, 0x04, 0x45, 0x6c, 0x69, 0x65,
	0x20, 0x41, 0x72, 0x61, 0x62, 0x69, 0x61, 0x6e,
	0x0a, 0x04, 0x45, 0x72, 0x69, 0x63, 0x20, 0x53,
	0x68, 0x75, 0x6c, 0x74, 0x73, 0x0a, 0x0a, 0x03,
	0x41, 0x75, 0x64, 0x69, 0x6f, 0x20, 0x44, 0x69,
	0x72, 0x65, 0x63, 0x74, 0x6f, 0x72, 0x3a, 0x05,
	0x04, 0x50, 0x61, 0x75, 0x6c, 0x20, 0x53, 0x2e,
	0x20, 0x4d, 0x75, 0x64, 0x72, 0x61, 0x0a, 0x0a,
	0x03, 0x4d, 0x75, 0x73, 0x69, 0x63, 0x3a, 0x05,
	0x04, 0x46, 0x72, 0x61, 0x6e, 0x6b, 0x20, 0x4b,
	0x6c, 0x65, 0x70, 0x61, 0x63, 0x6b, 0x69, 0x0a,
	0x0a, 0x03, 0x53, 0x6f, 0x75, 0x6e, 0x64, 0x20,
	0x45, 0x66, 0x66, 0x65, 0x63, 0x74, 0x73, 0x3a,
	0x05, 0x04, 0x44, 0x77, 0x69, 0x67, 0x68, 0x74,
	0x20, 0x4f, 0x6b, 0x61, 0x68, 0x61, 0x72, 0x61,
	0x0a, 0x0a, 0x03, 0x57, 0x72, 0x69, 0x74, 0x65,
	0x72, 0x3a, 0x05, 0x04, 0x22, 0x43, 0x6f, 0x63,
	0x6f, 0x22, 0x0a, 0x0a, 0x03, 0x51, 0x75, 0x61,
	0x6c, 0x69, 0x74, 0x79, 0x20, 0x41, 0x73, 0x73,
	0x75, 0x72, 0x61, 0x6e, 0x63, 0x65, 0x3a, 0x05,
	0x04, 0x47, 0x6c, 0x65, 0x6e, 0x6e, 0x20, 0x53,
	0x70, 0x65, 0x72, 0x72, 0x79, 0x0a, 0x04, 0x4d,
	0x61, 0x74, 0x74, 0x20, 0x43, 0x6f, 0x6c, 0x6c,
	0x69, 0x6e, 0x73, 0x0a, 0x04, 0x42, 0x69, 0x6c,
	0x6c, 0x20, 0x46, 0x6f, 0x73, 0x74, 0x65, 0x72,
	0x0a, 0x04, 0x4d, 0x69, 0x63, 0x68, 0x61, 0x65,
	0x6c, 0x20, 0x4c, 0x69, 0x67, 0x68, 0x74, 0x6e,
	0x65, 0x72, 0x0a, 0x04, 0x4d, 0x69, 0x63, 0x68,
	0x61, 0x65, 0x6c, 0x20, 0x47, 0x61, 0x74, 0x65,
	0x72, 0x0a, 0x04, 0x45, 0x75, 0x67, 0x65, 0x6e,
	0x65, 0x20, 0x4d, 0x61, 0x72, 0x74, 0x69, 0x6e,
	0x0a, 0x04, 0x4d, 0x69, 0x63, 0x68, 0x61, 0x65,
	0x6c, 0x20, 0x47, 0x6c, 0x6f, 0x73, 0x65, 0x63,
	0x6b, 0x69, 0x0a, 0x04, 0x4a, 0x75, 0x73, 0x74,
	0x69, 0x6e, 0x20, 0x4e, 0x6f, 0x72, 0x72, 0x0a,
	0x04, 0x54, 0x6f, 0x70, 0x20, 0x53, 0x74, 0x61,
	0x72, 0x20, 0x54, 0x65, 0x73, 0x74, 0x69, 0x6e,
	0x67, 0x0a, 0x04, 0x54, 0x65, 0x73, 0x74, 0x69,
	0x6e, 0x67, 0x20, 0x31, 0x2c, 0x20, 0x32, 0x2c,
	0x20, 0x33, 0x2e, 0x2e, 0x2e, 0x0a, 0x0a, 0x4d,
	0x61, 0x63, 0x69, 0x6e, 0x74, 0x6f, 0x73, 0x68,
	0x20, 0x56, 0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e,
	0x20, 0x42, 0x79, 0x0a, 0x54, 0x68, 0x65, 0x20,
	0x44, 0x72, 0x65, 0x61, 0x6d, 0x65, 0x72, 0x73,
	0x20, 0x47, 0x75, 0x69, 0x6c, 0x64, 0x2c, 0x20,
	0x49, 0x6e, 0x63, 0x2e, 0x0a, 0x0a, 0x03, 0x50,
	0x72, 0x6f, 0x67, 0x72, 0x61, 0x6d, 0x6d, 0x65,
	0x72, 0x73, 0x3a, 0x05, 0x04, 0x4a, 0x6f, 0x65,
	0x20, 0x50, 0x65, 0x61, 0x72, 0x63, 0x65, 0x0a,
	0x04, 0x4a, 0x6f, 0x65, 0x20, 0x42, 0x75, 0x72,
	0x6b, 0x73, 0x0a, 0x03, 0x54, 0x65, 0x73, 0x74,
	0x69, 0x6e, 0x67, 0x3a, 0x05, 0x04, 0x53, 0x74,
	0x65, 0x76, 0x65, 0x6e, 0x20, 0x4d, 0x63, 0x4e,
	0x61, 0x6c, 0x6c, 0x79, 0x0a, 0x0a, 0x0a, 0x0a,
	0x0a, 0x53, 0x70, 0x65, 0x63, 0x69, 0x61, 0x6c,
	0x20, 0x54, 0x68, 0x61, 0x6e, 0x6b, 0x73, 0x20,
	0x74, 0x6f, 0x0a, 0x4c, 0x79, 0x6c, 0x65, 0x20,
	0x4a, 0x2e, 0x20, 0x48, 0x61, 0x6c, 0x6c, 0x0a,
	0x44, 0x61, 0x76, 0x69, 0x64, 0x20, 0x42, 0x69,
	0x73, 0x68, 0x6f, 0x70, 0x0a, 0x22, 0x54, 0x68,
	0x65, 0x20, 0x44, 0x6f, 0x63, 0x22, 0x0a, 0x61,
	0x6e, 0x64, 0x0a, 0x46, 0x69, 0x72, 0x65, 0x62,
	0x65, 0x72, 0x72, 0x79, 0x20, 0x47, 0x72, 0x65,
	0x65, 0x6e, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x02, 0x43, 0x41, 0x53, 0x54, 0x20, 0x4f, 0x46,
	0x20, 0x43, 0x48, 0x41, 0x52, 0x41, 0x43, 0x54,
	0x45, 0x52, 0x53, 0x0a, 0x0a, 0x4d, 0x61, 0x6c,
	0x63, 0x6f, 0x6c, 0x6d, 0x20, 0x74, 0x68, 0x65,
	0x20, 0x4a, 0x65, 0x73, 0x74, 0x65, 0x72, 0x0a,
	0x42, 0x72, 0x61, 0x6e, 0x64, 0x6f, 0x6e, 0x20,
	0x74, 0x68, 0x65, 0x20, 0x42, 0x6f, 0x6c, 0x64,
	0x0a, 0x42, 0x72, 0x61, 0x6e, 0x64, 0x79, 0x77,
	0x69, 0x6e, 0x65, 0x20, 0x61, 0x6e, 0x64, 0x20,
	0x44, 0x61, 0x72, 0x6d, 0x0a, 0x50, 0x72, 0x69,
	0x65, 0x73, 0x74, 0x65, 0x73, 0x73, 0x20, 0x42,
	0x72, 0x79, 0x6e, 0x6e, 0x0a, 0x5a, 0x61, 0x6e,
	0x74, 0x68, 0x69, 0x61, 0x2c, 0x20, 0x4c, 0x61,
	0x64, 0x79, 0x20, 0x6f, 0x66, 0x20, 0x41, 0x6c,
	0x63, 0x68, 0x65, 0x6d, 0x79, 0x0a, 0x48, 0x65,
	0x72, 0x6d, 0x61, 0x6e, 0x20, 0x74, 0x68, 0x65,
	0x20, 0x48, 0x61, 0x6e, 0x64, 0x79, 0x6d, 0x61,
	0x6e, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x0a, 0x0a, 0x0a, 0x0a, 0x01, 0x41, 0x6e, 0x79,
	0x20, 0x63, 0x68, 0x61, 0x72, 0x61, 0x63, 0x74,
	0x65, 0x72, 0x73, 0x20, 0x69, 0x6e, 0x20, 0x74,
	0x68, 0x69, 0x73, 0x20, 0x67, 0x61, 0x6d, 0x65,
	0x20, 0x77, 0x68, 0x6f, 0x20, 0x62, 0x65, 0x61,
	0x72, 0x0a, 0x01, 0x72, 0x65, 0x73, 0x65, 0x6d,
	0x62, 0x6c, 0x61, 0x6e, 0x63, 0x65, 0x20, 0x74,
	0x6f, 0x20, 0x70, 0x65, 0x72, 0x73, 0x6f, 0x6e,
	0x73, 0x20, 0x6c, 0x69, 0x76, 0x69, 0x6e, 0x67,
	0x20, 0x6f, 0x72, 0x0a, 0x01, 0x64, 0x65, 0x61,
	0x64, 0x20, 0x61, 0x72, 0x65, 0x20, 0x70, 0x75,
	0x72, 0x65, 0x6c, 0x79, 0x20, 0x63, 0x6f, 0x69,
	0x6e, 0x63, 0x69, 0x64, 0x65, 0x6e, 0x74, 0x61,
	0x6c, 0x2e, 0x0a, 0x0a, 0x0a, 0x01, 0x43, 0x6f,
	0x70, 0x79, 0x72, 0x69, 0x67, 0x68, 0x74, 0x20,
	0x28, 0x43, 0x29, 0x20, 0x31, 0x39, 0x39, 0x32,
	0x2c, 0x31, 0x39, 0x39, 0x33, 0x20, 0x57, 0x65,
	0x73, 0x74, 0x77, 0x6f, 0x6f, 0x64, 0x20, 0x53,
	0x74, 0x75, 0x64, 0x69, 0x6f, 0x73, 0x2c, 0x20,
	0x49, 0x6e, 0x63, 0x2e, 0x0a, 0x01, 0x41, 0x6c,
	0x6c, 0x20, 0x72, 0x69, 0x67, 0x68, 0x74, 0x73,
	0x20, 0x72, 0x65, 0x73, 0x65, 0x72, 0x76, 0x65,
	0x64, 0x2e, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x0a, 0x00
};

static const ByteProvider k1CreditsStringsMacEnglishProvider = { ARRAYSIZE(k1CreditsStringsMacEnglish), k1CreditsStringsMacEnglish };
