/* ScummVM - Graphic Adventure Engine
 *
 * ScummVM is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Based on the original sources
 *   Faery Tale II -- The Halls of the Dead
 *   (c) 1993-1996 The Wyrmkeep Entertainment Co.
 */

#ifndef SAGA2_FONTLIB_H
#define SAGA2_FONTLIB_H

namespace Saga2 {

//  The Agate family

extern gFont        Agate6Font,
       Agate7Font,
       Agate8Font,
       Agate11Font;

//  The Helv family (Oh, yeah, like that's it's real name...)

extern gFont        Helv7Font,
       Helv8Font,
       Helv9Font,
       Helv11Font,
       Helv13Font,
       Helv15Font,
       Helv18Font,
       Helv24Font;

//  The Juno family

extern gFont        Juno10Font;

//  The Onyx family

extern gFont        Onyx10Font;

//  The Plate family

extern gFont        Plate18Font;

//  The Script family

extern gFont        Script10Font;

//  The Amber family

extern gFont        Amber9Font,
       Amber11Font,
       Amber13Font;

//  Fixed-width fonts

extern gFont        Fix8Font,
       ThinFix8Font;

} // namespace Saga2

#endif
