begin_section("TwinE");
	add_person("Alexandre Fontoura", "xesf", "(retired)");
	add_person("Vincent Hamm", "yaz0r", "(retired)");
	add_person("Felipe Sanches", "jucablues", "(retired)");
	add_person("Nikita Tereshin", "rumkex", "(retired)");
	add_person("Patrik Dahlstr&ouml;m", "risca", "(retired)");
	add_person("Arthur Blot", "arthur.blot78", "(retired)");
	add_person("Kyuubu", "wett", "(retired)");
	add_person("To&euml;l Hartmann", "toel__", "(retired)");
	add_person("Seb&aacute;stien Viannay", "", "(retired)");
	add_person("Martin Gerhardy", "mgerhardy", "");
end_section();
