begin_section("Dragons");
	add_person("Eric Fry", "yuv422", "");
	add_person("Benjamin Haisch", "john_doe", "Actor pathfinding");
	add_person("&Aacute;ngel Eduardo Garc&iacute;a Hern&aacute;ndez", "arcnor", "Help with reverse engineering");
end_section();
