static const char *const k2SeqplayStringsDOSCDSpanish[104] = {
	"Kyrandia esta desapareciendo!",
	"Roca tras roca...",
	"...y ""\xA0""rbol tras ""\xA0""rbol.",
	"Kyrandia deja de existir!",
	"Los M""\xA1""sticos Reales estan desconcertados.",
	"Cada decision tiene que ser consultada.",
	"Hasta Marko y su nuevo ayudante han sido autorizados a entrar en la conferencia.",
	"Por suerte, la Mano tenia experiencia en estos asuntos.",
	"Y finalmente el plan fue aprovado...",
	"...el plan requiria que una Piedra Ancla m""\xA0""gica...",
	"...fuera extraida del centro del mundo.",
	"Zanthia, la mas joven de los M""\xA1""sticos de Kyrandia, ha sido elegida para recoger la Piedra.",
	"Gracias por jugar The Hand of Fate.",
	"Estos ar""\xA0""ndanos deber""\xA1""an bastar para abrir un portal al centro del mundo.",
	" DUMMY STRING... ",
	" DUMMY STRING... ",
	"Hey! Me han robado todo mi equipo!",
	" DUMMY STRING... ",
	"Si piensan que voy a ir caminando hasta all""\xA1"" est""\xA0""n locos!",
	" DUMMY STRING... ",
	" DUMMY STRING... ",
	"Vamos Faun!",
	"Chico, eso estuvo cerca!",
	"Me lo dijiste. Por mi parte, no pienso volver a cazar jamas!",
	"Ribbit.",
	"Cuantas veces tengo que repetirtelo? Eres un sapo.",
	"Oh no! Nos hemos quedado sin queso!",
	"Probemos este cerumen. Es naranja.",
	"Mam""\xA0"", cuando tendr""\x82"" la hiedra?",
	"Fuera de aqu""\xA1"", Uhhh!",
	"Tu cortas, y yo elijo.",
	"No. TU cortas y YO elijo.",
	"Sigo diciendo que no tenia ni idea de lo que hablaba.",
	"No reconocerias un pent""\xA0""metro y""\xA0""mbico ni aunque te picara en el trasero!",
	"Productor Ejecutivo",
	"Brett W. Sperry",
	"Direcci""\xA2""n y Diseno",
	"Rick Gush",
	"Programador Principal",
	"Michael Legg",
	"Gestion artistica",
	"Louis Castle",
	"Joseph B. Hewitt IV",
	"Artista principal",
	"Rick Parks",
	"Programaci""\xA2""n adicional",
	"Philip W. Gorrow",
	"Mike Grayford",
	"Mark McCubbin",
	"Artistas",
	"Cameron Chun",
	"Cary Averett",
	"Cindy Chinn",
	"Elie Arabian",
	"Fei Cheng",
	"Ferby Miguel",
	"Frank Mendeola",
	"Jack Martin",
	"Jerry Moore",
	"DUMMY STRING... ",
	"Judith Peterson",
	"Larry Miller",
	"Lenny Lee",
	"Louise Sandoval",
	"Ren Olsen",
	"Musica y Sonidos",
	"Paul Mudra",
	"Frank Klepacki",
	"Dwight Okahara",
	"Pat Collins",
	"Asesores de calidad",
	"Glenn Sperry",
	"Michael Lightner",
	"William Foster",
	"Jesse Clemit",
	"Jeff Fillhaber",
	"Manual, Diseno",
	"y Empaquetado",
	"Eydie Laramore",
	"Lisa Marcinko",
	"Lauren Rifkin",
	"Felicidades!",
	"Gracias por jugar The Hand of Fate!",
	"Programadores externos",
	"Producer Liaison",
	"Scott Duckett",
	"Irvine Testers",
	"Chris McFarland",
	"Paul Moore",
	"Chad Soares",
	"Jared Brinkley",
	"Jon Willliams",
	"Chris Toft",
	"Joe Kucan's Hair by",
	"Theodore A. Morris",
	"Cargar Partida",
	"Introducci""\xA2""n",
	"Empezar nueva partida",
	"Salir del juego",
	"Agradecimientos especiales",
	"Sake Joe Bostic-san",
	"Tim Fritz",
	"Kenny Dunne",
	"Gracias por jugar \"The Hand of Fate.\"\n"
};

static const StringListProvider k2SeqplayStringsDOSCDSpanishProvider = { ARRAYSIZE(k2SeqplayStringsDOSCDSpanish), k2SeqplayStringsDOSCDSpanish };

static const char *const k2SeqplayTlkFilesDOSCDSpanish[14] = {
	"EINTRO1",
	"EINTRO2",
	"EINTRO3",
	"EINTRO4",
	"EINTRO5",
	"EINTRO6",
	"EINTRO7",
	"EINTRO8",
	"EINTRO9",
	"EINTRO10",
	"EINTRO11",
	"EINTRO12",
	"EGLOW",
	""
};

static const StringListProvider k2SeqplayTlkFilesDOSCDSpanishProvider = { ARRAYSIZE(k2SeqplayTlkFilesDOSCDSpanish), k2SeqplayTlkFilesDOSCDSpanish };
