static const char *const kEoB2ChargenStrings1FMTownsJapanese[9] = {
	"\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x83\x81\x83\x93\x83""o""\x81""[""\x82\xAA\x91""S""\x88\xF5""\r""\x82\xBB\x82\xEB\x82\xA2\x82\xDC\x82\xB5\x82\xBD\x81""B""\x82""o""\x82""k""\x82""`""\x82""x""\x83""{""\x83""^""\x83\x93""\r""\x82\xA9\x81""A""\x83""L""\x81""[""\x83""{""\x81""[""\x83""h""\x82\xCC\x82""o""\x83""L""\x81""[""\x82\xF0""\r""\x89\x9F\x82\xB7\x82\xC6\x83""Q""\x81""[""\x83\x80\x82\xAA\x8E""n""\x82\xDC\x82\xE8\x82\xDC\x82\xB7\x81""B",
	"          ",
	"AC\rHP\rLVL",
	"%s\r%d\r%d\r%d\r%d\r%d",
	"%d\r%d",
	"%d",
	"%d/%d",
	"%d/%d/%d",
	"\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[""\x82\xF0\x8D\xEC\x82\xE9\x82\xBD\x82\xDF\x82\xCC""\r""\x83""{""\x83""b""\x83""N""\x83""X""\x82\xF0\x8E""w""\x92\xE8\x82\xB5\x82\xC4\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B"
};

static const StringListProvider kEoB2ChargenStrings1FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenStrings1FMTownsJapanese), kEoB2ChargenStrings1FMTownsJapanese };

static const char *const kEoB2ChargenStrings2FMTownsJapanese[12] = {
	"%s",
	"%d",
	"%s",
	"%d",
	"%d",
	"%d",
	"%s",
	"%d",
	"SELECT RACE:",
	"SELECT CLASS:",
	"SELECT ALIGNMENT:",
	"Name:"
};

static const StringListProvider kEoB2ChargenStrings2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenStrings2FMTownsJapanese), kEoB2ChargenStrings2FMTownsJapanese };

static const char *const kEoB2ChargenStatStringsFMTownsJapanese[12] = {
	"STR",
	"INT",
	"WIS",
	"DEX",
	"CON",
	"CHA",
	"STR",
	"INT",
	"WIS",
	"DEX",
	"CON",
	"CHA"
};

static const StringListProvider kEoB2ChargenStatStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenStatStringsFMTownsJapanese), kEoB2ChargenStatStringsFMTownsJapanese };

static const char *const kEoB2ChargenRaceSexStringsFMTownsJapanese[12] = {
	"HUMAN MALE",
	"HUMAN FEMALE",
	"ELF MALE",
	"ELF FEMALE",
	"HALF-ELF MALE",
	"HALF-ELF FEMALE",
	"DWARF MALE",
	"DWARF FEMALE",
	"GNOME MALE",
	"GNOME FEMALE",
	"HALFLING MALE",
	"HALFLING FEMALE"
};

static const StringListProvider kEoB2ChargenRaceSexStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenRaceSexStringsFMTownsJapanese), kEoB2ChargenRaceSexStringsFMTownsJapanese };

static const char *const kEoB2ChargenClassStringsFMTownsJapanese[21] = {
	"FIGHTER",
	"RANGER",
	"PALADIN",
	"MAGE",
	"CLERIC",
	"THIEF",
	"FIGHTER/CLERIC",
	"FIGHTER/THIEF",
	"FIGHTER/MAGE",
	"FIGHTER/MAGE/THIEF",
	"THIEF/MAGE",
	"CLERIC/THIEF",
	"FIGHTER/CLERIC/MAGE",
	"RANGER/CLERIC",
	"CLERIC/MAGE",
	"FIGHTER",
	"MAGE",
	"CLERIC",
	"THIEF",
	"PALADIN",
	"RANGER"
};

static const StringListProvider kEoB2ChargenClassStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenClassStringsFMTownsJapanese), kEoB2ChargenClassStringsFMTownsJapanese };

static const char *const kEoB2ChargenAlignmentStringsFMTownsJapanese[9] = {
	"LAWFUL GOOD",
	"NEUTRAL GOOD",
	"CHAOTIC GOOD",
	"LAWFUL NEUTRAL",
	"TRUE NEUTRAL",
	"CHAOTIC NEUTRAL",
	"LAWFUL EVIL",
	"NEUTRAL EVIL",
	"CHAOTIC EVIL"
};

static const StringListProvider kEoB2ChargenAlignmentStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenAlignmentStringsFMTownsJapanese), kEoB2ChargenAlignmentStringsFMTownsJapanese };

static const char *const kEoB2ChargenEnterGameStringsFMTownsJapanese[1] = {
	"  ""\x83""Q""\x81""[""\x83\x80\x82\xF0\x8E""n""\x82\xDF\x82\xDC\x82\xB7\x81""B\r  ""\x8F\xAD\x81""X""\x82\xA8\x91\xD2\x82\xBF\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B"
};

static const StringListProvider kEoB2ChargenEnterGameStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ChargenEnterGameStringsFMTownsJapanese), kEoB2ChargenEnterGameStringsFMTownsJapanese };

static const char *const kEoB2PryDoorStringsFMTownsJapanese[8] = {
	"\r",
	"\x97\xCD\x82\xB8\x82\xAD\x82\xC5\x83""h""\x83""A""\x82\xF0\x8A""J""\x82\xAF\x82\xE9\x82\xB1\x82\xC6\x82\xCD\x82\xC5\x82\xAB\x82\xBB\x82\xA4\x82\xC9\x82\xC8\x82\xA2\x81""B\r",
	"\x06\x04\x83""h""\x83""A""\x82\xF0\x82\xB1\x82\xB6\x8A""J""\x82\xAF\x82\xBD\x81""B\r",
	"\x06\x06\x83""h""\x83""A""\x82\xF0\x97\xCD\x82\xB8\x82\xAD\x82\xC5\x8A""J""\x82\xAF\x82\xE6\x82\xA4\x82\xC6\x82\xB5\x82\xBD\x82\xAA\x8E\xB8\x94""s""\x82\xB5\x82\xBD\x81""B\r",
	"\x83""A""\x83""C""\x83""e""\x83\x80\x82\xF0\x82\xBB\x82\xB1\x82\xC9\x92""u""\x82\xAD\x82\xB1\x82\xC6\x82\xCD\x82\xC5\x82\xAB\x82\xC8\x82\xA2\x81""B\r",
	"\x82\xBB\x82\xCC\x83""A""\x83""C""\x83""e""\x83\x80\x82\xCD\x91\xE5\x82\xAB\x82\xB7\x82\xAC\x82\xC4\x93\xFC\x82\xE7\x82\xC8\x82\xA2\x81""B\r",
	"\x92""N""\x82\xE0\x83""h""\x83""A""\x82\xF0\x82\xB1\x82\xB6\x8A""J""\x82\xAF\x82\xE9\x82\xB1\x82\xC6\x82\xAA\x82\xC5\x82\xAB\x82\xC8\x82\xA2\x81""B\r",
	"\r"
};

static const StringListProvider kEoB2PryDoorStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2PryDoorStringsFMTownsJapanese), kEoB2PryDoorStringsFMTownsJapanese };

static const char *const kEoB2WarningStringsFMTownsJapanese[4] = {
	"\x82\xBB\x82\xBF\x82\xE7\x82\xD6\x82\xCD\x8D""s""\x82\xAF\x82\xC8\x82\xA2\x81""B\r",
	"%s""\x82\xCD\x95\xA8\x82\xF0\x90""H""\x82\xD7\x82\xE7\x82\xEA\x82\xE9\x8F\xF3\x91\xD4\x82\xC5\x82\xCD\x82\xC8\x82\xA2\x81""I\r",
	"\x82\xBB\x82\xCC\x90""H""\x97\xBF\x82\xCD\x95\x85\x82\xC1\x82\xC4\x82\xA2\x82\xE9\x81""I""\x90""H""\x82\xD7\x82\xE9\x8B""C""\x82\xAA\x82\xB5\x82\xC8\x82\xA2\x81""B\r",
	"\x90""H""\x82\xD7\x95\xA8\x88\xC8\x8A""O""\x82\xCD\x90""H""\x82\xD7\x82\xE7\x82\xEA\x82\xC8\x82\xA2\x81""I\r"
};

static const StringListProvider kEoB2WarningStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2WarningStringsFMTownsJapanese), kEoB2WarningStringsFMTownsJapanese };

static const char *const kEoB2ItemSuffixStringsRingsFMTownsJapanese[4] = {
	"\x83""A""\x83""h""\x81""[""\x83\x93\x83\x81\x83\x93\x83""g",
	"\x83""E""\x83""C""\x83""U""\x81""[""\x83""h""\x83\x8A\x81""[",
	"\x83""T""\x83""X""\x83""e""\x83""B""\x83""i""\x83\x93\x83""X",
	"\x83""t""\x83""F""\x83""U""\x81""[""\x83""t""\x83""H""\x81""[""\x83\x8B"
};

static const StringListProvider kEoB2ItemSuffixStringsRingsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ItemSuffixStringsRingsFMTownsJapanese), kEoB2ItemSuffixStringsRingsFMTownsJapanese };

static const char *const kEoB2ItemSuffixStringsPotionsFMTownsJapanese[8] = {
	"\x83""W""\x83\x83\x83""C""\x83""A""\x83\x93\x83""g""\x81""E""\x83""X""\x83""g""\x83\x8C\x83\x93\x83""O""\x83""X",
	"\x83""q""\x81""[""\x83\x8A\x83\x93\x83""O",
	"\x83""G""\x83""L""\x83""X""\x83""g""\x83\x89\x81""E""\x83""q""\x81""[""\x83\x8A\x83\x93\x83""O",
	"\x83""|""\x83""C""\x83""Y""\x83\x93",
	"\x83""o""\x83""C""\x83""^""\x83\x8A\x83""e""\x83""B",
	"\x83""X""\x83""s""\x81""[""\x83""h",
	"\x83""C""\x83\x93\x83""r""\x83""W""\x83""r""\x83\x8A\x83""e""\x83""B",
	"\x83""L""\x83\x85\x83""A""\x81""E""\x83""|""\x83""C""\x83""Y""\x83\x93"
};

static const StringListProvider kEoB2ItemSuffixStringsPotionsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ItemSuffixStringsPotionsFMTownsJapanese), kEoB2ItemSuffixStringsPotionsFMTownsJapanese };

static const char *const kEoB2ItemSuffixStringsWandsFMTownsJapanese[8] = {
	"\x82""r""\x82\x94\x82\x89\x82\x83\x82\x8B",
	"\x82""k""\x82\x89\x82\x87\x82\x88\x82\x94\x82\x8E\x82\x89\x82\x8E\x82\x87",
	"\x82""e""\x82\x92\x82\x8F\x82\x93\x82\x94",
	"\x82""b""\x82\x95\x82\x92\x82\x89\x82\x8E\x82\x87",
	"\x82""e""\x82\x89\x82\x92\x82\x85\x82\x82\x82\x81\x82\x8C\x82\x8C",
	"\x82""r""\x82\x94\x82\x81\x82\x92\x82\x86\x82\x89\x82\x92\x82\x85",
	"\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83\x81""@""\x82""l""\x82\x89\x82\x93\x82\x93\x82\x89\x82\x8C\x82\x85",
	"\x82""c""\x82\x89\x82\x93\x82\x90\x82\x85\x82\x8C\x81""@""\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83"
};

static const StringListProvider kEoB2ItemSuffixStringsWandsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ItemSuffixStringsWandsFMTownsJapanese), kEoB2ItemSuffixStringsWandsFMTownsJapanese };

static const char *const kEoB2RipItemStringsFMTownsJapanese[3] = {
	"%s""\x82\xCC",
	"%s""\x82\xCC",
	"\x82\xCD\x96\xB3\x82\xAD\x82\xC8\x82\xC1\x82\xBD\x81""B\r"
};

static const StringListProvider kEoB2RipItemStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2RipItemStringsFMTownsJapanese), kEoB2RipItemStringsFMTownsJapanese };

static const char *const kEoB2CursedStringFMTownsJapanese[1] = {
	"%d""\x82\xCC\x8E\xF4\x82\xED\x82\xEA\x82\xBD""%s"
};

static const StringListProvider kEoB2CursedStringFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CursedStringFMTownsJapanese), kEoB2CursedStringFMTownsJapanese };

static const char *const kEoB2MagicObjectStringsFMTownsJapanese[5] = {
	"\x83\x81\x83""C""\x83""W""\x83""X""\x83""N""\x83\x8D\x81""[""\x83\x8B",
	"\x83""N""\x83\x8C\x83\x8A\x83""b""\x83""N""\x83""X""\x83""N""\x83\x8D\x81""[""\x83\x8B",
	"\x83\x8A\x83\x93\x83""O",
	"\x83""|""\x81""[""\x83""V""\x83\x87\x83\x93",
	"\x83\x8F\x83\x93\x83""h"
};

static const StringListProvider kEoB2MagicObjectStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicObjectStringsFMTownsJapanese), kEoB2MagicObjectStringsFMTownsJapanese };

static const char *const kEoB2MagicObjectString5FMTownsJapanese[1] = {
	"\x82""r""\x82\x94\x82\x89\x82\x83\x82\x8B"
};

static const StringListProvider kEoB2MagicObjectString5FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicObjectString5FMTownsJapanese), kEoB2MagicObjectString5FMTownsJapanese };

static const char *const kEoB2PatternSuffixFMTownsJapanese[1] = {
	"%s""\x82\xCC""%s"
};

static const StringListProvider kEoB2PatternSuffixFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2PatternSuffixFMTownsJapanese), kEoB2PatternSuffixFMTownsJapanese };

static const char *const kEoB2PatternGrFix1FMTownsJapanese[1] = {
	"%s""\x82\xCC""%s"
};

static const StringListProvider kEoB2PatternGrFix1FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2PatternGrFix1FMTownsJapanese), kEoB2PatternGrFix1FMTownsJapanese };

static const char *const kEoB2PatternGrFix2FMTownsJapanese[1] = {
	"%s""\x82\xCC""%s"
};

static const StringListProvider kEoB2PatternGrFix2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2PatternGrFix2FMTownsJapanese), kEoB2PatternGrFix2FMTownsJapanese };

static const char *const kEoB2ValidateArmorStringFMTownsJapanese[1] = {
	"%s""\x82\xCD\x82\xBB\x82\xCC\x83""^""\x83""C""\x83""v""\x82\xCC\x96""h""\x8B\xEF\x82\xF0\x90""g""\x82\xC9\x82\xC2\x82\xAF\x82\xE7\x82\xEA\x82\xC8\x82\xA2\x81""B\r",
};

static const StringListProvider kEoB2ValidateArmorStringFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ValidateArmorStringFMTownsJapanese), kEoB2ValidateArmorStringFMTownsJapanese };

static const char *const kEoB2ValidateCursedStringFMTownsJapanese[1] = {
	"%s""\x82\xCD\x95\x90\x8A\xED\x82\xF0\x8A""O""\x82\xB7\x82\xB1\x82\xC6\x82\xAA\x82\xC5\x82\xAB\x82\xC8\x82\xA2\x81""I""\x82\xBB\x82\xCC\x95\x90\x8A\xED\x82\xCD\x8E\xF4\x82\xED\x82\xEA\x82\xC4\x82\xA2\x82\xE9\x81""I\r",
};

static const StringListProvider kEoB2ValidateCursedStringFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ValidateCursedStringFMTownsJapanese), kEoB2ValidateCursedStringFMTownsJapanese };

static const char *const kEoB2ValidateNoDropStringFMTownsJapanese[1] = {
	"\x82\xB1\x82\xCC\x83""A""\x83""C""\x83""e""\x83\x80\x82\xCD\x82\xBB\x82\xB1\x82\xC9\x92""u""\x82\xAF\x82\xC8\x82\xA2\x81""B\r"
};

static const StringListProvider kEoB2ValidateNoDropStringFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ValidateNoDropStringFMTownsJapanese), kEoB2ValidateNoDropStringFMTownsJapanese };

static const char *const kEoB2PotionStringsFMTownsJapanese[2] = {
	"\x93\xC5\x82\xF0\x8E\xF3\x82\xAF\x82\xBD",
	"%s""\x82\xCD""%s""\x81""I\r"
};

static const StringListProvider kEoB2PotionStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2PotionStringsFMTownsJapanese), kEoB2PotionStringsFMTownsJapanese };

static const char *const kEoB2WandStringsFMTownsJapanese[1] = {
	"\x82\xBB\x82\xCC\x83\x8F\x83\x93\x83""h""\x82\xC9\x82\xCD\x96\x82\x97\xCD\x82\xAA\x8E""c""\x82\xC1\x82\xC4\x82\xA2\x82\xC8\x82\xA2\x81""B\r"
};

static const StringListProvider kEoB2WandStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2WandStringsFMTownsJapanese), kEoB2WandStringsFMTownsJapanese };

static const char *const kEoB2ItemMisuseStringsFMTownsJapanese[3] = {
	"%s""\x82\xCD\x82\xBB\x82\xCC\x83""A""\x83""C""\x83""e""\x83\x80\x82\xF0\x8E""g""\x82\xA6\x82\xC8\x82\xA2\x81""B\r",
	"\x82\xB1\x82\xCC\x83""A""\x83""C""\x83""e""\x83\x80\x82\xCD\x90""g""\x82\xC9\x82\xC2\x82\xAF\x82\xC4\x82\xA2\x82\xEA\x82\xCE\x8E\xA9\x93\xAE\x93""I""\x82\xC9\x8C\xF8\x89\xCA\x82\xF0\x94\xAD\x8A\xF6\x82\xB7\x82\xE9\x81""B\r",
	"\x83""A""\x83""C""\x83""e""\x83\x80\x82\xCC\x8E""g""\x82\xA2\x95\xFB\x82\xAA\x8A\xD4\x88\xE1\x82\xC1\x82\xC4\x82\xA2\x82\xE9\x81""B\r"
};

static const StringListProvider kEoB2ItemMisuseStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ItemMisuseStringsFMTownsJapanese), kEoB2ItemMisuseStringsFMTownsJapanese };

static const char *const kEoB2TakenStringsFMTownsJapanese[1] = {
	"\x82\xF0\x8E\xE6\x82\xC1\x82\xBD\x81""B\r"
};

static const StringListProvider kEoB2TakenStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2TakenStringsFMTownsJapanese), kEoB2TakenStringsFMTownsJapanese };

static const char *const kEoB2PotionEffectStringsFMTownsJapanese[8] = {
	"\x97\xCD\x82\xAA\x82\xDD\x82\xC8\x82\xAC\x82\xC1\x82\xC4\x82\xAB\x82\xBD",
	"\x8B""C""\x95\xAA\x82\xAA\x82\xE6\x82\xAD\x82\xC8\x82\xC1\x82\xBD",
	"\x8B""C""\x95\xAA\x82\xAA\x82\xA9\x82\xC8\x82\xE8\x82\xE6\x82\xAD\x82\xC8\x82\xC1\x82\xBD",
	"\x8B""C""\x95\xAA\x82\xAA\x88\xAB\x82\xAD\x82\xC8\x82\xC1\x82\xBD",
	"\x8B\xF3\x95\xA0\x82\xC5\x82\xC8\x82\xAD\x82\xC8\x82\xC1\x82\xBD",
	"\x91""f""\x91\x81\x82\xB3\x82\xAA\x91\x9D\x82\xB5\x82\xBD",
	"\x91\xCC\x82\xAA\x93\xA7\x96\xBE\x82\xC9\x82\xC8\x82\xC1\x82\xBD",
	"\x8B""C""\x95\xAA\x82\xAA\x82\xE6\x82\xAD\x82\xC8\x82\xC1\x82\xBD"
};

static const StringListProvider kEoB2PotionEffectStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2PotionEffectStringsFMTownsJapanese), kEoB2PotionEffectStringsFMTownsJapanese };

static const char *const kEoB2YesNoStringsFMTownsJapanese[2] = {
	"\x82\xCD\x82\xA2",
	"\x82\xA2\x82\xA2\x82\xA6"
};

static const StringListProvider kEoB2YesNoStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2YesNoStringsFMTownsJapanese), kEoB2YesNoStringsFMTownsJapanese };

static const char *const kEoB2MoreStringsFMTownsJapanese[1] = {
	"\x82""l""\x82""n""\x82""q""\x82""d"
};

static const StringListProvider kEoB2MoreStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MoreStringsFMTownsJapanese), kEoB2MoreStringsFMTownsJapanese };

static const char *const kEoB2NpcMaxStringsFMTownsJapanese[1] = {
	"\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x95\xD2\x90\xAC\x82\xCD\x82""U""\x90""l""\x82\xDC\x82\xC5\x82\xC5\x82\xB7\x81""B""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xA9\x82\xE7\x8A""O""\x82\xB7\x90""l""\x82\xF0\x82""P""\x90""l""\x91""I""\x82\xF1\x82\xC5\x89\xBA\x82\xB3\x82\xA2\x81""B"
};

static const StringListProvider kEoB2NpcMaxStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2NpcMaxStringsFMTownsJapanese), kEoB2NpcMaxStringsFMTownsJapanese };

static const char *const kEoB2OkStringsFMTownsJapanese[1] = {
	"\x82""n""\x82""j"
};

static const StringListProvider kEoB2OkStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2OkStringsFMTownsJapanese), kEoB2OkStringsFMTownsJapanese };

static const char *const kEoB2NpcJoinStringsFMTownsJapanese[1] = {
	"%s""\x82\xAA\x83""p""\x81""[""\x83""e""\x83""B""\x82\xC9\x89\xC1\x82\xED\x82\xC1\x82\xBD\x81""B\r"
};

static const StringListProvider kEoB2NpcJoinStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2NpcJoinStringsFMTownsJapanese), kEoB2NpcJoinStringsFMTownsJapanese };

static const char *const kEoB2CancelStringsFMTownsJapanese[1] = {
	"\x92\x86\x8E""~"
};

static const StringListProvider kEoB2CancelStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CancelStringsFMTownsJapanese), kEoB2CancelStringsFMTownsJapanese };

static const char *const kEoB2AbortStringsFMTownsJapanese[1] = {
	"\x92\x86\x8E""~"
};

static const StringListProvider kEoB2AbortStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2AbortStringsFMTownsJapanese), kEoB2AbortStringsFMTownsJapanese };

static const char *const kEoB2MenuStringsMainFMTownsJapanese[8] = {
	"\x83""I""\x83""v""\x83""V""\x83\x87\x83\x93\x91""I""\x91\xF0"":",
	"\x8B""x""\x91\xA7\x82\xB7\x82\xE9",
	"\x8E\xF4\x95\xB6\x82\xF0\x8B""L""\x89\xAF\x82\xB7\x82\xE9",
	"\x8B""F""\x82\xE8\x82\xF0\x95\xF9\x82\xB0\x82\xE9",
	"\x83""X""\x83""N""\x83\x8D\x81""[""\x83\x8B\x82\xF0\x8F\x91\x82\xAB\x8E\xCA\x82\xB7",
	"\x8A\xC2\x8B\xAB\x90\xDD\x92\xE8",
	"\x83""I""\x83""v""\x83""V""\x83\x87\x83\x93",
	"EXIT",
};

static const StringListProvider kEoB2MenuStringsMainFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsMainFMTownsJapanese), kEoB2MenuStringsMainFMTownsJapanese };

static const char *const kEoB2MenuStringsSaveLoadFMTownsJapanese[8] = {
	"\x83""Q""\x81""[""\x83\x80\x83\x8D\x81""[""\x83""h",
	"\x83""Q""\x81""[""\x83\x80\x83""Z""\x81""[""\x83""u",
	"\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[""\x82\xF0\x8A""O""\x82\xB7",
	"\x83""Q""\x81""[""\x83\x80\x82\xF0\x8F""I""\x97\xB9\x82\xB7\x82\xE9",
	"\x83""I""\x83""v""\x83""V""\x83\x87\x83\x93"":",
	"\r  ""\x83""Z""\x81""[""\x83""u""\x8F""I""\x97\xB9",
	"\r""\x81""@""\x83""f""\x81""[""\x83""^""\x82\xCC\x83""Z""\x81""[""\x83""u""\x82\xC9\x8E\xB8\x94""s""\x82\xB5\x82\xDC\x82\xB5\x82\xBD\x81""I",
	"\r""\x81""@""\x83""Z""\x81""[""\x83""u""\x83""f""\x81""[""\x83""^""\x82\xAA\x93\xC7\x82\xDD\x8D\x9E\x82\xDF\x82\xDC\x82\xB9\x82\xF1\x81""B\r""\x81""@""\x83""t""\x83""@""\x83""C""\x83\x8B\x82\xAA\x89\xF3\x82\xEA\x82\xC4\x82\xA2\x82\xDC\x82\xB7\x81""I\r",
};

static const StringListProvider kEoB2MenuStringsSaveLoadFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsSaveLoadFMTownsJapanese), kEoB2MenuStringsSaveLoadFMTownsJapanese };

static const char *const kEoB2MenuStringsOnOffFMTownsJapanese[2] = {
	"\x82""n""\x82""m",
	"\x82""n""\x82""e""\x82""e"
};

static const StringListProvider kEoB2MenuStringsOnOffFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsOnOffFMTownsJapanese), kEoB2MenuStringsOnOffFMTownsJapanese };

static const char *const kEoB2MenuStringsSpellsFMTownsJapanese[17] = {
	"\r\r  ""\x8E\xF4\x95\xB6\x82\xF0\x8B""L""\x89\xAF\x82\xB3\x82\xB9\x82\xBD\x82\xA2\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[""\x82\xF0""\r  ""\x91""I""\x82\xF1\x82\xC5\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B",
	"\r  ""\x82\xBB\x82\xCC\x83""p""\x83\x89\x83""f""\x83""B""\x83\x93\x82\xCD\x83\x8C\x83""x""\x83\x8B\x82\xAA\x92\xE1\x82\xB7\x82\xAC""\r  ""\x82\xC4\x8E\xF4\x95\xB6\x82\xF0\x8A""o""\x82\xA6\x82\xE9\x82\xB1\x82\xC6\x82\xAA\x82\xC5\x82\xAB\x82\xC8\x82\xA2\x81""B",
	"\r\r  ""\x82\xBB\x82\xCC\x83\x81\x83""C""\x83""W""\x82\xCD\x83""X""\x83""y""\x83\x8B\x83""u""\x83""b""\x83""N""\x82\xF0\x8E\x9D\x82\xC1\x82\xC4\x82\xA2""\r  ""\x82\xC8\x82\xA2\x81""I",
	"\r\r\r  ""\x8E\xF4\x95\xB6\x82\xCC\x82\xBD\x82\xDF\x82\xC9\x8B""F""\x82\xE8\x82\xF0\x95\xF9\x82\xB0\x82\xE9\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[\r  ""\x82\xF0\x91""I""\x82\xF1\x82\xC5\x89\xBA\x82\xB3\x82\xA2\x81""B",
	"\r  ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x92\x86\x82\xC9\x8B""F""\x82\xE8\x82\xF0\x95\xF9\x82\xB0\x82\xE7\x82\xEA\x82\xE9""\r  ""\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[""\x82\xAA\x82\xA2\x82\xC8\x82\xA2\x81""B",
	"\r  ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x92\x86\x82\xC9\x82\xCD\x90""V""\x82\xB5\x82\xA2\x8E\xF4\x95\xB6\x82\xF0""\r  ""\x8A""o""\x82\xA6\x82\xE9\x82\xB1\x82\xC6\x82\xAA\x82\xC5\x82\xAB\x82\xE9\x8E\xD2\x82\xAA\x82\xA2\x82\xC8\x82\xA2\x81""B",
	"  ""\x8B""C""\x90\xE2\x82\xDC\x82\xBD\x82\xCD\x8E\x80\x82\xF1\x82\xBE\x83\x81\x83""C""\x83""W""\x82\xCD\x8E\xF4\x95\xB6\x82\xF0\x8B""L""\x89\xAF""\r  ""\x82\xB7\x82\xE9\x82\xB1\x82\xC6\x82\xAA\x82\xC5\x82\xAB\x82\xC8\x82\xA2\x81""B",
	"  ""\x8B""C""\x90\xE2\x82\xDC\x82\xBD\x82\xCD\x8E\x80\x82\xF1\x82\xBE\x83""N""\x83\x8C\x83\x8A\x83""b""\x83""N""\x82\xCD\x8B""F""\x82\xE8\x82\xF0""\r  ""\x95\xF9\x82\xB0\x82\xE9\x82\xB1\x82\xC6\x82\xCD\x82\xC5\x82\xAB\x82\xC8\x82\xA2\x81""B",
	"1",
	"2",
	"3",
	"4",
	"5",
	"\x83""N""\x83\x8A\x83""A",
	"\x8E""g""\x97""p""\x89\xC2\x94""\\""\x82\xC8\x8E\xF4\x95\xB6"":",
	"\x82\xCD\x82\xA2",
	"\x82\xA2\x82\xA2\x82\xA6"
};

static const StringListProvider kEoB2MenuStringsSpellsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsSpellsFMTownsJapanese), kEoB2MenuStringsSpellsFMTownsJapanese };

static const char *const kEoB2MenuStringsRestFMTownsJapanese[5] = {
	"\r  ""\x83""N""\x83\x8C\x83\x8A\x83""b""\x83""N""\x82\xC9\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x8E\xA1\x97\xC3\x82\xF0""\r  ""\x82\xB3\x82\xB9\x82\xDC\x82\xB7\x82\xA9\x81""H",
	" ""\x82\xDC\x82\xBE\x89\xF6\x89\xE4\x82\xF0\x82\xB5\x82\xC4\x82\xA2\x82\xE9\x90""l""\x82\xAA\x82\xA2\x82\xDC\x82\xB7\x81""B\r  ""\x8B""x""\x91\xA7\x82\xF0\x91\xB1\x82\xAF\x82\xDC\x82\xB7\x82\xA9\x81""H",
	"  ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCD\x8B""x""\x8C""e""\x82\xB5\x82\xC4\x82\xA2\x82\xDC\x82\xB7""..",
	"\r""\x81""@""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCD\x91""S""\x88\xF5\x8A\xAE\x91""S""\x82\xC9\x89\xF1\x95\x9C\x82\xB5\x82\xDC\x82\xB5\x82\xBD\x81""B",
	"  ""\x8E\xF4\x95\xB6\x82\xF0\x8E""g""\x82\xA6\x82\xE9\x97""l""\x82\xC9\x82\xC8\x82\xE9\x82\xBD\x82\xDF\x82\xC9\x82\xCD\x81""A\r  ""\x8B""x""\x91\xA7\x82\xAA\x95""K""\x97""v""\x82\xC5\x82\xB7\x81""B\r"
};

static const StringListProvider kEoB2MenuStringsRestFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsRestFMTownsJapanese), kEoB2MenuStringsRestFMTownsJapanese };

static const char *const kEoB2MenuStringsDropFMTownsJapanese[1] = {
	" ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x83\x81\x83\x93\x83""o""\x81""[""\x82\xF0\x82""S""\x90""l""\x96\xA2\x96\x9E""\r""\x82\xC9\x82\xB7\x82\xE9\x82\xB1\x82\xC6\x82\xCD\x82\xC5\x82\xAB\x82\xDC\x82\xB9\x82\xF1\x81""B"
};

static const StringListProvider kEoB2MenuStringsDropFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsDropFMTownsJapanese), kEoB2MenuStringsDropFMTownsJapanese };

static const char *const kEoB2MenuStringsExitFMTownsJapanese[1] = {
	"\r""\x83""Q""\x81""[""\x83\x80\x82\xF0\x8F""I""\x97\xB9\x82\xB5\x82\xC4\x82\xE0\x82\xA2\x82\xA2\x82\xC5\x82\xB7\x82\xA9\x81""H"
};

static const StringListProvider kEoB2MenuStringsExitFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsExitFMTownsJapanese), kEoB2MenuStringsExitFMTownsJapanese };

static const char *const kEoB2MenuStringsStarveFMTownsJapanese[1] = {
	"\r ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x92\x86\x82\xC9\x8B\xF3\x95\xA0\x82\xCC\x83\x81\x83\x93\x83""o""\x81""[""\x82\xAA""\r ""\x82\xA2\x82\xDC\x82\xB7\x81""B\r ""\x82\xBB\x82\xEA\x82\xC5\x82\xE0\x8B""x""\x91\xA7\x82\xF0\x91\xB1\x82\xAF\x82\xDC\x82\xB7\x82\xA9\x81""H"
};

static const StringListProvider kEoB2MenuStringsStarveFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsStarveFMTownsJapanese), kEoB2MenuStringsStarveFMTownsJapanese };

static const char *const kEoB2MenuStringsScribeFMTownsJapanese[5] = {
	"  ""\x8F\x91\x82\xAB\x8E\xCA\x82\xB5\x82\xBD\x82\xA2\x8E\xF4\x95\xB6\x82\xCC\x83""X""\x83""N""\x83\x8D\x81""[""\x83\x8B\x82\xF0""\r  ""\x91""I""\x82\xF1\x82\xC5\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B",
	"\r\r\r  ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCC\x92\x86\x82\xA9\x82\xE7\x8E\xF4\x95\xB6\x82\xF0\x8F\x91\x82\xAB\x8E\xCA\x82\xB5\x82\xBD\x82\xA2""\r  ""\x83\x81\x83""C""\x83""W""\x82\xF0\x91""I""\x82\xF1\x82\xC5\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B",
	"  ""\x8E\xF4\x95\xB6\x82\xF0\x8F\xA5\x82\xA6\x82\xE9\x82\xBD\x82\xDF\x82\xCC\x83""X""\x83""N""\x83\x8D\x81""[""\x83\x8B""\r ""\x82\xF0\x8E\x9D\x82\xC1\x82\xC4\x82\xA2\x82\xC8\x82\xA2\x81""B",
	"  ""\x82\xB1\x82\xCC\x83\x81\x83""C""\x83""W""\x82\xAA\x95""K""\x97""v""\x82\xC6\x82\xB7\x82\xE9\x83""X""\x83""N""\x83\x8D\x81""[""\x83\x8B""\r ""\x82\xF0\x8E\x9D\x82\xC1\x82\xC4\x82\xA2\x82\xC8\x82\xA2\x81""B",
	"\r  ""\x90""V""\x82\xB5\x82\xA2\x8E\xF4\x95\xB6\x82\xF0\x8F\x91\x82\xAB\x8E\xCA\x82\xB7\x82\xB1\x82\xC6\x82\xCC\x82\xC5\x82\xAB\x82\xE9""\r  ""\x83\x81\x83""C""\x83""W""\x82\xAA\x82\xA2\x82\xC8\x82\xA2\x81""B"
};

static const StringListProvider kEoB2MenuStringsScribeFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsScribeFMTownsJapanese), kEoB2MenuStringsScribeFMTownsJapanese };

static const char *const kEoB2MenuStringsDrop2FMTownsJapanese[3] = {
	"  ""\x95\xCA\x82\xEA\x82\xE9\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[""\x82\xF0\x91""I""\x82\xF1\x82\xC5\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B\r",
	"  ""\x83""Z""\x81""[""\x83""u""\x83""t""\x83""@""\x83""C""\x83\x8B\x82\xC9\x82\xCD\x96\xBC\x91""O""\x82\xAA\x95""K""\x97""v""\x82\xC5\x82\xB7\x81""B",
	"\r  ""\x91""O""\x82\xCC\x83""Z""\x81""[""\x83""u""\x83""f""\x81""[""\x83""^""\x82\xC9\x8F\xE3\x8F\x91\x82\xAB\x82\xB5\x82\xC4\x82\xE0""\r  ""\x82\xA2\x82\xA2\x82\xC5\x82\xB7\x82\xA9\x81""H"
};

static const StringListProvider kEoB2MenuStringsDrop2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsDrop2FMTownsJapanese), kEoB2MenuStringsDrop2FMTownsJapanese };

static const char *const kEoB2MenuStringsHeadFMTownsJapanese[3] = {
	"\x83""L""\x83\x83\x83\x93\x83""v:",
	"\x8A\xC2\x8B\xAB\x90\xDD\x92\xE8"":",
	"\x83""I""\x83""v""\x83""V""\x83\x87\x83\x93"":"
};

static const StringListProvider kEoB2MenuStringsHeadFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsHeadFMTownsJapanese), kEoB2MenuStringsHeadFMTownsJapanese };

static const char *const kEoB2MenuStringsPoisonFMTownsJapanese[1] = {
	"\x81""@""\x93\xC5\x82\xF0\x8E\xF3\x82\xAF\x82\xBD\x83""L""\x83\x83\x83\x89\x83""N""\x83""^""\x81""[""\x82\xAA\x8E\x80\x82\xF1\x82\xC5""\r""\x82\xB5\x82\xDC\x82\xA2\x82\xDC\x82\xB7\x81""I\r""\x81""@""\x82\xBB\x82\xEA\x82\xC5\x82\xE0\x8B""x""\x82\xDD\x82\xDC\x82\xB7\x82\xA9\x81""H"
};

static const StringListProvider kEoB2MenuStringsPoisonFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsPoisonFMTownsJapanese), kEoB2MenuStringsPoisonFMTownsJapanese };

static const char *const kEoB2MenuStringsMgcFMTownsJapanese[2] = {
	"",
	"%d""\x8C\xC2\x92\x86\x81""@%d""\x8C\xC2\x8E""g""\x97""p""\x89\xC2\x94""\\"
};

static const StringListProvider kEoB2MenuStringsMgcFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsMgcFMTownsJapanese), kEoB2MenuStringsMgcFMTownsJapanese };

static const char *const kEoB2MenuStringsPrefsFMTownsJapanese[4] = {
	"\x82""a""\x82""f""\x82""l %-6s",
	"\x8C\xF8\x89\xCA\x89\xB9"" %-6s",
	"\x83""o""\x81""[""\x83""O""\x83\x89\x83""t""\x95""\\""\x8E\xA6"" %-6s",
	"\x89\xE6\x96\xCA\x83\x82\x81""[""\x83""h""\x81""@%-10s"
};

static const StringListProvider kEoB2MenuStringsPrefsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsPrefsFMTownsJapanese), kEoB2MenuStringsPrefsFMTownsJapanese };

static const char *const kEoB2MenuStringsRest2FMTownsJapanese[5] = {
	"%s""\x82\xCD""%s""\x82\xCC\x8E\xF4\x95\xB6\x82\xF0\x8E\xF6\x82\xA9\x82\xC1\x82\xBD\x81""B\r",
	"%s""\x82\xCD""%s""\x82\xCC\x8E\xF4\x95\xB6\x82\xF0\x8B""L""\x89\xAF\x82\xB5\x82\xBD\x81""B\r",
	"%s""\x82\xCD""%s""\x82\xC9\x8E\xA1\x96\xFC\x82\xCC\x8E\xF4\x95\xB6\x82\xF0\x82\xC6\x82\xC8\x82\xA6\x82\xBD\x81""B\r",
	"\x8B""x""\x91\xA7\x82\xB5\x82\xBD\x8E\x9E\x8A\xD4\x81""F %-4d",
	"\r%s\r"
};

static const StringListProvider kEoB2MenuStringsRest2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsRest2FMTownsJapanese), kEoB2MenuStringsRest2FMTownsJapanese };

static const char *const kEoB2MenuStringsRest3FMTownsJapanese[1] = {
	"\x06\x06\x82\xB1\x82\xB1\x82\xC5\x82\xCD\x88\xC0\x90""S""\x82\xB5\x82\xC4\x8B""x""\x82\xDF\x82\xBB\x82\xA4\x82\xC9\x82\xC8\x82\xA2\x81""B"
};

static const StringListProvider kEoB2MenuStringsRest3FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsRest3FMTownsJapanese), kEoB2MenuStringsRest3FMTownsJapanese };

static const char *const kEoB2MenuStringsRest4FMTownsJapanese[1] = {
	"\x06\x06\x83\x82\x83\x93\x83""X""\x83""^""\x81""[""\x82\xAA\x8B\xDF\x82\xAD\x82\xC9\x82\xA2\x82\xC4\x8B""x""\x82\xDF\x82\xC8\x82\xA2\x81""B"
};

static const StringListProvider kEoB2MenuStringsRest4FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsRest4FMTownsJapanese), kEoB2MenuStringsRest4FMTownsJapanese };

static const char *const kEoB2MenuStringsDefeatFMTownsJapanese[1] = {
	"\r ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCD\x91""S""\x96\xC5\x82\xB5\x82\xBD\x81""I\r ""\x83""Z""\x81""[""\x83""u""\x83""Q""\x81""[""\x83\x80\x82\xF0\x93\xC7\x82\xDD\x8D\x9E\x82\xF1\x82\xC5\x91\xB1\x82\xAB\x82\xF0""\r ""\x82\xB5\x82\xDC\x82\xB7\x82\xA9\x81""H"
};

static const StringListProvider kEoB2MenuStringsDefeatFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsDefeatFMTownsJapanese), kEoB2MenuStringsDefeatFMTownsJapanese };

static const char *const kEoB2MenuStringsTransferFMTownsJapanese[5] = {
	"You must transfer\rfour party members\rbefore you may play!",
	"You may only transfer\rfour party members to\rEOB II!",
	"Items which are not\ruseful or unbalance\rthe game will be\rdeleted.",
	" \r ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xc9\x83\x81\x83""C""\x83""W""\x82\xaa\x82\xa2\x82\xdc\x82\xb9\x82\xf1\x81""B",
	" \r  ""\x83""p""\x81""[""\x83""e""\x83""B""\x82\xc9\x83""N""\x83\x8c\x83\x8a\x83""b""\x83""N""\x82\xe0\x83""p""\x83\x89\x83""f""\x83""B""\x83\x93""\r  ""\x82\xe0\x82\xa2\x82\xdc\x82\xb9\x82\xf1\x81""B"
};

static const StringListProvider kEoB2MenuStringsTransferFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsTransferFMTownsJapanese), kEoB2MenuStringsTransferFMTownsJapanese };


static const char *const kEoB2MenuStringsSpecFMTownsJapanese[2] = {
	"\r ""\x96\xDA\x82\xAA\x8A""o""\x82\xDF\x82\xE9\x82\xC6\x81""A""\x83""V""\x81""[""\x83""t""\x82\xCC\x83""C""\x83\x93\x83""T""\x83\x8B\x82\xAA""\r ""\x82\xA2\x82\xC8\x82\xAD\x82\xC8\x82\xC1\x82\xC4\x82\xA2\x82\xBD\x81""I",
	"  ""\x8E\x9D\x82\xBF\x95\xA8\x82\xAA\x82\xC8\x82\xAD\x82\xC8\x82\xC1\x82\xC4\x82\xA2\x82\xE9\x81""I"
};

static const StringListProvider kEoB2MenuStringsSpecFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuStringsSpecFMTownsJapanese), kEoB2MenuStringsSpecFMTownsJapanese };

static const char *const kEoB2MenuYesNoStringsFMTownsJapanese[2] = {
	"\x82\xCD\x82\xA2",
	"\x82\xA2\x82\xA2\x82\xA6"
};

static const StringListProvider kEoB2MenuYesNoStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MenuYesNoStringsFMTownsJapanese), kEoB2MenuYesNoStringsFMTownsJapanese };

static const char *const kEoB2CharGuiStringsHpFMTownsJapanese[2] = {
	"HP",
	"%3d of %-3d"
};

static const StringListProvider kEoB2CharGuiStringsHpFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharGuiStringsHpFMTownsJapanese), kEoB2CharGuiStringsHpFMTownsJapanese };

static const char *const kEoB2CharGuiStringsWp2FMTownsJapanese[3] = {
	"MISS",
	"HACK",
	"BASH"
};

static const StringListProvider kEoB2CharGuiStringsWp2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharGuiStringsWp2FMTownsJapanese), kEoB2CharGuiStringsWp2FMTownsJapanese };

static const char *const kEoB2CharGuiStringsWrFMTownsJapanese[4] = {
	"CAN'T",
	"REACH",
	"NO",
	"AMMO"
};

static const StringListProvider kEoB2CharGuiStringsWrFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharGuiStringsWrFMTownsJapanese), kEoB2CharGuiStringsWrFMTownsJapanese };

static const char *const kEoB2CharGuiStringsSt2FMTownsJapanese[7] = {
	"Swapping",
	"\x8E\x80\x96""S",
	"\x8B""C""\x90\xE2",
	"\x93\xC5\x81""i""\x92""x""\x82\xA2\x81""j",
	"\x93\xC5",
	"\x96\x83\xE1\x83",
	"\x90\xCE\x89\xBB"
};

static const StringListProvider kEoB2CharGuiStringsSt2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharGuiStringsSt2FMTownsJapanese), kEoB2CharGuiStringsSt2FMTownsJapanese };

static const char *const kEoB2CharGuiStringsInFMTownsJapanese[4] = {
	"CHARACTER INFO",
	"AC",
	"EXP",
	"LVL"
};

static const StringListProvider kEoB2CharGuiStringsInFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharGuiStringsInFMTownsJapanese), kEoB2CharGuiStringsInFMTownsJapanese };

static const char *const kEoB2CharStatusStrings7FMTownsJapanese[1] = {
	"%s""\x82\xCC\x83""W""\x83\x83\x83""C""\x83""A""\x83\x93\x83""g""\x81""E""\x83""X""\x83""g""\x83\x8C\x83\x93\x83""O""\x83""X""\x82\xCC\x8C\xF8\x89\xCA\x82\xAA\x90\xD8\x82\xEA\x82\xBD\x81""B\r"
};

static const StringListProvider kEoB2CharStatusStrings7FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharStatusStrings7FMTownsJapanese), kEoB2CharStatusStrings7FMTownsJapanese };

static const char *const kEoB2CharStatusStrings82FMTownsJapanese[1] = {
	"\x06\x06""%s""\x82\xCD\x93\xC5\x82\xAA\x89\xF1\x82\xC1\x82\xBD\x81""I\r"
};

static const StringListProvider kEoB2CharStatusStrings82FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharStatusStrings82FMTownsJapanese), kEoB2CharStatusStrings82FMTownsJapanese };

static const char *const kEoB2CharStatusStrings9FMTownsJapanese[1] = {
	"\x06\x04""%s""\x82\xCC\x96\x83\xE1\x83\x82\xCD\x8E\xA1\x82\xC1\x82\xBD\x81""I\r"
};

static const StringListProvider kEoB2CharStatusStrings9FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharStatusStrings9FMTownsJapanese), kEoB2CharStatusStrings9FMTownsJapanese };

static const char *const kEoB2CharStatusStrings12FMTownsJapanese[1] = {
	"%s""\x82\xCC\x93\xAE\x82\xAB\x82\xCD\x93\xDD\x82\xAD\x82\xC8\x82\xC1\x82\xBD\x81""B\r"
};

static const StringListProvider kEoB2CharStatusStrings12FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharStatusStrings12FMTownsJapanese), kEoB2CharStatusStrings12FMTownsJapanese };

static const char *const kEoB2CharStatusStrings132FMTownsJapanese[1] = {
	"\x06\x06""%s""\x82\xCD""%s""\x81""I\r"
};

static const StringListProvider kEoB2CharStatusStrings132FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2CharStatusStrings132FMTownsJapanese), kEoB2CharStatusStrings132FMTownsJapanese };

static const char *const kEoB2LevelGainStringsFMTownsJapanese[1] = {
	"\x06\x01""%s""\x82\xCD\x83\x8C\x83""x""\x83\x8B\x82\xAA\x8F\xE3\x82\xAA\x82\xC1\x82\xBD\x81""B""\x06\x0F""\r"
};

static const StringListProvider kEoB2LevelGainStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2LevelGainStringsFMTownsJapanese), kEoB2LevelGainStringsFMTownsJapanese };

static const char *const kEoB2BookNumbersFMTownsJapanese[5] = {
	"First",
	"Second",
	"Third",
	"Fourth",
	"Fifth"
};

static const StringListProvider kEoB2BookNumbersFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2BookNumbersFMTownsJapanese), kEoB2BookNumbersFMTownsJapanese };

static const char *const kEoB2MageSpellsListFMTownsJapanese[33] = {
	"",
	"\x82""`""\x82""q""\x82""l""\x82""n""\x82""q",
	"\x82""a""\x82""t""\x82""q""\x82""m""\x82""h""\x82""m""\x82""f""\x81""@""\x82""g""\x82""`""\x82""m""\x82""c""\x82""r",
	"\x82""c""\x82""d""\x82""s""\x82""d""\x82""b""\x82""s""\x81""@""\x82""l""\x82""`""\x82""f""\x82""h""\x82""b",
	"\x82""l""\x82""`""\x82""f""\x82""h""\x82""b""\x81""@""\x82""l""\x82""h""\x82""r""\x82""r""\x82""h""\x82""k""\x82""d",
	"\x82""r""\x82""g""\x82""h""\x82""d""\x82""k""\x82""c",
	"\x82""r""\x82""g""\x82""n""\x82""b""\x82""j""\x82""h""\x82""m""\x82""f""\x81""@""\x82""f""\x82""q""\x82""`""\x82""r""\x82""o",
	"\x82""a""\x82""k""\x82""t""\x82""q",
	"\x82""c""\x82""d""\x82""s""\x82""d""\x82""b""\x82""s""\x81""@""\x82""h""\x82""m""\x82""u""\x82""h""\x82""r""\x82""h""\x82""a""\x82""k""\x82""d",
	"\x82""h""\x82""l""\x82""o""\x81""@""\x82""h""\x82""c""\x82""d""\x82""m""\x82""s""\x82""h""\x82""e""\x82""x",
	"\x82""h""\x82""m""\x82""u""\x82""h""\x82""r""\x82""h""\x82""a""\x82""h""\x82""k""\x82""h""\x82""s""\x82""x",
	"\x82""l""\x81""f""\x82""r""\x81""@""\x82""`""\x82""b""\x82""h""\x82""c""\x81""@""\x82""`""\x82""q""\x82""q""\x82""n""\x82""v",
	"\x82""c""\x82""h""\x82""r""\x82""o""\x82""d""\x82""k""\x81""@""\x82""l""\x82""`""\x82""f""\x82""h""\x82""b",
	"\x82""e""\x82""h""\x82""q""\x82""d""\x81""@""\x82""a""\x82""`""\x82""k""\x82""k",
	"\x82""g""\x82""`""\x82""r""\x82""s""\x82""d",
	"\x82""g""\x82""n""\x82""k""\x82""c""\x81""@""\x82""o""\x82""d""\x82""q""\x82""r""\x82""n""\x82""m",
	"\x82""h""\x82""m""\x82""u""\x82""h""\x82""r""\x82""h""\x82""a""\x82""h""\x82""k""\x82""h""\x82""s""\x82""x""\x81""@""\x82""P""\x82""O""\x81""f",
	"\x82""k""\x82""h""\x82""f""\x82""g""\x82""s""\x82""m""\x82""h""\x82""m""\x82""f""\x81""@""\x82""a""\x82""n""\x82""k""\x82""s",
	"\x82""u""\x82""`""\x82""l""\x82""o""\x82""h""\x82""q""\x82""h""\x82""b""\x81""@""\x82""s""\x82""n""\x82""t""\x82""b""\x82""g",
	"\x82""e""\x82""d""\x82""`""\x82""q",
	"\x82""h""\x82""b""\x82""d""\x81""@""\x82""r""\x82""s""\x82""n""\x82""q""\x82""l",
	"\x82""h""\x82""l""\x82""o""\x81""@""\x82""h""\x82""m""\x82""u""\x82""h""\x82""r""\x82""h""\x82""a""\x82""h""\x82""k""\x82""h""\x82""s""\x82""x",
	"\x82""q""\x82""d""\x82""l""\x82""n""\x82""u""\x82""d""\x81""@""\x82""b""\x82""t""\x82""q""\x82""r""\x82""d",
	"\x82""b""\x82""n""\x82""m""\x82""d""\x81""@""\x82""n""\x82""e""\x81""@""\x82""b""\x82""n""\x82""k""\x82""c",
	"\x82""g""\x82""n""\x82""k""\x82""c""\x81""@""\x82""l""\x82""n""\x82""m""\x82""r""\x82""s""\x82""d""\x82""q",
	"\x82""v""\x82""`""\x82""k""\x82""k""\x81""@""\x82""n""\x82""e""\x81""@""\x82""e""\x82""n""\x82""q""\x82""b""\x82""d",
	"\x82""c""\x82""h""\x82""r""\x82""h""\x82""m""\x82""s""\x82""d""\x82""f""\x82""q""\x82""`""\x82""s""\x82""d",
	"\x82""e""\x82""k""\x82""d""\x82""r""\x82""g""\x81""@""\x82""s""\x82""n""\x81""@""\x82""r""\x82""s""\x82""n""\x82""m""\x82""d",
	"\x82""r""\x82""s""\x82""n""\x82""m""\x82""d""\x81""@""\x82""s""\x82""n""\x81""@""\x82""e""\x82""k""\x82""d""\x82""r""\x82""g",
	"\x82""s""\x82""q""\x82""t""\x82""d""\x81""@""\x82""r""\x82""d""\x82""d""\x82""h""\x82""m""\x82""f",
	"\x82""e""\x82""h""\x82""m""\x82""f""\x82""d""\x82""q""\x81""@""\x82""n""\x82""e""\x81""@""\x82""c""\x82""d""\x82""`""\x82""s""\x82""g",
	"\x82""o""\x82""n""\x82""v""\x82""d""\x82""q""\x81""@""\x82""v""\x82""n""\x82""q""\x82""c""\x81""@""\x82""r""\x82""s""\x82""t""\x82""m",
	"\x82""a""\x82""h""\x82""f""\x82""a""\x82""x""\x81""f""\x82""r""\x81""@""\x82""e""\x82""h""\x82""r""\x82""s"
};

static const StringListProvider kEoB2MageSpellsListFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MageSpellsListFMTownsJapanese), kEoB2MageSpellsListFMTownsJapanese };

static const char *const kEoB2ClericSpellsListFMTownsJapanese[30] = {
	"",
	"\x82""a""\x82""k""\x82""d""\x82""r""\x82""r",
	"\x82""b""\x82""`""\x82""t""\x82""r""\x82""d""\x81""@""\x82""k""\x82""h""\x82""f""\x82""g""\x82""s""\x81""@""\x82""v""\x82""m""\x82""c""\x82""r",
	"\x82""b""\x82""t""\x82""q""\x82""d""\x81""@""\x82""k""\x82""h""\x82""f""\x82""g""\x82""s""\x81""@""\x82""v""\x82""m""\x82""c""\x82""r",
	"\x82""c""\x82""d""\x82""s""\x82""d""\x82""b""\x82""s""\x81""@""\x82""l""\x82""`""\x82""f""\x82""h""\x82""b",
	"\x82""o""\x82""q""\x82""n""\x82""s""\x82""d""\x82""b""\x82""s""\x81""|""\x82""d""\x82""u""\x82""h""\x82""k",
	"\x82""`""\x82""h""\x82""c",
	"\x82""e""\x82""k""\x82""`""\x82""l""\x82""d""\x81""@""\x82""a""\x82""k""\x82""`""\x82""c""\x82""d",
	"\x82""g""\x82""n""\x82""k""\x82""c""\x81""@""\x82""o""\x82""d""\x82""q""\x82""r""\x82""n""\x82""m",
	"\x82""r""\x82""k""\x82""n""\x82""v""\x81""@""\x82""o""\x82""n""\x82""h""\x82""r""\x82""n""\x82""m",
	"\x82""b""\x82""q""\x82""d""\x82""`""\x82""s""\x82""d""\x81""@""\x82""e""\x82""n""\x82""n""\x82""c",
	"\x82""c""\x82""h""\x82""r""\x82""o""\x82""d""\x82""k""\x81""@""\x82""l""\x82""`""\x82""f""\x82""h""\x82""b",
	"\x82""l""\x82""`""\x82""f""\x82""h""\x82""b""\x82""`""\x82""k""\x81""@""\x82""u""\x82""d""\x82""r""\x82""s""\x82""l""\x82""d""\x82""m""\x82""s",
	"\x82""o""\x82""q""\x82""`""\x82""x""\x82""d""\x82""q",
	"\x82""q""\x82""d""\x82""l""\x82""n""\x82""u""\x82""d""\x81""@""\x82""o""\x82""`""\x82""q""\x82""`""\x82""k""\x82""x""\x82""r""\x82""h""\x82""r",
	"\x82""b""\x82""`""\x82""t""\x82""r""\x82""d""\x81""@""\x82""r""\x82""d""\x82""q""\x82""h""\x82""n""\x82""t""\x82""r",
	"\x82""b""\x82""t""\x82""q""\x82""d""\x81""@""\x82""r""\x82""d""\x82""q""\x82""h""\x82""n""\x82""t""\x82""r",
	"\x82""m""\x82""d""\x82""t""\x82""s""\x82""q""\x82""`""\x82""k""\x81""|""\x82""o""\x82""n""\x82""h""\x82""r""\x82""n""\x82""m",
	"\x82""o""\x82""q""\x82""n""\x82""s""\x82""d""\x82""b""\x82""s""\x81""|""\x82""d""\x82""u""\x82""h""\x82""k""\x81""@""\x82""P""\x82""O""\x81""f",
	"\x82""b""\x82""`""\x82""t""\x82""r""\x82""d""\x81""@""\x82""b""\x82""q""\x82""h""\x82""s""\x82""h""\x82""b""\x82""`""\x82""k",
	"\x82""b""\x82""t""\x82""q""\x82""d""\x81""@""\x82""b""\x82""q""\x82""h""\x82""s""\x82""h""\x82""b""\x82""`""\x82""k",
	"\x82""e""\x82""k""\x82""`""\x82""l""\x82""d""\x81""@""\x82""r""\x82""s""\x82""q""\x82""h""\x82""j""\x82""d",
	"\x82""q""\x82""`""\x82""h""\x82""r""\x82""d""\x81""@""\x82""c""\x82""d""\x82""`""\x82""c",
	"\x82""r""\x82""k""\x82""`""\x82""x""\x81""@""\x82""k""\x82""h""\x82""u""\x82""h""\x82""m""\x82""f",
	"\x82""s""\x82""q""\x82""t""\x82""d""\x81""@""\x82""r""\x82""d""\x82""d""\x82""h""\x82""m""\x82""f",
	"\x82""g""\x82""`""\x82""q""\x82""l",
	"\x82""g""\x82""d""\x82""`""\x82""k",
	"\x82""q""\x82""d""\x82""r""\x82""r""\x82""t""\x82""q""\x82""d""\x82""b""\x82""s""\x82""h""\x82""n""\x82""m",
	"\x82""k""\x82""`""\x82""x""\x81""@""\x82""n""\x82""m""\x81""@""\x82""g""\x82""`""\x82""m""\x82""c""\x82""r",
	"\x82""s""\x82""t""\x82""q""\x82""m""\x81""@""\x82""t""\x82""m""\x82""c""\x82""d""\x82""`""\x82""c"
};

static const StringListProvider kEoB2ClericSpellsListFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2ClericSpellsListFMTownsJapanese), kEoB2ClericSpellsListFMTownsJapanese };

static const char *const kEoB2SpellNamesFMTownsJapanese[68] = {
	"\x82""`""\x82\x92\x82\x8D\x82\x8F\x82\x92",
	"\x82""a""\x82\x95\x82\x92\x82\x8E\x82\x89\x82\x8E\x82\x87\x81""@""\x82""g""\x82\x81\x82\x8E\x82\x84\x82\x93",
	"\x82""c""\x82\x85\x82\x94\x82\x85\x82\x83\x82\x94\x81""@""\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83",
	"\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83\x81""@""\x82""l""\x82\x89\x82\x93\x82\x93\x82\x89\x82\x8C\x82\x85",
	"\x82""r""\x82\x88\x82\x89\x82\x85\x82\x8C\x82\x84",
	"\x82""r""\x82\x88\x82\x8F\x82\x83\x82\x8B\x82\x89\x82\x8E\x82\x87\x81""@""\x82""f""\x82\x92\x82\x81\x82\x93\x82\x90",
	"\x82""a""\x82\x8C\x82\x95\x82\x92",
	"\x82""c""\x82\x85\x82\x94\x82\x85\x82\x83\x82\x94\x81""@""\x82""h""\x82\x8E\x82\x96\x82\x89\x82\x93\x82\x89\x82\x82\x82\x89\x82\x8C\x82\x89\x82\x94\x82\x99",
	"\x82""h""\x82\x8D\x82\x90\x82\x92\x82\x8F\x82\x96\x82\x85\x82\x84\x81""@""\x82""h""\x82\x84\x82\x85\x82\x8E\x82\x94\x82\x89\x82\x86\x82\x99",
	"\x82""h""\x82\x8E\x82\x96\x82\x89\x82\x93\x82\x89\x82\x82\x82\x89\x82\x8C\x82\x89\x82\x94\x82\x99",
	"\x82""l""\x82\x85\x82\x8C\x82\x86\x81""f""\x82\x93\x81""@""\x82""`""\x82\x83\x82\x89\x82\x84\x81""@""\x82""`""\x82\x92\x82\x92\x82\x8F\x82\x97",
	"\x82""c""\x82\x89\x82\x93\x82\x90\x82\x85\x82\x8C\x81""@""\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83",
	"\x82""e""\x82\x89\x82\x92\x82\x85\x82\x82\x82\x81\x82\x8C\x82\x8C",
	"\x82""g""\x82\x81\x82\x93\x82\x94\x82\x85",
	"\x82""g""\x82\x8F\x82\x8C\x82\x84\x81""@""\x82""o""\x82\x85\x82\x92\x82\x93\x82\x8F\x82\x8E",
	"\x82""h""\x82\x8E\x82\x96\x82\x89\x82\x93\x82\x89\x82\x82\x82\x89\x82\x8C\x82\x89\x82\x94\x82\x99\x81""@""\x82""P""\x82""O""\x81""f""\x82\x93\x81""@""\x82\x92\x82\x81\x82\x84\x82\x89\x82\x95\x82\x93",
	"\x82""k""\x82\x89\x82\x87\x82\x88\x82\x94\x82\x8E\x82\x89\x82\x8E\x82\x87\x81""@""\x82""a""\x82\x8F\x82\x8C\x82\x94",
	"\x82""u""\x82\x81\x82\x8D\x82\x90\x82\x89\x82\x92\x82\x89\x82\x83\x81""@""\x82""s""\x82\x8F\x82\x95\x82\x83\x82\x88",
	"\x82""e""\x82\x85\x82\x81\x82\x92",
	"\x82""h""\x82\x83\x82\x85\x81""@""\x82""r""\x82\x94\x82\x8F\x82\x92\x82\x8D",
	"\x82""h""\x82\x8D\x82\x90\x82\x92\x82\x8F\x82\x96\x82\x85\x82\x84\x81""@""\x82""h""\x82\x8E\x82\x96\x82\x89\x82\x93\x82\x89\x82\x82\x82\x89\x82\x8C\x82\x89\x82\x94\x82\x99",
	"\x82""q""\x82\x85\x82\x8D\x82\x8F\x82\x96\x82\x85\x81""@""\x82""b""\x82\x95\x82\x92\x82\x93\x82\x85",
	"\x82""b""\x82\x8F\x82\x8E\x82\x85\x81""@""\x82\x8F\x82\x86\x81""@""\x82""b""\x82\x8F\x82\x8C\x82\x84",
	"\x82""g""\x82\x8F\x82\x8C\x82\x84\x81""@""\x82""l""\x82\x8F\x82\x8E\x82\x93\x82\x94\x82\x85\x82\x92",
	"\x82""v""\x82\x81\x82\x8C\x82\x8C\x81""@""\x82\x8F\x82\x86\x81""@""\x82""e""\x82\x8F\x82\x92\x82\x83\x82\x85",
	"\x82""c""\x82\x89\x82\x93\x82\x89\x82\x8E\x82\x94\x82\x85\x82\x87\x82\x92\x82\x81\x82\x94\x82\x85",
	"\x82""e""\x82\x8C\x82\x85\x82\x93\x82\x88\x81""@""\x82\x94\x82\x8F\x81""@""\x82""r""\x82\x94\x82\x8F\x82\x8E\x82\x85",
	"\x82""r""\x82\x94\x82\x8F\x82\x8E\x82\x85\x81""@""\x82\x94\x82\x8F\x81""@""\x82""e""\x82\x8C\x82\x85\x82\x93\x82\x88",
	"\x82""s""\x82\x92\x82\x95\x82\x85\x81""@""\x82""r""\x82\x85\x82\x85\x82\x89\x82\x8E\x82\x87",
	"\x82""e""\x82\x89\x82\x8E\x82\x87\x82\x85\x82\x92\x81""@""\x82\x8F\x82\x86\x81""@""\x82\x84\x82\x85\x82\x81\x82\x94\x82\x88",
	"\x82""o""\x82\x8F\x82\x97\x82\x85\x82\x92\x81""@""\x82""v""\x82\x8F\x82\x92\x82\x84\x81""@""\x82""r""\x82\x94\x82\x95\x82\x8E",
	"\x82""a""\x82\x89\x82\x87\x82\x82\x82\x99\x81""f""\x82\x93\x81""@""\x82""b""\x82\x8C\x82\x85\x82\x8E\x82\x83\x82\x88\x82\x85\x82\x84\x81""@""\x82""e""\x82\x89\x82\x93\x82\x94",
	"\x82""a""\x82\x8C\x82\x85\x82\x93\x82\x93",
	"\x82""b""\x82\x81\x82\x95\x82\x93\x82\x85\x81""@""\x82""k""\x82\x89\x82\x87\x82\x88\x82\x94\x81""@""\x82""v""\x82\x8F\x82\x95\x82\x8E\x82\x84\x82\x93",
	"\x82""b""\x82\x95\x82\x92\x82\x85\x81""@""\x82""k""\x82\x89\x82\x87\x82\x88\x82\x94\x81""@""\x82""v""\x82\x8F\x82\x95\x82\x8E\x82\x84\x82\x93",
	"\x82""c""\x82\x85\x82\x94\x82\x85\x82\x83\x82\x94\x81""@""\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83",
	"\x82""o""\x82\x92\x82\x8F\x82\x94\x82\x85\x82\x83\x82\x94\x82\x89\x82\x8F\x82\x8E\x81""@""\x82\x86\x82\x92\x82\x8F\x82\x8D\x81""@""\x82""d""\x82\x96\x82\x89\x82\x8C",
	"\x82""`""\x82\x89\x82\x84",
	"\x82""e""\x82\x8C\x82\x81\x82\x8D\x82\x85\x81""@""\x82""a""\x82\x8C\x82\x81\x82\x84\x82\x85",
	"\x82""g""\x82\x8F\x82\x8C\x82\x84\x81""@""\x82""o""\x82\x85\x82\x92\x82\x93\x82\x8F\x82\x8E",
	"\x82""r""\x82\x8C\x82\x8F\x82\x97\x81""@""\x82""o""\x82\x8F\x82\x89\x82\x93\x82\x8F\x82\x8E",
	"\x82""b""\x82\x92\x82\x85\x82\x81\x82\x94\x82\x85\x81""@""\x82""e""\x82\x8F\x82\x8F\x82\x84",
	"\x82""c""\x82\x89\x82\x93\x82\x90\x82\x85\x82\x8C\x81""@""\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83",
	"\x82""l""\x82\x81\x82\x87\x82\x89\x82\x83\x82\x81\x82\x8C\x81""@""\x82""u""\x82\x85\x82\x93\x82\x94\x82\x8D\x82\x85\x82\x8E\x82\x94",
	"\x82""o""\x82\x92\x82\x81\x82\x99\x82\x85\x82\x92",
	"\x82""q""\x82\x85\x82\x8D\x82\x8F\x82\x96\x82\x85\x81""@""\x82""o""\x82\x81\x82\x92\x82\x81\x82\x8C\x82\x99\x82\x93\x82\x89\x82\x93",
	"\x82""b""\x82\x81\x82\x95\x82\x93\x82\x85\x81""@""\x82""r""\x82\x85\x82\x92\x82\x89\x82\x8F\x82\x95\x82\x93\x81""@""\x82""v""\x82\x8F\x82\x95\x82\x8E\x82\x84\x82\x93",
	"\x82""b""\x82\x95\x82\x92\x82\x85\x81""@""\x82""r""\x82\x85\x82\x92\x82\x89\x82\x8F\x82\x95\x82\x93\x81""@""\x82""v""\x82\x8F\x82\x95\x82\x8E\x82\x84\x82\x93",
	"\x82""m""\x82\x85\x82\x95\x82\x94\x82\x92\x82\x81\x82\x8C\x82\x89\x82\x9A\x82\x85\x81""@""\x82""o""\x82\x8F\x82\x89\x82\x93\x82\x8F\x82\x8E",
	"\x82""o""\x82\x92\x82\x8F\x82\x94\x82\x85\x82\x83\x82\x94\x82\x89\x82\x8F\x82\x8E\x81""@""\x82\x86\x82\x92\x82\x8F\x82\x8D\x81""@""\x82\x85\x82\x96\x82\x89\x82\x8C\x81""@""\x82""P""\x82""O""\x81""f""\x82\x93\x81""@""\x82\x92\x82\x81\x82\x84\x82\x89\x82\x95\x82\x93",
	"\x82""b""\x82\x81\x82\x95\x82\x93\x82\x85\x81""@""\x82""b""\x82\x92\x82\x89\x82\x94\x82\x89\x82\x83\x82\x81\x82\x8C\x81""@""\x82""v""\x82\x8F\x82\x95\x82\x8E\x82\x84\x82\x93",
	"\x82""b""\x82\x95\x82\x92\x82\x85\x81""@""\x82""b""\x82\x92\x82\x89\x82\x94\x82\x89\x82\x83\x82\x81\x82\x8C\x81""@""\x82""v""\x82\x8F\x82\x95\x82\x8E\x82\x84\x82\x93",
	"\x82""e""\x82\x8C\x82\x81\x82\x8D\x82\x85\x81""@""\x82""r""\x82\x94\x82\x92\x82\x89\x82\x8B\x82\x85",
	"\x82""q""\x82\x81\x82\x89\x82\x93\x82\x85\x81""@""\x82""c""\x82\x85\x82\x81\x82\x84",
	"\x82""r""\x82\x8C\x82\x81\x82\x99\x81""@""\x82""k""\x82\x89\x82\x96\x82\x89\x82\x8E\x82\x87",
	"\x82""s""\x82\x92\x82\x95\x82\x85\x81""@""\x82""r""\x82\x85\x82\x85\x82\x89\x82\x8E\x82\x87",
	"\x82""g""\x82\x81\x82\x92\x82\x8D",
	"\x82""g""\x82\x85\x82\x81\x82\x8C",
	"\x82""q""\x82\x85\x82\x93\x82\x93\x82\x95\x82\x92\x82\x85\x82\x83\x82\x94\x82\x89\x82\x8F\x82\x8E",
	"\x82""k""\x82\x81\x82\x99\x81""@""\x82\x8F\x82\x8E\x81""@""\x82""g""\x82\x81\x82\x8E\x82\x84\x82\x93",
	"\x82""s""\x82\x95\x82\x92\x82\x8E\x81""@""\x82""t""\x82\x8E\x82\x84\x82\x85\x82\x81\x82\x84",
	"",
	"\x82""l""\x82\x99\x82\x93\x82\x94\x82\x89\x82\x83\x81""@""\x82""c""\x82\x85\x82\x86\x82\x85\x82\x8E\x82\x83\x82\x85",
	"",
	"",
	"",
	"",
	""
};

static const StringListProvider kEoB2SpellNamesFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2SpellNamesFMTownsJapanese), kEoB2SpellNamesFMTownsJapanese };

static const char *const kEoB2MagicStrings1FMTownsJapanese[6] = {
	"ABORT SPELL",
	"ABORT SPELL",
	"\x82\xB1\x82\xCC\x8E\xF4\x95\xB6\x82\xCD\x8E\xE8\x82\xAA\x8B\xF3\x82\xA2\x82\xC4\x82\xA2\x82\xC8\x82\xAF\x82\xEA\x82\xCE\x82\xC8\x82\xE7\x82\xC8\x82\xA2\x81""B\r",
	"\x82\xB1\x82\xCC\x83""^""\x83""C""\x83""v""\x82\xCC\x96\x82\x96""@""\x82\xCD\x82""Q""\x82\xC2\x93\xAF\x8E\x9E\x82\xC9\x82\xA9\x82\xAF\x82\xE7\x82\xEA\x82\xC8\x82\xA2\x81""B\r",
	"%s""\x82\xCD""%s""\x82\xCC\x8E\xF4\x95\xB6\x82\xF0\x8F\xA5\x82\xA6\x82\xBD\x81""B\r",
	"\rok\r"
};

static const StringListProvider kEoB2MagicStrings1FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings1FMTownsJapanese), kEoB2MagicStrings1FMTownsJapanese };

static const char *const kEoB2MagicStrings2FMTownsJapanese[5] = {
	"\x8E\xF4\x95\xB6\x82\xCD\x8E\xB8\x94""s""\x82\xB5\x82\xBD\x81""I\r",
	"%s""\x82\xCD\x95\xAA\x89\xF0\x82\xB3\x82\xEA\x82\xBD\x81""I\r",
	"\x06\x06\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCD\x83""f""\x83""X""\x82\xCC\x8E\xF4\x95\xB6\x82\xF0\x8E\xF3\x82\xAF\x82\xBD\x81""I\r",
	"\x06\x06""%s""\x82\xCD\x83""R""\x81""[""\x83""Y""\x81""E""\x83""V""\x83\x8A\x83""A""\x83""X""\x81""E""\x83""E""\x81""[""\x83\x93\x83""Y""\x82\xCC\x8E\xF4\x95\xB6\x82\xF0\x8E\xF3\x82\xAF\x82\xBD\x81""I\r",
	"\x90\xCE\x89\xBB"
};

static const StringListProvider kEoB2MagicStrings2FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings2FMTownsJapanese), kEoB2MagicStrings2FMTownsJapanese };

static const char *const kEoB2MagicStrings3FMTownsJapanese[5] = {
	"\x92""N""\x82\xC9\x8E\xF4\x95\xB6\x82\xF0\x8F\xA5\x82\xA6\x82\xDC\x82\xB7\x82\xA9\x81""H ",
	"\r""\x8E\xF4\x95\xB6\x82\xCD\x92\x86\x92""f""\x82\xB3\x82\xEA\x82\xDC\x82\xB5\x82\xBD\x81""B\r",
	"%s""\x82\xCC""%s""\x82\xCC\x8E\xF4\x95\xB6\x82\xCC\x8C\xF8\x89\xCA\x82\xAA\x90\xD8\x82\xEA\x82\xBD\x81""B\r",
	"%s""\x82\xCD\x83\x82\x83\x93\x83""X""\x83""^""\x81""[""\x82\xC9\x93\x96\x82\xBD\x82\xE7\x82\xC8\x82\xA9\x82\xC1\x82\xBD\x81""B\r",
	"%s""\x82\xCD\x91""O""\x97\xF1\x82\xC9\x82\xA2\x82\xC8\x82\xAF\x82\xEA\x82\xCE\x8C\xF8\x89\xCA\x82\xAA\x82\xC8\x82\xA2\x81""I\r"
};

static const StringListProvider kEoB2MagicStrings3FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings3FMTownsJapanese), kEoB2MagicStrings3FMTownsJapanese };

static const char *const kEoB2MagicStrings4FMTownsJapanese[1] = {
	"\x8C\xF8\x89\xCA\x82\xAA\x82\xC8\x82\xA9\x82\xC1\x82\xBD\x81""B\r"
};

static const StringListProvider kEoB2MagicStrings4FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings4FMTownsJapanese), kEoB2MagicStrings4FMTownsJapanese };

static const char *const kEoB2MagicStrings6FMTownsJapanese[1] = {
	"%s""\x82\xCC\x83""A""\x81""[""\x83""}""\x81""[""\x82\xCC\x83""N""\x83\x89\x83""X""\x82\xCD\x8A\xF9\x82\xC9\x82""U""\x88\xC8\x89\xBA\x82\xBE\x81""B"
};

static const StringListProvider kEoB2MagicStrings6FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings6FMTownsJapanese), kEoB2MagicStrings6FMTownsJapanese };

static const char *const kEoB2MagicStrings7FMTownsJapanese[2] = {
	"%s""\x82\xC9\x82\xCD\x8A\xF9\x82\xC9""%s""\x82\xCC\x8C\xF8\x89\xCA\x82\xAA\x82\xA0\x82\xE9\x81""B\r",
	"\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCD\x8A\xF9\x82\xC9""%s""\x82\xCC\x8C\xF8\x89\xCA\x82\xAA\x82\xA0\x82\xE9\x81""B\r"
};

static const StringListProvider kEoB2MagicStrings7FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings7FMTownsJapanese), kEoB2MagicStrings7FMTownsJapanese };

static const char *const kEoB2MagicStrings8FMTownsJapanese[6] = {
	"\x8E\xF4\x95\xB6\x82\xCD\x8E\xB8\x94""s""\x82\xB5\x82\xBD\x81""B\r",
	"\x90""g""\x93\xAE\x82\xAB\x82\xAA\x82\xC6\x82\xEA\x82\xC8\x82\xA2",
	"\x8E\xF4\x95\xB6\x82\xCD\x8E\xB8\x94""s""\x82\xB5\x82\xBD\x81""I\r",
	"\x8E\xF4\x95\xB6\x82\xCD\x8E\xB8\x94""s""\x82\xB5\x82\xBD""\r",
	"\x93\xCD\x82\xAD\x94\xCD\x88\xCD\x82\xC9\x83\x82\x83\x93\x83""X""\x83""^""\x81""[""\x82\xAA\x82\xA2\x82\xC8\x82\xA2\x81""I\r",
	"%s""\x82\xC9\x82\xCD\x8A\xF9\x82\xC9\x83""G""\x83""C""\x83""h""\x82\xCC\x8E\xF4\x95\xB6\x82\xCC\x8C\xF8\x89\xCA\x82\xAA\x82\xA0\x82\xE9\x81""B\r"
};

static const StringListProvider kEoB2MagicStrings8FMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MagicStrings8FMTownsJapanese), kEoB2MagicStrings8FMTownsJapanese };

static const char *const kEoB2MainMenuStringsFMTownsJapanese[5] = {
	"\x83""Z""\x81""[""\x83""u""\x83""Q""\x81""[""\x83\x80\x82\xA9\x82\xE7\x8E""n""\x82\xDF\x82\xE9""\r",
	"\x90""V""\x82\xB5\x82\xAD\x83""p""\x81""[""\x83""e""\x83""B""\x82\xF0\x8D\xEC\x82\xC1\x82\xC4\x8E""n""\x82\xDF\x82\xE9""\r",
	"\x83\x86\x81""[""\x83""e""\x83""B""\x83\x8A\x83""e""\x83""B""\x81""[\r",
	"\x83""C""\x83\x93\x83""g""\x83\x8D\x83""_""\x83""N""\x83""V""\x83\x87\x83\x93""\r",
	"\x83""Q""\x81""[""\x83\x80\x82\xF0\x8F""I""\x97\xB9\x82\xB7\x82\xE9""\r"
};

static const StringListProvider kEoB2MainMenuStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MainMenuStringsFMTownsJapanese), kEoB2MainMenuStringsFMTownsJapanese };

static const char *const kEoB2MainMenuUtilStringsFMTownsJapanese[2] = {
	"\x89\xE6\x96\xCA\x83\x82\x81""[""\x83""h""\x82\xCC\x95\xCF\x8D""X\r",
	"\x83\x81\x83""C""\x83\x93\x83\x81\x83""j""\x83\x85\x81""[""\x82\xC9\x96\xDF\x82\xE9""\r"
};

static const StringListProvider kEoB2MainMenuUtilStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MainMenuUtilStringsFMTownsJapanese), kEoB2MainMenuUtilStringsFMTownsJapanese };

static const char *const kEoB2IntroStringsFMTownsJapanese[20] = {
	"\x82\xBB\x82\xCC\x94\xD3\x81""A""\x89\xE4\x81""X""\x82\xCD\x92""g""\x98""F""\x82\xCC\x82\xBB\x82\xCE\x82\xC5""\r""\x82\xA2\x82\xC2\x82\xE0\x82\xCC\x82\xE6\x82\xA4\x82\xC9\x82\xAD\x82\xC2\x82\xEB\x82\xA2\x82\xC5\x82\xA2\x82\xBD\x81""B",
	"\x82\xBB\x82\xB1\x82\xD6\x93\xCB\x91""R""\x82\xCC\x92""m""\x82\xE7\x82\xB9\x82\xAA""\r""\x95\x91\x82\xA2\x8D\x9E\x82\xF1\x82\xC5\x82\xAB\x82\xBD\x81""B",
	"\x82\xBB\x82\xEA\x82\xCD\x81""A""\x89\xE4\x81""X""\x82\xCC\x93\xAF\x8E""u""\x82\xC5\x82\xA0\x82\xE9""\r""\x83""E""\x83""H""\x81""[""\x83""^""\x81""[""\x83""f""\x83""B""\x81""[""\x83""v""\x82\xCC\x83""A""\x81""[""\x83""`""\x83\x81\x83""C""\x83""W\r""\x83""P""\x83\x8B\x83""x""\x83\x93\x81\x81\x83""u""\x83\x89\x83""b""\x83""N""\x83""X""\x83""^""\x83""b""\x83""t""\x82\xA9\x82\xE7\x82\xBE\x82\xC1\x82\xBD\x81""B",
	"\x8B\xD9\x8B""}""\x82\xCC\x97""p""\x8C\x8F\x82\xC6\x82\xCC\x82\xB1\x82\xC6\x82\xBE\x82\xAA\x81""E""\x81""E",
	"\x82\xA2\x82\xC1\x82\xBD\x82\xA2\x83""P""\x83\x8B\x83""x""\x83\x93\x82\xCD""\r""\x89\xBD\x82\xF0\x96""]""\x82\xF1\x82\xC5\x82\xA2\x82\xE9\x82\xCC\x82\xA9\x81""B",
	"\x82\xC7\x82\xA4\x82\xBC\x82\xA8\x93\xFC\x82\xE8\x82\xAD\x82\xBE\x82\xB3\x82\xA2\x81""B",
	"\x83""P""\x83\x8B\x83""x""\x83\x93\x97""l""\x82\xAA\x8F\x91\x8D\xD6\x82\xC5\x82\xA8\x91\xD2\x82\xBF\x82\xC5\x82\xB7\x81""B",
	"\x91""f""\x91\x81\x82\xA2\x93\x9E\x92\x85\x82\xC9\x8A\xB4\x8E\xD3\x82\xB7\x82\xE9\x81""B",
	"\x8D\xA2\x82\xC1\x82\xBD\x8E\x96\x82\xAA\x8B""N""\x82\xAB\x82\xC4\x82\xB5\x82\xDC\x82\xC1\x82\xBD\x81""B",
	"\x83""_""\x81""[""\x83""N""\x83\x80\x81""[""\x83\x93\x8E\x9B\x89""@""\x82\xC5""\r""\x8A\xF5\x82\xDC\x82\xED\x82\xB5\x82\xAB\x82\xE0\x82\xCC\x82\xAA""\r""\x93\xAE\x82\xAB\x8F""o""\x82\xB5\x82\xBD\x82\xE6\x82\xA4\x82\xBE\x81""B",
	"\x83""E""\x83""H""\x81""[""\x83""^""\x81""[""\x83""f""\x83""B""\x81""[""\x83""v""\x82\xCC\x88\xC0\x91""S""\x82\xAA""\r""\x8B\xBA\x82\xA9\x82\xB3\x82\xEA\x82\xC4\x82\xA2\x82\xE9\x81""B",
	"\x8C""N""\x92""B""\x82\xCC\x8F\x95\x82\xAF\x82\xAA\x95""K""\x97""v""\x82\xBE\x81""B",
	"\x8E""O""\x93\xFA\x91""O""\x92\xE3\x8E""@""\x82\xF0\x88\xEA\x90""l""\x94""h""\x8C\xAD\x82\xB5\x82\xBD\x82\xCC\x82\xBE\x82\xAA\x81""A",
	"\x94\xDE\x8F\x97\x82\xCD\x82\xA2\x82\xDC\x82\xBE\x82\xC9\x96\xDF\x82\xC1\x82\xC4\x82\xAB\x82\xC4\x82\xA2\x82\xC8\x82\xA2\x81""B",
	"\x94\xDE\x8F\x97\x82\xCC\x82\xB1\x82\xC6\x82\xE0\x90""S""\x94""z""\x82\xBE\x81""B",
	"\x82\xB1\x82\xCC\x83""R""\x83""C""\x83\x93\x82\xF0\x8E\x9D\x82\xC1\x82\xC4\x82\xA2\x82\xC1\x82\xC4\x82\xAD\x82\xEA\x81""B",
	"\x8C""N""\x92""B""\x82\xC6\x98""A""\x97\x8D\x82\xF0\x8E\xE6\x82\xE9\x8E\x9E\x82\xC9\x8E""g""\x82\xA4\x82\xE0\x82\xCC\x82\xBE\x81""B",
	"\x8E\x96\x82\xCD\x88\xEA\x8D\x8F\x82\xF0\x91\x88\x82\xA4\x81""B",
	"\x83""_""\x81""[""\x83""N""\x83\x80\x81""[""\x83\x93\x82\xCC\x8B\xDF\x82\xAD\x82\xDC\x82\xC5""\r""\x8C""N""\x92""B""\x82\xF0\x83""e""\x83\x8C\x83""|""\x81""[""\x83""g""\x82\xB7\x82\xE9\x81""B",
	"\x93\xAF\x8E""u""\x82\xC9\x90""_""\x82\xCC\x82\xB2\x89\xC1\x8C\xEC\x82\xAA\x82\xA0\x82\xE7\x82\xF1\x82\xB1\x82\xC6\x82\xF0\x81""B"
};

static const StringListProvider kEoB2IntroStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2IntroStringsFMTownsJapanese), kEoB2IntroStringsFMTownsJapanese };

static const char *const kEoB2FinaleStringsFMTownsJapanese[20] = {
	"\x82\xC2\x82\xA2\x82\xC9\x81""A""\x83""h""\x83\x89\x83\x93\x82\xCD\x97\xCD\x82\xC2\x82\xAB\x82\xBD\x81""B",
	"\x82\xBB\x82\xCC\x8E\x9E\x81""A""\x93\xAF\x8E""u""\x83""P""\x83\x8B\x83""x""\x83\x93\x82\xAA\x96\xDA\x82\xCC\x91""O""\x82\xC9\x8C\xBB\x82\xEA\x82\xBD\x81""B",
	"\x82\xA8\x82\xDF\x82\xC5\x82\xC6\x82\xA4\x81""A""\x8C\xD6\x82\xE8\x8D\x82\x82\xAB\x93\xAF\x8E""u""\x92""B""\x82\xE6\x81""B",
	"\x8C""N""\x92""B""\x82\xC9\x82\xE6\x82\xC1\x82\xC4\x83""h""\x83\x89\x83\x93\x82\xCD\x93""|""\x82\xB3\x82\xEA\x82\xBD\x81""I",
	"\x83""h""\x83\x89\x83\x93\x82\xAA\x83""h""\x83\x89\x83""S""\x83\x93\x82\xBE\x82\xC1\x82\xBD\x82\xC6\x82\xCD""\r""\x8E\x84\x82\xE0\x92""m""\x82\xE7\x82\xC8\x82\xA9\x82\xC1\x82\xBD\x81""B",
	"\x82\xB1\x82\xA2\x82\xC2\x82\xCD\x82""R""\x82""O""\x82""O""\x94""N""\x88\xC8\x8F\xE3""\r""\x90\xB6\x82\xAB\x82\xC4\x82\xA2\x82\xBD\x82\xC9\x88\xE1\x82\xA2\x82\xC8\x82\xA2\x81""I",
	"\x83""h""\x83\x89\x83\x93\x82\xCC\x97\xCD\x82\xCD\x8F\xC1\x82\xA6\x82\xBD\x81""B",
	"\x82\xB5\x82\xA9\x82\xB5\x81""A""\x83""_""\x81""[""\x83""N""\x83\x80\x81""[""\x83\x93\x82\xCD""\r""\x96\xA2\x82\xBE\x82\xC9\x8B\x90\x91\xE5\x82\xC8\x88\xAB\x82\xCC\x97\xCD\x82\xCC\x89\xF2\x82\xBE\x81""B",
	"\x82\xBB\x82\xEA\x82\xC9\x81""A""\x94\xDE\x82\xCC\x89\xBA\x96""l""\x92""B""\x82\xAA\x8E""c""\x82\xC1\x82\xC4\x82\xA2\x82\xE9\x81""B",
	"\x82\xB3\x82\xA0\x81""A""\x82\xB1\x82\xCC\x8F\xEA\x8F\x8A\x82\xF0\x97\xA3\x82\xEA\x82\xE9\x82\xC6\x82\xB5\x82\xE6\x82\xA4\x81""B",
	"\x8E\x84\x82\xCC\x97\xCD\x82\xC5\x94""j""\x89\xF3\x82\xB7\x82\xEA\x82\xCE\x81""A\r""\x8F""I""\x8E""~""\x95\x84\x82\xF0\x91\xC5\x82\xC2\x82\xB1\x82\xC6\x82\xAA\x82\xC5\x82\xAB\x82\xE9\x82\xBE\x82\xEB\x82\xA4\x81""B",
	"\x82\xC2\x82\xA2\x82\xC4\x97\x88\x82\xC8\x82\xB3\x82\xA2\x81""B",
	"\x8B\xAD\x97\xCD\x82\xC8\x83\x81\x83""C""\x83""W""\x92""B""\x82\xAA\x81""A""\x8D\xC5\x8C\xE3\x82\xCC\x8C\x88\x90\xED\x82\xCC\x88\xD7\x82\xC9""\r""\x83""_""\x81""[""\x83""N""\x83\x80\x81""[""\x83\x93\x82\xCC\x89\xF1\x82\xE8\x82\xC9\x97\xA7\x82\xC1\x82\xC4\x82\xA2\x82\xE9\x81""B",
	"\x83""_""\x81""[""\x83""N""\x83\x80\x81""[""\x83\x93\x82\xCC\x88\xAB\x82\xCC\x97\xCD\x82\xCD\x8B\xAD\x97\xCD\x82\xBE\x81""B",
	"\x90\xB6\x82\xA9\x82\xB5\x82\xC4\x82\xA8\x82\xA2\x82\xC4\x82\xCD\x82\xC8\x82\xE7\x82\xC8\x82\xA2\x81""I",
	"\x83""_""\x81""[""\x83""N""\x83\x80\x81""[""\x83\x93\x82\xCD\x8F\xC1\x82\xA6\x8B\x8E\x82\xC1\x82\xC4\x82\xA2\x82\xC1\x82\xBD\x81""B",
	"\x93\xAF\x8E""u""\x82\xE6\x81""B""\x82\xB1\x82\xEA\x82\xC5\x81""A""\x91""S""\x82\xC4\x82\xAA\x8F""I""\x82\xED\x82\xC1\x82\xBD\x81""B",
	"\x90""S""\x82\xA9\x82\xE7\x8A\xB4\x8E\xD3\x82\xB7\x82\xE9\x81""B",
	"\x8C""N""\x92""B""\x82\xC9\x8E\x84\x82\xCC\x8D\xC5\x8D\x82\x82\xCC\x8C""h""\x88\xD3\x82\xF0\x95\xA5\x82\xA8\x82\xA4\x81""B",
	"\x89\xE4\x81""X""\x82\xCD\x8C""N""\x92""B""\x82\xCC\x82\xB1\x82\xC6\x82\xF0""\r""\x82\xA2\x82\xC2\x82\xE0\x96""Y""\x82\xEA\x82\xC8\x82\xA2\x82\xBE\x82\xEB\x82\xA4\x81""B"
};

static const StringListProvider kEoB2FinaleStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2FinaleStringsFMTownsJapanese), kEoB2FinaleStringsFMTownsJapanese };

static const char *const kEoB2MonsterDistAttStringsFMTownsJapanese[5] = {
	"\x83""p""\x81""[""\x83""e""\x83""B""\x82\xCD\x83""T""\x83""C""\x83""L""\x83""b""\x83""N""\x83""}""\x83""C""\x83\x93\x83""h""\x83""u""\x83\x89\x83""X""\x83""g""\x82\xF0\x8E\xF3\x82\xAF\x82\xBD\x81""I\r",
	"\x96\x83\xE1\x83\x82\xB5\x82\xBD",
	"\x93\xC5\x82\xF0\x8E\xF3\x82\xAF\x82\xBD",
	"\x96\x83\xE1\x83\x82\xB5\x82\xBD",
	"\x90\xCE\x89\xBB\x82\xB5\x82\xBD"
};

static const StringListProvider kEoB2MonsterDistAttStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MonsterDistAttStringsFMTownsJapanese), kEoB2MonsterDistAttStringsFMTownsJapanese };

static const char *const kEoB2Npc1StringsFMTownsJapanese[2] = {
	"\x98""b""\x82\xB7",
	"\x8B\x8E\x82\xE9"
};

static const StringListProvider kEoB2Npc1StringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2Npc1StringsFMTownsJapanese), kEoB2Npc1StringsFMTownsJapanese };

static const char *const kEoB2Npc2StringsFMTownsJapanese[2] = {
	"\x8D\xBD\x82\xF0\x8A""O""\x82\xB7",
	"\x8B\x8E\x82\xE9"
};

static const StringListProvider kEoB2Npc2StringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2Npc2StringsFMTownsJapanese), kEoB2Npc2StringsFMTownsJapanese };

static const char *const kEoB2MonsterDustStringsFMTownsJapanese[1] = {
	"\x83\x82\x83\x93\x83""X""\x83""^""\x81""|""\x82\xCD\x8D\xD3\x82\xAF\x8E""U""\x82\xC1\x82\xBD\x81""I\r"
};

static const StringListProvider kEoB2MonsterDustStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2MonsterDustStringsFMTownsJapanese), kEoB2MonsterDustStringsFMTownsJapanese };

static const char *const kEoB2KheldranStringsFMTownsJapanese[1] = {
	"    ""\x82\xB1\x82\xCC\x8F\xAC\x91""m""\x82\xDF\x82\xAA\x81""I"
};

static const StringListProvider kEoB2KheldranStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2KheldranStringsFMTownsJapanese), kEoB2KheldranStringsFMTownsJapanese };

static const char *const kEoB2HornStringsFMTownsJapanese[4] = {
	"\x83""z""\x81""[""\x83\x93\x82\xA9\x82\xE7\x82\xA4\x82\xC8\x82\xE9\x82\xE6\x82\xA4\x82\xC8\x89\xB9\x82\xAA\x8B\xBF\x82\xA2\x82\xBD\x81""B\r",
	"\x83""z""\x81""[""\x83\x93\x82\xA9\x82\xE7\x82\xA4\x82\xC2\x82\xEB\x82\xC8\x89\xB9\x82\xAA\x95\xB7\x82\xB1\x82\xA6\x82\xE9\x81""B\r",
	"\x83""z""\x81""[""\x83\x93\x82\xA9\x82\xE7\x94\xFC\x82\xB5\x82\xA2\x89\xB9\x82\xAA\x97\xAC\x82\xEA\x82\xE9\x81""B\r",
	"\x83""z""\x81""[""\x83\x93\x82\xA9\x82\xE7\x95""s""\x8B""C""\x96\xA1\x82\xC8\x89\xB9\x82\xAA\x82\xB7\x82\xE9\x81""B\r"
};

static const StringListProvider kEoB2HornStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2HornStringsFMTownsJapanese), kEoB2HornStringsFMTownsJapanese };

static const char *const kEoB2UtilMenuStringsFMTownsJapanese[2] = {
	"\x89\xE6\x96\xCA\x83\x82\x81\x5B\x83\x68\x82\xCC\x95\xCF\x8D\x58\r",
	"\x83\x81\x83\x43\x83\x93\x83\x81\x83\x6A\x83\x85\x81\x5B\x82\xC9\x96\xDF\x82\xE9\r"
};

static const StringListProvider kEoB2UtilMenuStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2UtilMenuStringsFMTownsJapanese), kEoB2UtilMenuStringsFMTownsJapanese };

static const char *const kEoB2Config2431StringsFMTownsJapanese[2] = {
	"\x82\x51\x82\x53\x82\x6A\x82\x67\x82\x9A",
	"\x82\x52\x82\x50\x82\x6A\x82\x67\x82\x9A"
};

static const StringListProvider kEoB2Config2431StringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2Config2431StringsFMTownsJapanese), kEoB2Config2431StringsFMTownsJapanese };

static const char *const kEoB2TextInputCharacterLinesFMTownsJapanese[12] = {
	"\x83""A""\x83""C""\x83""E""\x83""G""\x83""I""\x81""@""\x83""J""\x83""L""\x83""N""\x83""P""\x83""R""\x81""@""\x83""T""\x83""V""\x83""X""\x83""Z""\x83""\\",
	"\x83""^""\x83""`""\x83""c""\x83""e""\x83""g""\x81""@""\x83""i""\x83""j""\x83""k""\x83""l""\x83""m""\x81""@""\x83""n""\x83""q""\x83""t""\x83""w""\x83""z",
	"\x83""}""\x83""~""\x83\x80\x83\x81\x83\x82\x81""@""\x83\x84\x81""@""\x83\x86\x81""@""\x83\x88\x81""@""\x83\x89\x83\x8a\x83\x8b\x83\x8c\x83\x8d",
	"\x83\x8f\x83\x92\x83\x93\x81""@""\x81""[""\x81""@""\x81""Q""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@",
	"\x83""K""\x83""M""\x83""O""\x83""Q""\x83""S""\x81""@""\x83""U""\x83""W""\x83""Y""\x83""[""\x83""]""\x81""@""\x83""_""\x83""a""\x83""Y""\x83""f""\x83""h",
	"\x83""o""\x83""r""\x83""u""\x83""x""\x83""{""\x81""@""\x83""p""\x83""s""\x83""v""\x83""y""\x83""|""\x81""@""\x83""@""\x83""B""\x83""D""\x83""F""\x83""H",
	"\x83\x83\x83\x85\x83\x87\x81""@""\x83""b""\x81""@""\x81""Q""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@",
	"\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@",
	"\x82""`""\x82""a""\x82""b""\x82""c""\x82""d""\x81""@""\x82""e""\x82""f""\x82""g""\x82""h""\x82""i""\x81""@""\x82""j""\x82""k""\x82""l""\x82""m""\x82""n",
	"\x82""o""\x82""p""\x82""q""\x82""r""\x82""s""\x81""@""\x82""t""\x82""u""\x82""v""\x82""w""\x82""x""\x81""@""\x82""y""\x81""@""\x81""D""\x81""C""\x81""Q",
	"\x82""O""\x82""P""\x82""Q""\x82""R""\x82""S""\x82""T""\x82""U""\x82""V""\x82""W""\x82""X""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@",
	"\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@""\x81""@"
};

static const StringListProvider kEoB2TextInputCharacterLinesFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2TextInputCharacterLinesFMTownsJapanese), kEoB2TextInputCharacterLinesFMTownsJapanese };

static const char *const kEoB2TextInputSelectStringsFMTownsJapanese[3] = {
	"NEXT",
	"END",
	"DEL"
};

static const StringListProvider kEoB2TextInputSelectStringsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2TextInputSelectStringsFMTownsJapanese), kEoB2TextInputSelectStringsFMTownsJapanese };

static const char *const kEoB2NpcPresetsNamesFMTownsJapanese[6] = {
	 "\x83""C""\x83\x93\x83""T""\x83\x8B",
	 "\x83""J""\x83\x89\x83\x93\x83""h""\x83\x89",
	"\x83""V""\x83\x87\x81""[""\x83\x93",
	"\x83""T""\x83\x93\x83\x89\x81""[""\x83\x8B",
	"\x83""^""\x83\x93\x83""O""\x83\x8D\x81""[",
	"\x83""A""\x83\x93\x83""o""\x81""["
};

static const StringListProvider kEoB2NpcPresetsNamesFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2NpcPresetsNamesFMTownsJapanese), kEoB2NpcPresetsNamesFMTownsJapanese };

static const char *const kEoB2SaveNamePatternsFMTownsJapanese[34] = {
	"NULL",
	"NULL",
	"CATACOMB LV1 %2d/%2d %2d:%02d ",
	"CATACOMB LV1 %2d/%2d %2d:%02d ",
	"CATACOMB LV2 %2d/%2d %2d:%02d ",
	"CATACOMB LV2 %2d/%2d %2d:%02d ",
	"CATACOMB LV3 %2d/%2d %2d:%02d ",
	"CATACOMB LV4 %2d/%2d %2d:%02d ",
	"FOREST       %2d/%2d %2d:%02d ",
	"FOREST       %2d/%2d %2d:%02d ",
	"TEMPLE   LV1 %2d/%2d %2d:%02d ",
	"TEMPLE   LV1 %2d/%2d %2d:%02d ",
	"TEMPLE   LV2 %2d/%2d %2d:%02d ",
	"TEMPLE   LV2 %2d/%2d %2d:%02d ",
	"SILVER   LV1 %2d/%2d %2d:%02d ",
	"SILVER   LV1 %2d/%2d %2d:%02d ",
	"SILVER   LV2 %2d/%2d %2d:%02d ",
	"SILVER   LV2 %2d/%2d %2d:%02d ",
	"SILVER   LV3 %2d/%2d %2d:%02d ",
	"SILVER   LV3 %2d/%2d %2d:%02d ",
	"AZURE    LV1 %2d/%2d %2d:%02d ",
	"AZURE    LV1 %2d/%2d %2d:%02d ",
	"AZURE    LV2 %2d/%2d %2d:%02d ",
	"AZURE    LV2 %2d/%2d %2d:%02d ",
	"AZURE    LV3 %2d/%2d %2d:%02d ",
	"AZURE    LV3 %2d/%2d %2d:%02d ",
	"AZURE    LV4 %2d/%2d %2d:%02d ",
	"AZURE    LV4 %2d/%2d %2d:%02d ",
	"FROST GIANT  %2d/%2d %2d:%02d ",
	"FROST GIANT  %2d/%2d %2d:%02d ",
	"CRIMSON  LV1 %2d/%2d %2d:%02d ",
	"CRIMSON  LV1 %2d/%2d %2d:%02d ",
	"CRIMSON  LV2 %2d/%2d %2d:%02d ",
	"CRIMSON  LV2 %2d/%2d %2d:%02d "
};

static const StringListProvider kEoB2SaveNamePatternsFMTownsJapaneseProvider = { ARRAYSIZE(kEoB2SaveNamePatternsFMTownsJapanese), kEoB2SaveNamePatternsFMTownsJapanese };
