static const char *const k3MainMenuStringsDOSCDItalian[4] = {
	"Nuova Partita",
	"Introduzione",
	"Carica una partita",
	"Esci dal gioco"
};

static const StringListProvider k3MainMenuStringsDOSCDItalianProvider = { ARRAYSIZE(k3MainMenuStringsDOSCDItalian), k3MainMenuStringsDOSCDItalian };

static const char *const k3VqaSubtitlesIntroDOSCDItalian[] = {
	""// TODO
};

static const StringListProvider k3VqaSubtitlesIntroDOSCDItalianProvider = { ARRAYSIZE(k3VqaSubtitlesIntroDOSCDItalian), k3VqaSubtitlesIntroDOSCDItalian };

static const char *const k3VqaSubtitlesBoatDOSCDItalian[] = {
	""// TODO
};

static const StringListProvider k3VqaSubtitlesBoatDOSCDItalianProvider = { ARRAYSIZE(k3VqaSubtitlesBoatDOSCDItalian), k3VqaSubtitlesBoatDOSCDItalian };
