static const char *const k1IntroStringsDOSItalian[50] = {
	"This is a text test - 1",
	"This is a text test - 2",
	"Salute, Kallak...",
	"...Potente capo dei Mistici reali!",
	"Ti ho spaventato?",
	"Malcolm!",
	"Ieri ho saputo della tua fuga.",
	"E ti aspettavo prima di adesso.",
	"Perch""\x82"" affrettarmi?",
	"Sono io che governo questa terra!",
	"La tua gracile maledizione non mi ha trattenuto affatto.",
	"Quindi, uccidimi adesso!",
	"Mi ""\x8A"" rimasta poca magia.",
	"Intendo danneggiarti, non ho dubbi...",
	"...Adesso tu sei di pietra...",
	"...tuttavia, lascio i tuoi occhi.",
	"Non ho versato alcuna lacrima per Kyrandia...",
	"Ma non posso negarti le tue.",
	"Malcolm il buffone si ""\x8A"" liberato!",
	"Adesso ha il controllo della GemmaKyra...",
	"...fonte di tutta la magia in Kyrandia!",
	"",
	"Come osi farMi lo sgambetto?",
	"Ma niente, qualcosa di speciale per te...",
	"Un malocchio spassoso!",
	"Dar""\x95"" un giusto avvertimento.",
	"Non saltare su QUELL'albero!",
	"Non salire su QUELL'albero piccolo scoiattolo!",
	"Bo!",
	"E' divertente in questo modo...",
	"...non ""\x8A"" vero?",
	"Ah ah ah...",
	"Non sono fantastiche le dita?",
	"E' meraviglioso essere vivo!",
	"Brandon!",
	"Nonno! Ce l'abbiamo fatta!",
	"No! TU ce l'hai fatta!",
	"Lunga vita a Re Brandon!",
	"Lunga vita ai Mistici Reali!",
	"Adesso, rimettiamo insieme Kyrandia!",
	"Buona idea!",
	"Come mia prima Proclamazione Reale...",
	"Nomino i sandali la calzatura ufficiale di Kyrandia!",
	"Ben fatto, Brandon!",
	"La Terra non versa pi""\x97"" lacrime...",
	"E i tuoi amici sono stati recuperati!",
	"Caro Brynn...",
	"Malcolm si ""\x8A"" liberato!",
	"Presto mi verr""\x85"" a cercare.",
	"Per favore aiuta Brandon..."
};

static const StringListProvider k1IntroStringsDOSItalianProvider = { ARRAYSIZE(k1IntroStringsDOSItalian), k1IntroStringsDOSItalian };

static const char *const k1ItemNamesDOSItalian[107] = {
	"Granito rosso",
	"Ametista",
	"Acquamarina",
	"Diamante",
	"Smeraldo",
	"Perla",
	"Rubino",
	"Olivina",
	"Zaffiro",
	"Opale",
	"Topazio",
	"Onice",
	"Pietra Solare",
	"Pietra Lunare",
	"Pietra-arcobaleno",
	"Magnetite",
	"Rosa",
	"Tulipano",
	"Orchidea",
	"Rosa d'Argento",
	"Statuetta d'Argento",
	"Moneta d'Argento",
	"Moneta d'Oro",
	"Anello d'Oro",
	"Calice Reale",
	"Pigna",
	"Ghianda",
	"Noce",
	"Baccheinfuocate Sempreardenti",
	"Baccheinfuocate",
	"Baccheinfuocate",
	"Baccheinfuocate",
	"Baccheinfuocate",
	"Baccheinfuocate",
	"Pesce",
	"Lisca di Pesce",
	"Zampa di Montone",
	"Osso",
	"Mela",
	"Torsolo di Mela",
	"Mirtilli",
	"Fungo",
	"Nota",
	"Marmo",
	"Sega",
	"Croce ansata",
	"Piuma",
	"Uovo",
	"Foglia",
	"Trifoglio",
	"Stella Cadente",
	"Sfera di Cristallo",
	"Lacrima",
	"Specchio",
	"Frammento di ghiaccio",
	"Flauto",
	"Clessidra",
	"Chiave di Ferro",
	"Chiave di Giada",
	"Chiave di Ossidione",
	"Pozione Rossa",
	"Pozione Rossa",
	"Pozione Azzurra",
	"Pozione Azzurra",
	"Pozione Gialla",
	"Pozione Gialla",
	"Pozione Verde",
	"Pozione Arancione",
	"Pozione viola",
	"Pozione di Arcobaleni",
	"Acqua Fresca",
	"Acqua Fresca",
	"Acqua Salata",
	"Acqua Salata",
	"Acqua Minerale",
	"Acqua Minerale",
	"Acqua Magica",
	"Acqua Magica",
	"Bottiglietta Vuota",
	"Bottiglietta Vuota",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Pergamena",
	"Frammento di Pergamena",
	"Frammento di Pergamena",
	"Frammento di Pergamena",
	"Frammento di Pergamena",
	"Frammento di Pergamena",
	"Pietramagica Rossa",
	"Pietramagica Arancione",
	"Pietramagica Gialla",
	"Pietramagica Verde",
	"Pietramagica Verde-Azzurra",
	"Pietramagica Azzurra",
	"Pietramagica Viola",
	"Roccia Pesante",
	"Corona Reale",
	"Scettro Reale",
	"Chiave d'oro",
	"Articolo non conosciuto"
};

static const StringListProvider k1ItemNamesDOSItalianProvider = { ARRAYSIZE(k1ItemNamesDOSItalian), k1ItemNamesDOSItalian };

static const char *const k1TakenStringsDOSItalian[2] = {
	" - preso.",
	" - preso."
};

static const StringListProvider k1TakenStringsDOSItalianProvider = { ARRAYSIZE(k1TakenStringsDOSItalian), k1TakenStringsDOSItalian };

static const char *const k1PlacedStringsDOSItalian[1] = {
	" - piazzato."
};

static const StringListProvider k1PlacedStringsDOSItalianProvider = { ARRAYSIZE(k1PlacedStringsDOSItalian), k1PlacedStringsDOSItalian };

static const char *const k1DroppedStringsDOSItalian[1] = {
	" - caduto."
};

static const StringListProvider k1DroppedStringsDOSItalianProvider = { ARRAYSIZE(k1DroppedStringsDOSItalian), k1DroppedStringsDOSItalian };

static const char *const k1NoDropStringsDOSItalian[2] = {
	"Qui, non ci puoi lasciare altri oggetti.",
	"Non lo puoi mettere laggiu."
};

static const StringListProvider k1NoDropStringsDOSItalianProvider = { ARRAYSIZE(k1NoDropStringsDOSItalian), k1NoDropStringsDOSItalian };

static const char *const k1PutDownStringDOSItalian[1] = {
	"Forse prima sarebbe meglio che mettessi gi""\x97"" questo."
};

static const StringListProvider k1PutDownStringDOSItalianProvider = { ARRAYSIZE(k1PutDownStringDOSItalian), k1PutDownStringDOSItalian };

static const char *const k1WaitAmuletStringDOSItalian[1] = {
	"Suppongo che dovr""\x95"" aspettare che il mio Amuleto riacquisti il suo potere."
};

static const StringListProvider k1WaitAmuletStringDOSItalianProvider = { ARRAYSIZE(k1WaitAmuletStringDOSItalian), k1WaitAmuletStringDOSItalian };

static const char *const k1BlackJewelStringDOSItalian[1] = {
	"E' un gioiello, ma perch""\x82"" ""\x8A"" nero?"
};

static const StringListProvider k1BlackJewelStringDOSItalianProvider = { ARRAYSIZE(k1BlackJewelStringDOSItalian), k1BlackJewelStringDOSItalian };

static const char *const k1HealingTipStringDOSItalian[1] = {
	"Accidenti, questo potrebbe essere utile nel caso fossi ferito."
};

static const StringListProvider k1HealingTipStringDOSItalianProvider = { ARRAYSIZE(k1HealingTipStringDOSItalian), k1HealingTipStringDOSItalian };

static const char *const k1PoisonGoneStringDOSItalian[2] = {
	"Incredibile!",
	"Gli effetti del veleno sono scomparsi!"
};

static const StringListProvider k1PoisonGoneStringDOSItalianProvider = { ARRAYSIZE(k1PoisonGoneStringDOSItalian), k1PoisonGoneStringDOSItalian };

static const char *const k1ThePoisonStringsDOSItalian[4] = {
	"Il veleno...",
	"Non posso respirare...",
	"Non mi sento molto bene...",
	"Quel serpente deve\ressere velenoso!"
};

static const StringListProvider k1ThePoisonStringsDOSItalianProvider = { ARRAYSIZE(k1ThePoisonStringsDOSItalian), k1ThePoisonStringsDOSItalian };

static const char *const k1FluteStringsDOSItalian[2] = {
	"Un suono non molto impressionante.",
	"Quell'ultima nota era sicuramente acuta!"
};

static const StringListProvider k1FluteStringsDOSItalianProvider = { ARRAYSIZE(k1FluteStringsDOSItalian), k1FluteStringsDOSItalian };

static const char *const k1WispJewelStringsDOSItalian[3] = {
	"Non mi sento effettivamente in grado di farlo, per adesso.",
	"Probabilmente dovrei metter gi""\x97"" il mio",
	", per prima cosa."
};

static const StringListProvider k1WispJewelStringsDOSItalianProvider = { ARRAYSIZE(k1WispJewelStringsDOSItalian), k1WispJewelStringsDOSItalian };

static const char *const k1MagicJewelStringsDOSItalian[1] = {
	"Questo mi ""\x8A"" sembrato strano."
};

static const StringListProvider k1MagicJewelStringsDOSItalianProvider = { ARRAYSIZE(k1MagicJewelStringsDOSItalian), k1MagicJewelStringsDOSItalian };

static const char *const k1FlaskFullStringDOSItalian[1] = {
	"Questa bottiglietta ""\x8A"" gi""\x85"" piena."
};

static const StringListProvider k1FlaskFullStringDOSItalianProvider = { ARRAYSIZE(k1FlaskFullStringDOSItalian), k1FlaskFullStringDOSItalian };

static const char *const k1FullFlaskStringDOSItalian[4] = {
	"Adesso la bottiglietta ""\x8A"" riempita\rcon acqua fresca e gasata.",
	"Adesso la bottiglietta ""\x8A"" riempita\rcon acqua salata.",
	"Adesso la bottiglietta ""\x8A"" riempita\rcon acqua minerale.",
	"Acqua magica."
};

static const StringListProvider k1FullFlaskStringDOSItalianProvider = { ARRAYSIZE(k1FullFlaskStringDOSItalian), k1FullFlaskStringDOSItalian };

static const char *const k1OutroHomeStringDOSItalian[1] = {
	"Casa"
};

static const StringListProvider k1OutroHomeStringDOSItalianProvider = { ARRAYSIZE(k1OutroHomeStringDOSItalian), k1OutroHomeStringDOSItalian };

static const char *const k1VeryCleverStringDOSItalian[1] = {
	"Molto intelligente!  Ma i tuoi deboli sforzi sono vani."
};

static const StringListProvider k1VeryCleverStringDOSItalianProvider = { ARRAYSIZE(k1VeryCleverStringDOSItalian), k1VeryCleverStringDOSItalian };

static const char *const k1GUIStringsDOSItalian[60] = {
	"The Legend of Kyrandia",
	"Carica un gioco",
	"Salva questo gioco",
	"Comandi di Gioco",
	"Abbandona il gioco",
	"Ricomincia il gioco",
	"Comandi di gioco",
	"Quale gioco desideri ricaricare?",
	"Seleziona una posizione da salvare:",
	"[ GRIGLIA VUOTA ]",
	"Annulla",
	"Inserisci descrizione del tuo gioco salvato:",
	"Salva",
	"Riposa in pace, Brandon.",
	"Sei sicuro di voler abbandonare il gioco?",
	"XXXXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXXXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXXXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"XXXXXXXXX",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"Menu principale",
	"acceso",
	"spento",
	"Si",
	"No",
	"Velocit""\x85"" di movimento",
	"Velocit""\x85"" del testo",
	"Musica \x8A",
	"Suoni sono"
};

static const StringListProvider k1GUIStringsDOSItalianProvider = { ARRAYSIZE(k1GUIStringsDOSItalian), k1GUIStringsDOSItalian };

static const char *const k1NewGameStringDOSItalian[1] = {
	"[ INIZIARE UN NUOVO GIOCO ]"
};

static const StringListProvider k1NewGameStringDOSItalianProvider = { ARRAYSIZE(k1NewGameStringDOSItalian), k1NewGameStringDOSItalian };

static const char *const k1ConfigStringsDOSItalian[9] = {
	"molto lento",
	"lento",
	"normale",
	"veloce",
	"molto veloce",
	"molto lento",
	"normale",
	"veloce",
	"cliccabile"
};

static const StringListProvider k1ConfigStringsDOSItalianProvider = { ARRAYSIZE(k1ConfigStringsDOSItalian), k1ConfigStringsDOSItalian };

