/* Copyright (c) <2003-2011> <Julio Jerez, Newton Game Dynamics>
*
* This software is provided 'as-is', without any express or implied
* warranty. In no event will the authors be held liable for any damages
* arising from the use of this software.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
* claim that you wrote the original software. If you use this software
* in a product, an acknowledgment in the product documentation would be
* appreciated but is not required.
*
* 2. Altered source versions must be plainly marked as such, and must not be
* misrepresented as being the original software.
*
* 3. This notice may not be removed or altered from any source distribution.
*/

#if !defined(AFX_dgConnectorConstraint_H__FE88214E_43D0_4913_89B1_C2D5216E5997__INCLUDED_)
#define AFX_dgConnectorConstraint_H__FE88214E_43D0_4913_89B1_C2D5216E5997__INCLUDED_


#include "dgBallConstraint.h"

//template<class T> class dgPool;

/*
class dgConnectorConstraint: public dgBallConstraint
{
//  void SetPivotPoint (const dgVector& pivot){}

private:
    void Setup ();
    virtual dgUnsigned32 JacobianDerivative (dgContraintDescritor& params);

    dgConnectorConstraint();
    virtual ~dgConnectorConstraint();

    static dgConnectorConstraint* Create();
    virtual void Remove();

    friend class dgWorld;
    friend class dgConnectorConstraintArray;
    friend class dgPool<dgConnectorConstraint>;
};

*/
#endif // !defined(AFX_dgConnectorConstraint_H__FE88214E_43D0_4913_89B1_C2D5216E5997__INCLUDED_)

