static const char *const k2SeqplayStringsDOSPolish[103] = {
    "Kyrandia znika!",
    "Ska""\x88""a po skale...",
    "...i drzewo po drzewie.",
    "Kyrandia przestaje istnie""\x86""!",
    "Kr""\xA2""lewscy Mistycy s""\xA5"" zbici z tropu.",
    "Zbadano ka""\xBE""de przes""\x88""anki.",
    "Na konferencj""\xA9"" zaproszono nawet Marka wraz ze sw""\xA5"" now""\xA5"" przyjaci""\xA2""""\x88""k""\xA5"", R""\xA9""k""\xA5"".",
    "Na szcz""\xA9""""\x98""cie, R""\xA9""ka mia""\x88""a do""\x98""wiadczenie w tych sprawach,",
    "I w ko""\xE4""cu plan zosta""\x88"" zatwierdzony.",
    "Wymaga""\x88"" on zdobycia magicznego kamienia...",
    "...znajduj""\xA5""cego si""\xA9"" w centrum ""\x98""wiata.",
    "Zanthia, najm""\x88""odsza z mistyk""\xA2""w Kyrnadii, zosta""\x88""a wybrana do tego zadania.",
    "Dzi""\xA9""kujemy za granie w Hand of Fate.",
    "Powinno wystarczy""\x86"" jag""\xA2""d, by otworzy""\x86"" magiczne przej""\x98""cie do centrum ""\x98""wiata.",
	" DUMMY STRING... ",
	" DUMMY STRING... ",
    "Hej! Ukradziono wszystkie moje rzeczy!",
	" DUMMY STRING... ",
    "Je""\xBE""eli Rada s""\xA5""dzi, ""\xBE""e p""\xA2""jd""\xA9"" tam pieszo, to si""\xA9"" grubo myli!",
	" DUMMY STRING... ",
	" DUMMY STRING... ",
    "Pospiesz si""\xA9"" Faun!",
    "O rany, ma""\x88""o brakowa""\x88""o!",
    "Masz racj""\xA9"", bracie. Ja ju""\xBE"" nie zamierzam wi""\xA9""cej polowa""\x86""!",
    "Kraaa.",
    "Ile razy mam Ci powtarza""\x86"", ""\xBE""e jeste""\x98"" ropuch""\xA5""?",
    "Ojej! Ser nam si""\xA9"" sko""\xE4""czy""\x88""!",
    "Spr""\xA2""bujmy tego miodu z uszu. Jest ""\xBE""""\xA2""""\x88""ty.",
    "Mamo, kiedy dostan""\xA9"" mleczko?",
    "Sio! Wynocha!",
    "Ty tniesz, a ja wybieram.",
    "Nie. To ty tniesz, a ja wybieram.",
    "Dalej s""\xA5""dz""\xA9"", ""\xBE""e to by""\x88""y rymy cz""\xA9""stochowskie.",
    "Och, nie rozpozna""\x88""by""\x98"" czterowiersza nawet gdybym Ci""\xA9"" ugryz""\x88"" w ty""\x88""ek!",
    "Producent",
    "Brett W. Sperry",
    "Projekt i re""\xBE""yseria",
    "Rick Gush",
    "Programowanie",
    "Michael Legg",
    "Opracowanie artystyczne",
    "Louis Castle",
    "Joseph B. Hewitt IV",
    "G""\x88""""\xA2""wny artysta",
    "Rick Parks",
    "Dodatkowe kodowanie",
    "Philip W. Gorrow",
    "Matt Collins",
    "Mark McCubbin",
    "Arty""\x98""ci",
    "Cameron Chun",
    "Cary Averett",
    "Cindy Chinn",
    "Elie Arabian",
    "Fei Cheng",
    "Ferby Miguel",
    "Frank Mendeola",
    "Jack Martin",
    "Jerry Moore",
    "",
    "Judith Peterson",
    "Larry Miller",
    "Lenny Lee",
    "Louise Sandoval",
    "Ren Olsen",
    "Muzyka i d""\xAB""wi""\xA9""ki",
    "Paul Mudra",
    "Frank Klepacki",
    "Dwight Okahara",
    "Pat Collins",
    "Poprawa jako""\x98""ci",
    "Glenn Sperry",
    "Michael Lightner",
    "William Foster",
    "Jesse Clemit",
    "Jeff Fillhaber",
    "Podr""\xA9""cznik, opakowanie",
    "i instrukcja.",
    "Eydie Laramore",
    "Lisa Marcinko",
    "Lauren Rifkin",
    "Gratulacje!",
    "Dzi""\xA9""kujemy za granie w Hand of Fate!",
    "Go""\x98""cinne kodowanie",
    "Wsp""\xA2""""\x88""praca",
    "Scott Duckett",
    "Testerzy",
    "Chris McFarland",
    "Paul Moore",
    "Chad Soares",
    "Jared Brinkley",
    "Jon Willliams",
    "Chris Toft",
    "T""\x88""umaczenie",
    "Jacek Zadr""\xA5""g, Marcin Kretowicz, Pawe""\x88"" Cybulski",
    "Za""\x88""aduj gr""\xA9""",
    "Wprowadzenie",
    "Nowa gra",
    "Zako""\xE4""cz gr""\xA9""",
    "Specjalne podzi""\xA9""kowania dla",
    "IPS Computer Group",
    "Tim Fritz",
    "Kenny Dunne"
};

static const StringListProvider k2SeqplayStringsDOSPolishProvider = { ARRAYSIZE(k2SeqplayStringsDOSPolish), k2SeqplayStringsDOSPolish };

