/* This file is automatically generated by create_project */
/* DO NOT EDIT MANUALLY */
// This file is being included by "base/plugins.cpp"
#if defined(ENABLE_SCUMM) || defined(DETECTION_FULL)
LINK_PLUGIN(SCUMM_DETECTION)
#endif
#if defined(ENABLE_ACCESS) || defined(DETECTION_FULL)
LINK_PLUGIN(ACCESS_DETECTION)
#endif
#if defined(ENABLE_ADL) || defined(DETECTION_FULL)
LINK_PLUGIN(ADL_DETECTION)
#endif
#if defined(ENABLE_AGI) || defined(DETECTION_FULL)
LINK_PLUGIN(AGI_DETECTION)
#endif
#if defined(ENABLE_AGOS) || defined(DETECTION_FULL)
LINK_PLUGIN(AGOS_DETECTION)
#endif
#if defined(ENABLE_AGS) || defined(DETECTION_FULL)
LINK_PLUGIN(AGS_DETECTION)
#endif
#if defined(ENABLE_ASYLUM) || defined(DETECTION_FULL)
LINK_PLUGIN(ASYLUM_DETECTION)
#endif
#if defined(ENABLE_AVALANCHE) || defined(DETECTION_FULL)
LINK_PLUGIN(AVALANCHE_DETECTION)
#endif
#if defined(ENABLE_BAGEL) || defined(DETECTION_FULL)
LINK_PLUGIN(BAGEL_DETECTION)
#endif
#if defined(ENABLE_BBVS) || defined(DETECTION_FULL)
LINK_PLUGIN(BBVS_DETECTION)
#endif
#if defined(ENABLE_BLADERUNNER) || defined(DETECTION_FULL)
LINK_PLUGIN(BLADERUNNER_DETECTION)
#endif
#if defined(ENABLE_BURIED) || defined(DETECTION_FULL)
LINK_PLUGIN(BURIED_DETECTION)
#endif
#if defined(ENABLE_CGE) || defined(DETECTION_FULL)
LINK_PLUGIN(CGE_DETECTION)
#endif
#if defined(ENABLE_CGE2) || defined(DETECTION_FULL)
LINK_PLUGIN(CGE2_DETECTION)
#endif
#if defined(ENABLE_CHAMBER) || defined(DETECTION_FULL)
LINK_PLUGIN(CHAMBER_DETECTION)
#endif
#if defined(ENABLE_CHEWY) || defined(DETECTION_FULL)
LINK_PLUGIN(CHEWY_DETECTION)
#endif
#if defined(ENABLE_CINE) || defined(DETECTION_FULL)
LINK_PLUGIN(CINE_DETECTION)
#endif
#if defined(ENABLE_COMPOSER) || defined(DETECTION_FULL)
LINK_PLUGIN(COMPOSER_DETECTION)
#endif
#if defined(ENABLE_CRAB) || defined(DETECTION_FULL)
LINK_PLUGIN(CRAB_DETECTION)
#endif
#if defined(ENABLE_CRUISE) || defined(DETECTION_FULL)
LINK_PLUGIN(CRUISE_DETECTION)
#endif
#if defined(ENABLE_CRYO) || defined(DETECTION_FULL)
LINK_PLUGIN(CRYO_DETECTION)
#endif
#if defined(ENABLE_CRYOMNI3D) || defined(DETECTION_FULL)
LINK_PLUGIN(CRYOMNI3D_DETECTION)
#endif
#if defined(ENABLE_DARKSEED) || defined(DETECTION_FULL)
LINK_PLUGIN(DARKSEED_DETECTION)
#endif
#if defined(ENABLE_DGDS) || defined(DETECTION_FULL)
LINK_PLUGIN(DGDS_DETECTION)
#endif
#if defined(ENABLE_DIRECTOR) || defined(DETECTION_FULL)
LINK_PLUGIN(DIRECTOR_DETECTION)
#endif
#if defined(ENABLE_DM) || defined(DETECTION_FULL)
LINK_PLUGIN(DM_DETECTION)
#endif
#if defined(ENABLE_DRACI) || defined(DETECTION_FULL)
LINK_PLUGIN(DRACI_DETECTION)
#endif
#if defined(ENABLE_DRAGONS) || defined(DETECTION_FULL)
LINK_PLUGIN(DRAGONS_DETECTION)
#endif
#if defined(ENABLE_DRASCULA) || defined(DETECTION_FULL)
LINK_PLUGIN(DRASCULA_DETECTION)
#endif
#if defined(ENABLE_DREAMWEB) || defined(DETECTION_FULL)
LINK_PLUGIN(DREAMWEB_DETECTION)
#endif
#if defined(ENABLE_EFH) || defined(DETECTION_FULL)
LINK_PLUGIN(EFH_DETECTION)
#endif
#if defined(ENABLE_FREESCAPE) || defined(DETECTION_FULL)
LINK_PLUGIN(FREESCAPE_DETECTION)
#endif
#if defined(ENABLE_GLK) || defined(DETECTION_FULL)
LINK_PLUGIN(GLK_DETECTION)
#endif
#if defined(ENABLE_GNAP) || defined(DETECTION_FULL)
LINK_PLUGIN(GNAP_DETECTION)
#endif
#if defined(ENABLE_GOB) || defined(DETECTION_FULL)
LINK_PLUGIN(GOB_DETECTION)
#endif
#if defined(ENABLE_GRIFFON) || defined(DETECTION_FULL)
LINK_PLUGIN(GRIFFON_DETECTION)
#endif
#if defined(ENABLE_GRIM) || defined(DETECTION_FULL)
LINK_PLUGIN(GRIM_DETECTION)
#endif
#if defined(ENABLE_GROOVIE) || defined(DETECTION_FULL)
LINK_PLUGIN(GROOVIE_DETECTION)
#endif
#if defined(ENABLE_HADESCH) || defined(DETECTION_FULL)
LINK_PLUGIN(HADESCH_DETECTION)
#endif
#if defined(ENABLE_HDB) || defined(DETECTION_FULL)
LINK_PLUGIN(HDB_DETECTION)
#endif
#if defined(ENABLE_HOPKINS) || defined(DETECTION_FULL)
LINK_PLUGIN(HOPKINS_DETECTION)
#endif
#if defined(ENABLE_HPL1) || defined(DETECTION_FULL)
LINK_PLUGIN(HPL1_DETECTION)
#endif
#if defined(ENABLE_HUGO) || defined(DETECTION_FULL)
LINK_PLUGIN(HUGO_DETECTION)
#endif
#if defined(ENABLE_HYPNO) || defined(DETECTION_FULL)
LINK_PLUGIN(HYPNO_DETECTION)
#endif
#if defined(ENABLE_ICB) || defined(DETECTION_FULL)
LINK_PLUGIN(ICB_DETECTION)
#endif
#if defined(ENABLE_ILLUSIONS) || defined(DETECTION_FULL)
LINK_PLUGIN(ILLUSIONS_DETECTION)
#endif
#if defined(ENABLE_IMMORTAL) || defined(DETECTION_FULL)
LINK_PLUGIN(IMMORTAL_DETECTION)
#endif
#if defined(ENABLE_KINGDOM) || defined(DETECTION_FULL)
LINK_PLUGIN(KINGDOM_DETECTION)
#endif
#if defined(ENABLE_KYRA) || defined(DETECTION_FULL)
LINK_PLUGIN(KYRA_DETECTION)
#endif
#if defined(ENABLE_LAB) || defined(DETECTION_FULL)
LINK_PLUGIN(LAB_DETECTION)
#endif
#if defined(ENABLE_LASTEXPRESS) || defined(DETECTION_FULL)
LINK_PLUGIN(LASTEXPRESS_DETECTION)
#endif
#if defined(ENABLE_LILLIPUT) || defined(DETECTION_FULL)
LINK_PLUGIN(LILLIPUT_DETECTION)
#endif
#if defined(ENABLE_LURE) || defined(DETECTION_FULL)
LINK_PLUGIN(LURE_DETECTION)
#endif
#if defined(ENABLE_M4) || defined(DETECTION_FULL)
LINK_PLUGIN(M4_DETECTION)
#endif
#if defined(ENABLE_MACVENTURE) || defined(DETECTION_FULL)
LINK_PLUGIN(MACVENTURE_DETECTION)
#endif
#if defined(ENABLE_MADE) || defined(DETECTION_FULL)
LINK_PLUGIN(MADE_DETECTION)
#endif
#if defined(ENABLE_MADS) || defined(DETECTION_FULL)
LINK_PLUGIN(MADS_DETECTION)
#endif
#if defined(ENABLE_MM) || defined(DETECTION_FULL)
LINK_PLUGIN(MM_DETECTION)
#endif
#if defined(ENABLE_MOHAWK) || defined(DETECTION_FULL)
LINK_PLUGIN(MOHAWK_DETECTION)
#endif
#if defined(ENABLE_MORTEVIELLE) || defined(DETECTION_FULL)
LINK_PLUGIN(MORTEVIELLE_DETECTION)
#endif
#if defined(ENABLE_MTROPOLIS) || defined(DETECTION_FULL)
LINK_PLUGIN(MTROPOLIS_DETECTION)
#endif
#if defined(ENABLE_MUTATIONOFJB) || defined(DETECTION_FULL)
LINK_PLUGIN(MUTATIONOFJB_DETECTION)
#endif
#if defined(ENABLE_MYST3) || defined(DETECTION_FULL)
LINK_PLUGIN(MYST3_DETECTION)
#endif
#if defined(ENABLE_NANCY) || defined(DETECTION_FULL)
LINK_PLUGIN(NANCY_DETECTION)
#endif
#if defined(ENABLE_NEVERHOOD) || defined(DETECTION_FULL)
LINK_PLUGIN(NEVERHOOD_DETECTION)
#endif
#if defined(ENABLE_NGI) || defined(DETECTION_FULL)
LINK_PLUGIN(NGI_DETECTION)
#endif
#if defined(ENABLE_PARALLACTION) || defined(DETECTION_FULL)
LINK_PLUGIN(PARALLACTION_DETECTION)
#endif
#if defined(ENABLE_PEGASUS) || defined(DETECTION_FULL)
LINK_PLUGIN(PEGASUS_DETECTION)
#endif
#if defined(ENABLE_PETKA) || defined(DETECTION_FULL)
LINK_PLUGIN(PETKA_DETECTION)
#endif
#if defined(ENABLE_PINK) || defined(DETECTION_FULL)
LINK_PLUGIN(PINK_DETECTION)
#endif
#if defined(ENABLE_PLAYGROUND3D) || defined(DETECTION_FULL)
LINK_PLUGIN(PLAYGROUND3D_DETECTION)
#endif
#if defined(ENABLE_PLUMBERS) || defined(DETECTION_FULL)
LINK_PLUGIN(PLUMBERS_DETECTION)
#endif
#if defined(ENABLE_PRINCE) || defined(DETECTION_FULL)
LINK_PLUGIN(PRINCE_DETECTION)
#endif
#if defined(ENABLE_PRIVATE) || defined(DETECTION_FULL)
LINK_PLUGIN(PRIVATE_DETECTION)
#endif
#if defined(ENABLE_QDENGINE) || defined(DETECTION_FULL)
LINK_PLUGIN(QDENGINE_DETECTION)
#endif
#if defined(ENABLE_QUEEN) || defined(DETECTION_FULL)
LINK_PLUGIN(QUEEN_DETECTION)
#endif
#if defined(ENABLE_SAGA) || defined(DETECTION_FULL)
LINK_PLUGIN(SAGA_DETECTION)
#endif
#if defined(ENABLE_SAGA2) || defined(DETECTION_FULL)
LINK_PLUGIN(SAGA2_DETECTION)
#endif
#if defined(ENABLE_SCI) || defined(DETECTION_FULL)
LINK_PLUGIN(SCI_DETECTION)
#endif
#if defined(ENABLE_SHERLOCK) || defined(DETECTION_FULL)
LINK_PLUGIN(SHERLOCK_DETECTION)
#endif
#if defined(ENABLE_SKY) || defined(DETECTION_FULL)
LINK_PLUGIN(SKY_DETECTION)
#endif
#if defined(ENABLE_SLUDGE) || defined(DETECTION_FULL)
LINK_PLUGIN(SLUDGE_DETECTION)
#endif
#if defined(ENABLE_STARK) || defined(DETECTION_FULL)
LINK_PLUGIN(STARK_DETECTION)
#endif
#if defined(ENABLE_STARTREK) || defined(DETECTION_FULL)
LINK_PLUGIN(STARTREK_DETECTION)
#endif
#if defined(ENABLE_SUPERNOVA) || defined(DETECTION_FULL)
LINK_PLUGIN(SUPERNOVA_DETECTION)
#endif
#if defined(ENABLE_SWORD1) || defined(DETECTION_FULL)
LINK_PLUGIN(SWORD1_DETECTION)
#endif
#if defined(ENABLE_SWORD2) || defined(DETECTION_FULL)
LINK_PLUGIN(SWORD2_DETECTION)
#endif
#if defined(ENABLE_SWORD25) || defined(DETECTION_FULL)
LINK_PLUGIN(SWORD25_DETECTION)
#endif
#if defined(ENABLE_TANITA2) || defined(DETECTION_FULL)
LINK_PLUGIN(TANITA2_DETECTION)
#endif
#if defined(ENABLE_TEENAGENT) || defined(DETECTION_FULL)
LINK_PLUGIN(TEENAGENT_DETECTION)
#endif
#if defined(ENABLE_TESTBED) || defined(DETECTION_FULL)
LINK_PLUGIN(TESTBED_DETECTION)
#endif
#if defined(ENABLE_TETRAEDGE) || defined(DETECTION_FULL)
LINK_PLUGIN(TETRAEDGE_DETECTION)
#endif
#if defined(ENABLE_TINSEL) || defined(DETECTION_FULL)
LINK_PLUGIN(TINSEL_DETECTION)
#endif
#if defined(ENABLE_TITANIC) || defined(DETECTION_FULL)
LINK_PLUGIN(TITANIC_DETECTION)
#endif
#if defined(ENABLE_TOLTECS) || defined(DETECTION_FULL)
LINK_PLUGIN(TOLTECS_DETECTION)
#endif
#if defined(ENABLE_TONY) || defined(DETECTION_FULL)
LINK_PLUGIN(TONY_DETECTION)
#endif
#if defined(ENABLE_TOON) || defined(DETECTION_FULL)
LINK_PLUGIN(TOON_DETECTION)
#endif
#if defined(ENABLE_TOT) || defined(DETECTION_FULL)
LINK_PLUGIN(TOT_DETECTION)
#endif
#if defined(ENABLE_TOUCHE) || defined(DETECTION_FULL)
LINK_PLUGIN(TOUCHE_DETECTION)
#endif
#if defined(ENABLE_TRECISION) || defined(DETECTION_FULL)
LINK_PLUGIN(TRECISION_DETECTION)
#endif
#if defined(ENABLE_TSAGE) || defined(DETECTION_FULL)
LINK_PLUGIN(TSAGE_DETECTION)
#endif
#if defined(ENABLE_TUCKER) || defined(DETECTION_FULL)
LINK_PLUGIN(TUCKER_DETECTION)
#endif
#if defined(ENABLE_TWINE) || defined(DETECTION_FULL)
LINK_PLUGIN(TWINE_DETECTION)
#endif
#if defined(ENABLE_TWP) || defined(DETECTION_FULL)
LINK_PLUGIN(TWP_DETECTION)
#endif
#if defined(ENABLE_ULTIMA) || defined(DETECTION_FULL)
LINK_PLUGIN(ULTIMA_DETECTION)
#endif
#if defined(ENABLE_VCRUISE) || defined(DETECTION_FULL)
LINK_PLUGIN(VCRUISE_DETECTION)
#endif
#if defined(ENABLE_VOYEUR) || defined(DETECTION_FULL)
LINK_PLUGIN(VOYEUR_DETECTION)
#endif
#if defined(ENABLE_WAGE) || defined(DETECTION_FULL)
LINK_PLUGIN(WAGE_DETECTION)
#endif
#if defined(ENABLE_WATCHMAKER) || defined(DETECTION_FULL)
LINK_PLUGIN(WATCHMAKER_DETECTION)
#endif
#if defined(ENABLE_WINTERMUTE) || defined(DETECTION_FULL)
LINK_PLUGIN(WINTERMUTE_DETECTION)
#endif
#if defined(ENABLE_ZVISION) || defined(DETECTION_FULL)
LINK_PLUGIN(ZVISION_DETECTION)
#endif
