begin_section("Asylum");
	add_person("Alex Bevilacqua", "alexbevi", "");
	add_person("Alex Fontoura", "xesf", "");
	add_person("Alexander Panov", "alxpnv", "");
	add_person("Benjamin Haisch", "john_doe", "");
	add_person("Filippos Karapetis", "bluegr", "");
	add_person("Joseph Davies", "Deledrius", "");
	add_person("Julien Templier", "littleboy", "");
end_section();
