/* This file is automatically generated by create_project */
/* DO NOT EDIT MANUALLY */
// This file is being included by "base/plugins.cpp"
#if PLUGIN_ENABLED_STATIC(SCUMM)
LINK_PLUGIN(SCUMM)
#endif
#if PLUGIN_ENABLED_STATIC(ACCESS)
LINK_PLUGIN(ACCESS)
#endif
#if PLUGIN_ENABLED_STATIC(ADL)
LINK_PLUGIN(ADL)
#endif
#if PLUGIN_ENABLED_STATIC(AGI)
LINK_PLUGIN(AGI)
#endif
#if PLUGIN_ENABLED_STATIC(AGOS)
LINK_PLUGIN(AGOS)
#endif
#if PLUGIN_ENABLED_STATIC(AGS)
LINK_PLUGIN(AGS)
#endif
#if PLUGIN_ENABLED_STATIC(ASYLUM)
LINK_PLUGIN(ASYLUM)
#endif
#if PLUGIN_ENABLED_STATIC(AVALANCHE)
LINK_PLUGIN(AVALANCHE)
#endif
#if PLUGIN_ENABLED_STATIC(BAGEL)
LINK_PLUGIN(BAGEL)
#endif
#if PLUGIN_ENABLED_STATIC(BBVS)
LINK_PLUGIN(BBVS)
#endif
#if PLUGIN_ENABLED_STATIC(BLADERUNNER)
LINK_PLUGIN(BLADERUNNER)
#endif
#if PLUGIN_ENABLED_STATIC(BURIED)
LINK_PLUGIN(BURIED)
#endif
#if PLUGIN_ENABLED_STATIC(CGE)
LINK_PLUGIN(CGE)
#endif
#if PLUGIN_ENABLED_STATIC(CGE2)
LINK_PLUGIN(CGE2)
#endif
#if PLUGIN_ENABLED_STATIC(CHAMBER)
LINK_PLUGIN(CHAMBER)
#endif
#if PLUGIN_ENABLED_STATIC(CHEWY)
LINK_PLUGIN(CHEWY)
#endif
#if PLUGIN_ENABLED_STATIC(CINE)
LINK_PLUGIN(CINE)
#endif
#if PLUGIN_ENABLED_STATIC(COMPOSER)
LINK_PLUGIN(COMPOSER)
#endif
#if PLUGIN_ENABLED_STATIC(CRAB)
LINK_PLUGIN(CRAB)
#endif
#if PLUGIN_ENABLED_STATIC(CRUISE)
LINK_PLUGIN(CRUISE)
#endif
#if PLUGIN_ENABLED_STATIC(CRYO)
LINK_PLUGIN(CRYO)
#endif
#if PLUGIN_ENABLED_STATIC(CRYOMNI3D)
LINK_PLUGIN(CRYOMNI3D)
#endif
#if PLUGIN_ENABLED_STATIC(DARKSEED)
LINK_PLUGIN(DARKSEED)
#endif
#if PLUGIN_ENABLED_STATIC(DGDS)
LINK_PLUGIN(DGDS)
#endif
#if PLUGIN_ENABLED_STATIC(DIRECTOR)
LINK_PLUGIN(DIRECTOR)
#endif
#if PLUGIN_ENABLED_STATIC(DM)
LINK_PLUGIN(DM)
#endif
#if PLUGIN_ENABLED_STATIC(DRACI)
LINK_PLUGIN(DRACI)
#endif
#if PLUGIN_ENABLED_STATIC(DRAGONS)
LINK_PLUGIN(DRAGONS)
#endif
#if PLUGIN_ENABLED_STATIC(DRASCULA)
LINK_PLUGIN(DRASCULA)
#endif
#if PLUGIN_ENABLED_STATIC(DREAMWEB)
LINK_PLUGIN(DREAMWEB)
#endif
#if PLUGIN_ENABLED_STATIC(EFH)
LINK_PLUGIN(EFH)
#endif
#if PLUGIN_ENABLED_STATIC(FREESCAPE)
LINK_PLUGIN(FREESCAPE)
#endif
#if PLUGIN_ENABLED_STATIC(GLK)
LINK_PLUGIN(GLK)
#endif
#if PLUGIN_ENABLED_STATIC(GNAP)
LINK_PLUGIN(GNAP)
#endif
#if PLUGIN_ENABLED_STATIC(GOB)
LINK_PLUGIN(GOB)
#endif
#if PLUGIN_ENABLED_STATIC(GRIFFON)
LINK_PLUGIN(GRIFFON)
#endif
#if PLUGIN_ENABLED_STATIC(GRIM)
LINK_PLUGIN(GRIM)
#endif
#if PLUGIN_ENABLED_STATIC(GROOVIE)
LINK_PLUGIN(GROOVIE)
#endif
#if PLUGIN_ENABLED_STATIC(HADESCH)
LINK_PLUGIN(HADESCH)
#endif
#if PLUGIN_ENABLED_STATIC(HDB)
LINK_PLUGIN(HDB)
#endif
#if PLUGIN_ENABLED_STATIC(HOPKINS)
LINK_PLUGIN(HOPKINS)
#endif
#if PLUGIN_ENABLED_STATIC(HPL1)
LINK_PLUGIN(HPL1)
#endif
#if PLUGIN_ENABLED_STATIC(HUGO)
LINK_PLUGIN(HUGO)
#endif
#if PLUGIN_ENABLED_STATIC(HYPNO)
LINK_PLUGIN(HYPNO)
#endif
#if PLUGIN_ENABLED_STATIC(ICB)
LINK_PLUGIN(ICB)
#endif
#if PLUGIN_ENABLED_STATIC(ILLUSIONS)
LINK_PLUGIN(ILLUSIONS)
#endif
#if PLUGIN_ENABLED_STATIC(IMMORTAL)
LINK_PLUGIN(IMMORTAL)
#endif
#if PLUGIN_ENABLED_STATIC(KINGDOM)
LINK_PLUGIN(KINGDOM)
#endif
#if PLUGIN_ENABLED_STATIC(KYRA)
LINK_PLUGIN(KYRA)
#endif
#if PLUGIN_ENABLED_STATIC(LAB)
LINK_PLUGIN(LAB)
#endif
#if PLUGIN_ENABLED_STATIC(LASTEXPRESS)
LINK_PLUGIN(LASTEXPRESS)
#endif
#if PLUGIN_ENABLED_STATIC(LILLIPUT)
LINK_PLUGIN(LILLIPUT)
#endif
#if PLUGIN_ENABLED_STATIC(LURE)
LINK_PLUGIN(LURE)
#endif
#if PLUGIN_ENABLED_STATIC(M4)
LINK_PLUGIN(M4)
#endif
#if PLUGIN_ENABLED_STATIC(MACVENTURE)
LINK_PLUGIN(MACVENTURE)
#endif
#if PLUGIN_ENABLED_STATIC(MADE)
LINK_PLUGIN(MADE)
#endif
#if PLUGIN_ENABLED_STATIC(MADS)
LINK_PLUGIN(MADS)
#endif
#if PLUGIN_ENABLED_STATIC(MM)
LINK_PLUGIN(MM)
#endif
#if PLUGIN_ENABLED_STATIC(MOHAWK)
LINK_PLUGIN(MOHAWK)
#endif
#if PLUGIN_ENABLED_STATIC(MORTEVIELLE)
LINK_PLUGIN(MORTEVIELLE)
#endif
#if PLUGIN_ENABLED_STATIC(MTROPOLIS)
LINK_PLUGIN(MTROPOLIS)
#endif
#if PLUGIN_ENABLED_STATIC(MUTATIONOFJB)
LINK_PLUGIN(MUTATIONOFJB)
#endif
#if PLUGIN_ENABLED_STATIC(MYST3)
LINK_PLUGIN(MYST3)
#endif
#if PLUGIN_ENABLED_STATIC(NANCY)
LINK_PLUGIN(NANCY)
#endif
#if PLUGIN_ENABLED_STATIC(NEVERHOOD)
LINK_PLUGIN(NEVERHOOD)
#endif
#if PLUGIN_ENABLED_STATIC(NGI)
LINK_PLUGIN(NGI)
#endif
#if PLUGIN_ENABLED_STATIC(PARALLACTION)
LINK_PLUGIN(PARALLACTION)
#endif
#if PLUGIN_ENABLED_STATIC(PEGASUS)
LINK_PLUGIN(PEGASUS)
#endif
#if PLUGIN_ENABLED_STATIC(PETKA)
LINK_PLUGIN(PETKA)
#endif
#if PLUGIN_ENABLED_STATIC(PINK)
LINK_PLUGIN(PINK)
#endif
#if PLUGIN_ENABLED_STATIC(PLAYGROUND3D)
LINK_PLUGIN(PLAYGROUND3D)
#endif
#if PLUGIN_ENABLED_STATIC(PLUMBERS)
LINK_PLUGIN(PLUMBERS)
#endif
#if PLUGIN_ENABLED_STATIC(PRINCE)
LINK_PLUGIN(PRINCE)
#endif
#if PLUGIN_ENABLED_STATIC(PRIVATE)
LINK_PLUGIN(PRIVATE)
#endif
#if PLUGIN_ENABLED_STATIC(QDENGINE)
LINK_PLUGIN(QDENGINE)
#endif
#if PLUGIN_ENABLED_STATIC(QUEEN)
LINK_PLUGIN(QUEEN)
#endif
#if PLUGIN_ENABLED_STATIC(SAGA)
LINK_PLUGIN(SAGA)
#endif
#if PLUGIN_ENABLED_STATIC(SAGA2)
LINK_PLUGIN(SAGA2)
#endif
#if PLUGIN_ENABLED_STATIC(SCI)
LINK_PLUGIN(SCI)
#endif
#if PLUGIN_ENABLED_STATIC(SHERLOCK)
LINK_PLUGIN(SHERLOCK)
#endif
#if PLUGIN_ENABLED_STATIC(SKY)
LINK_PLUGIN(SKY)
#endif
#if PLUGIN_ENABLED_STATIC(SLUDGE)
LINK_PLUGIN(SLUDGE)
#endif
#if PLUGIN_ENABLED_STATIC(STARK)
LINK_PLUGIN(STARK)
#endif
#if PLUGIN_ENABLED_STATIC(STARTREK)
LINK_PLUGIN(STARTREK)
#endif
#if PLUGIN_ENABLED_STATIC(SUPERNOVA)
LINK_PLUGIN(SUPERNOVA)
#endif
#if PLUGIN_ENABLED_STATIC(SWORD1)
LINK_PLUGIN(SWORD1)
#endif
#if PLUGIN_ENABLED_STATIC(SWORD2)
LINK_PLUGIN(SWORD2)
#endif
#if PLUGIN_ENABLED_STATIC(SWORD25)
LINK_PLUGIN(SWORD25)
#endif
#if PLUGIN_ENABLED_STATIC(TANITA2)
LINK_PLUGIN(TANITA2)
#endif
#if PLUGIN_ENABLED_STATIC(TEENAGENT)
LINK_PLUGIN(TEENAGENT)
#endif
#if PLUGIN_ENABLED_STATIC(TESTBED)
LINK_PLUGIN(TESTBED)
#endif
#if PLUGIN_ENABLED_STATIC(TETRAEDGE)
LINK_PLUGIN(TETRAEDGE)
#endif
#if PLUGIN_ENABLED_STATIC(TINSEL)
LINK_PLUGIN(TINSEL)
#endif
#if PLUGIN_ENABLED_STATIC(TITANIC)
LINK_PLUGIN(TITANIC)
#endif
#if PLUGIN_ENABLED_STATIC(TOLTECS)
LINK_PLUGIN(TOLTECS)
#endif
#if PLUGIN_ENABLED_STATIC(TONY)
LINK_PLUGIN(TONY)
#endif
#if PLUGIN_ENABLED_STATIC(TOON)
LINK_PLUGIN(TOON)
#endif
#if PLUGIN_ENABLED_STATIC(TOT)
LINK_PLUGIN(TOT)
#endif
#if PLUGIN_ENABLED_STATIC(TOUCHE)
LINK_PLUGIN(TOUCHE)
#endif
#if PLUGIN_ENABLED_STATIC(TRECISION)
LINK_PLUGIN(TRECISION)
#endif
#if PLUGIN_ENABLED_STATIC(TSAGE)
LINK_PLUGIN(TSAGE)
#endif
#if PLUGIN_ENABLED_STATIC(TUCKER)
LINK_PLUGIN(TUCKER)
#endif
#if PLUGIN_ENABLED_STATIC(TWINE)
LINK_PLUGIN(TWINE)
#endif
#if PLUGIN_ENABLED_STATIC(TWP)
LINK_PLUGIN(TWP)
#endif
#if PLUGIN_ENABLED_STATIC(ULTIMA)
LINK_PLUGIN(ULTIMA)
#endif
#if PLUGIN_ENABLED_STATIC(VCRUISE)
LINK_PLUGIN(VCRUISE)
#endif
#if PLUGIN_ENABLED_STATIC(VOYEUR)
LINK_PLUGIN(VOYEUR)
#endif
#if PLUGIN_ENABLED_STATIC(WAGE)
LINK_PLUGIN(WAGE)
#endif
#if PLUGIN_ENABLED_STATIC(WATCHMAKER)
LINK_PLUGIN(WATCHMAKER)
#endif
#if PLUGIN_ENABLED_STATIC(WINTERMUTE)
LINK_PLUGIN(WINTERMUTE)
#endif
#if PLUGIN_ENABLED_STATIC(ZVISION)
LINK_PLUGIN(ZVISION)
#endif
