static const char *const k2SeqplayStringsDOSEnglish[103] = {
	"Kyrandia is disappearing!",
	"Rock by rock,",
	"And tree by tree.",
	"Kyrandia ceases to exist!",
	"The Royal Mystics are baffled.",
	"Every reference has been consulted.",
	"Even Marko and his new valet, have been allowed into the conference.",
	"Luckily, the Hand was experienced in these matters.",
	"And finally a plan was approved,",
	"That required a magic Anchor Stone,",
	"To be retrieved from the center of the world.",
	"Zanthia, youngest of the Kyrandian Mystics, has been selected to retrieve the Stone.",
	"Thank you for playing The Hand of Fate.",
	"This should be enough blueberries to open a portal to the center of the world.",
	" DUMMY STRING... ",
	" DUMMY STRING... ",
	"Hey! All my equipment has been stolen!",
	" DUMMY STRING... ",
	"If they think I'm going to walk all the way down there, they're nuts!",
	" DUMMY STRING... ",
	" DUMMY STRING... ",
	"Hurry up Faun!",
	"Boy, that was a close call!",
	"You said it pal. I, for one, am never going hunting again!",
	"Ribbit.",
	"How many times do I have to tell you? You're a toad.",
	"Oh no! We're out of cheese!",
	"Let's try this earwax. It's orange.",
	"Mommy, when do I get the ivy?",
	"Get out of here, shoo!",
	"You cut, and I'll choose.",
	"No. You cut and I'll choose.",
	"I still say it was derivative drivel.",
	"Aw, you still wouldn't recognize iambic pentameter if it bit you on the butt!",
	"Executive Producer",
	"Brett W. Sperry",
	"Designed & Directed by",
	"Rick Gush",
	"Lead Programmer",
	"Michael Legg",
	"Art Management",
	"Louis Castle",
	"Joseph B. Hewitt IV",
	"Lead Artist",
	"Rick Parks",
	"Additional Coding by",
	"Philip W. Gorrow",
	"Matt Collins",
	"Mark McCubbin",
	"Artists",
	"Cameron Chun",
	"Cary Averett",
	"Cindy Chinn",
	"Elie Arabian",
	"Fei Cheng",
	"Ferby Miguel",
	"Frank Mendeola",
	"Jack Martin",
	"Jerry Moore",
	"",
	"Judith Peterson",
	"Larry Miller",
	"Lenny Lee",
	"Louise Sandoval",
	"Ren Olsen",
	"Music & Sounds by",
	"Paul Mudra",
	"Frank Klepacki",
	"Dwight Okahara",
	"Pat Collins",
	"Qualilty Assurance by",
	"Glenn Sperry",
	"Michael Lightner",
	"William Foster",
	"Jesse Clemit",
	"Jeff Fillhaber",
	"Manual, Package Design",
	"& Fulfillment",
	"Eydie Laramore",
	"Lisa Marcinko",
	"Lauren Rifkin",
	"Congratulations!",
	"Thank you for playing The Hand of Fate!",
	"Guest Coding",
	"Producer Liaison",
	"Scott Duckett",
	"Irvine Testers",
	"Chris McFarland",
	"Paul Moore",
	"Chad Soares",
	"Jared Brinkley",
	"Jon Willliams",
	"Chris Toft",
	"Chris's Hair by",
	"Cumulo Nimbus",
	"Load a game",
	"Introduction",
	"Start a new game",
	"Exit the game",
	"Special Thanks to",
	"Sake Joe Bostic-san",
	"Tim Fritz",
	"Kenny Dunne"
};

static const StringListProvider k2SeqplayStringsDOSEnglishProvider = { ARRAYSIZE(k2SeqplayStringsDOSEnglish), k2SeqplayStringsDOSEnglish };

