begin_section("Groovie");
	add_person("Henry Bush", "spookypeanut", "");
	add_person("Ray Carro", "Die4Ever", "");
	add_person("Scott Thomas", "ST", "");
	add_person("Jordi Vilalta Prat", "jvprat", "");
end_section();
