static const char *const k1IntroStringsDOSEnglish[51] = {
	"This is a text test - 1",
	"This is a text test - 2",
	"Greetings, Kallak...",
	"...Powerful leader of the Royal Mystics!",
	"Did I frighten you?",
	"Malcolm!",
	"I heard yesterday of your escape.",
	"And have expected you before now.",
	"Why would I rush?",
	"I rule this land!",
	"Your puny curse restrains me naught.",
	"So, slay me now!",
	"I have little magic left.",
	"I mean to harm you, have no doubt...",
	"...Now you are stone...",
	"...Yet I leave your eyes.",
	"I shed no tears for Kyrandia...",
	"...But cannot deny you yours.",
	"Malcolm the Jester has broken free!",
	"He now controls the Kyragem...",
	"...source of all magic in Kyrandia!",
	"",
	"How dare you trip ME!",
	"But no, something special for you...",
	"A humorous hex!",
	"But I shall give fair warning.",
	"Don't jump on THAT tree!",
	"Don't climb up THAT tree little squirrel!",
	"Boo!",
	"Tis funnier that way...",
	"...Is it not?",
	"Ha ha ha...",
	"Aren't fingers wonderful?",
	"It's great to be alive!",
	"Brandon!",
	"Grandfather! We did it!",
	"No! YOU did it!",
	"Long live King Brandon!",
	"Long live the Royal Mystics!",
	"Now, let's go put Kyrandia back together!",
	"Good idea!",
	"As my first Royal Proclamation...",
	"I name sandals as the official footwear of Kyrandia!",
	"Well done, Brandon!",
	"The Land no longer weeps...",
	"And your friends have been restored!",
	"Dear Brynn...",
	"Malcolm has broken free!",
	"Soon he will come for me.",
	"Please help Brandon...",
	""
};

static const StringListProvider k1IntroStringsDOSEnglishProvider = { ARRAYSIZE(k1IntroStringsDOSEnglish), k1IntroStringsDOSEnglish };

static const char *const k1ItemNamesDOSEnglish[107] = {
	"Garnet",
	"Amethyst",
	"Aquamarine",
	"Diamond",
	"Emerald",
	"Pearl",
	"Ruby",
	"Peridot",
	"Sapphire",
	"Opal",
	"Topaz",
	"Onyx",
	"Sunstone",
	"Moonstone",
	"Rainbowstone",
	"Lodestone",
	"Rose",
	"Tulip",
	"Orchid",
	"Silver Rose",
	"Silver Statuette",
	"Silver Coin",
	"Gold Coin",
	"Gold Ring",
	"Royal Chalice",
	"Pinecone",
	"Acorn",
	"Walnut",
	"Everglowing Fireberries",
	"Fireberries",
	"Fireberries",
	"Fireberries",
	"Fireberries",
	"Fireberries",
	"Fish",
	"Fish Bone",
	"Mutton Leg",
	"Bone",
	"Apple",
	"Apple Core",
	"Blueberries",
	"Mushroom",
	"Note",
	"Marble",
	"Saw",
	"Ankh",
	"Feather",
	"Egg",
	"Leaf",
	"Shamrock",
	"Fallen Star",
	"Crystal Ball",
	"Teardrop",
	"Mirror",
	"Ice Shard",
	"Flute",
	"Hourglass",
	"Iron Key",
	"Jade Key",
	"Obsidion Key",
	"Red Potion",
	"Red Potion",
	"Blue Potion",
	"Blue Potion",
	"Yellow Potion",
	"Yellow Potion",
	"Green Potion",
	"Orange Potion",
	"Purple Potion",
	"Potion of Rainbows",
	"Fresh Water",
	"Fresh Water",
	"Salt Water",
	"Salt Water",
	"Mineral Water",
	"Mineral Water",
	"Magical Water",
	"Magical Water",
	"Empty Flask",
	"Empty Flask",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Scroll",
	"Parchment scrap",
	"Parchment scrap",
	"Parchment scrap",
	"Parchment scrap",
	"Parchment scrap",
	"Red Magestone",
	"Orange Magestone",
	"Yellow Magestone",
	"Green Magestone",
	"Blue-Green Magestone",
	"Blue Magestone",
	"Purple Magestone",
	"Heavy rock",
	"Royal Crown",
	"Royal Sceptre",
	"Gold key",
	"Unknown item"
};

static const StringListProvider k1ItemNamesDOSEnglishProvider = { ARRAYSIZE(k1ItemNamesDOSEnglish), k1ItemNamesDOSEnglish };

static const char *const k1TakenStringsDOSEnglish[2] = {
	" taken.",
	" taken."
};

static const StringListProvider k1TakenStringsDOSEnglishProvider = { ARRAYSIZE(k1TakenStringsDOSEnglish), k1TakenStringsDOSEnglish };

static const char *const k1PlacedStringsDOSEnglish[1] = {
	" placed."
};

static const StringListProvider k1PlacedStringsDOSEnglishProvider = { ARRAYSIZE(k1PlacedStringsDOSEnglish), k1PlacedStringsDOSEnglish };

static const char *const k1DroppedStringsDOSEnglish[1] = {
	" dropped."
};

static const StringListProvider k1DroppedStringsDOSEnglishProvider = { ARRAYSIZE(k1DroppedStringsDOSEnglish), k1DroppedStringsDOSEnglish };

static const char *const k1NoDropStringsDOSEnglish[2] = {
	"You can't drop any more items in this scene.",
	"You can't put that there."
};

static const StringListProvider k1NoDropStringsDOSEnglishProvider = { ARRAYSIZE(k1NoDropStringsDOSEnglish), k1NoDropStringsDOSEnglish };

static const char *const k1PutDownStringDOSEnglish[1] = {
	"Maybe I'd better put this down first."
};

static const StringListProvider k1PutDownStringDOSEnglishProvider = { ARRAYSIZE(k1PutDownStringDOSEnglish), k1PutDownStringDOSEnglish };

static const char *const k1WaitAmuletStringDOSEnglish[1] = {
	"I guess I'll have to wait for my Amulet to regain its power."
};

static const StringListProvider k1WaitAmuletStringDOSEnglishProvider = { ARRAYSIZE(k1WaitAmuletStringDOSEnglish), k1WaitAmuletStringDOSEnglish };

static const char *const k1BlackJewelStringDOSEnglish[1] = {
	"It's a jewel, but why is it black?"
};

static const StringListProvider k1BlackJewelStringDOSEnglishProvider = { ARRAYSIZE(k1BlackJewelStringDOSEnglish), k1BlackJewelStringDOSEnglish };

static const char *const k1HealingTipStringDOSEnglish[1] = {
	"Gee, this could come in handy if I get hurt."
};

static const StringListProvider k1HealingTipStringDOSEnglishProvider = { ARRAYSIZE(k1HealingTipStringDOSEnglish), k1HealingTipStringDOSEnglish };

static const char *const k1PoisonGoneStringDOSEnglish[2] = {
	"Amazing!",
	"The effects of the poison are gone!"
};

static const StringListProvider k1PoisonGoneStringDOSEnglishProvider = { ARRAYSIZE(k1PoisonGoneStringDOSEnglish), k1PoisonGoneStringDOSEnglish };

static const char *const k1ThePoisonStringsDOSEnglish[4] = {
	"The poison...",
	"I can't breathe...",
	"I don't feel so good...",
	"That snake must\rbe poisonous!"
};

static const StringListProvider k1ThePoisonStringsDOSEnglishProvider = { ARRAYSIZE(k1ThePoisonStringsDOSEnglish), k1ThePoisonStringsDOSEnglish };

static const char *const k1FluteStringsDOSEnglish[2] = {
	"Not very impressive sounding.",
	"That last note sure was high!"
};

static const StringListProvider k1FluteStringsDOSEnglishProvider = { ARRAYSIZE(k1FluteStringsDOSEnglish), k1FluteStringsDOSEnglish };

static const char *const k1WispJewelStringsDOSEnglish[3] = {
	"I'm not really feeling up to it right now.",
	"I probably should put down my ",
	", first."
};

static const StringListProvider k1WispJewelStringsDOSEnglishProvider = { ARRAYSIZE(k1WispJewelStringsDOSEnglish), k1WispJewelStringsDOSEnglish };

static const char *const k1MagicJewelStringsDOSEnglish[1] = {
	"That felt strange."
};

static const StringListProvider k1MagicJewelStringsDOSEnglishProvider = { ARRAYSIZE(k1MagicJewelStringsDOSEnglish), k1MagicJewelStringsDOSEnglish };

static const char *const k1FlaskFullStringDOSEnglish[1] = {
	"This flask is already full."
};

static const StringListProvider k1FlaskFullStringDOSEnglishProvider = { ARRAYSIZE(k1FlaskFullStringDOSEnglish), k1FlaskFullStringDOSEnglish };

static const char *const k1FullFlaskStringDOSEnglish[4] = {
	"The flask is now filled\rwith fresh, sparkling water.",
	"The flask is now filled\rwith salt water.",
	"The flask is now filled\rwith mineral water.",
	"Magic water."
};

static const StringListProvider k1FullFlaskStringDOSEnglishProvider = { ARRAYSIZE(k1FullFlaskStringDOSEnglish), k1FullFlaskStringDOSEnglish };

static const char *const k1OutroHomeStringDOSEnglish[1] = {
	"Home"
};

static const StringListProvider k1OutroHomeStringDOSEnglishProvider = { ARRAYSIZE(k1OutroHomeStringDOSEnglish), k1OutroHomeStringDOSEnglish };

static const char *const k1VeryCleverStringDOSEnglish[1] = {
	"Very clever!  But your feeble efforts are futile."
};

static const StringListProvider k1VeryCleverStringDOSEnglishProvider = { ARRAYSIZE(k1VeryCleverStringDOSEnglish), k1VeryCleverStringDOSEnglish };

static const char *const k1GUIStringsDOSEnglish[28] = {
	"The Legend of Kyrandia",
	"Load a game",
	"Save this game",
	"Game controls",
	"Quit playing",
	"Resume game",
	"Game Controls",
	"Which game would you like to reload?",
	"Select a position to save to:",
	"[ EMPTY SLOT ]",
	"Cancel",
	"Enter a description of your saved game:",
	"Save",
	"Rest in peace, Brandon.",
	"Are you sure you want to quit playing?",
	"XXX",
	"XXX",
	"XXXXXXX",
	"XXXXXXXXX",
	"Main menu",
	"on",
	"off",
	"Yes",
	"No",
	"Walk speed ",
	"Text speed ",
	"Music is ",
	"Sounds are "
};

static const StringListProvider k1GUIStringsDOSEnglishProvider = { ARRAYSIZE(k1GUIStringsDOSEnglish), k1GUIStringsDOSEnglish };

static const char *const k1NewGameStringDOSEnglish[1] = {
	"[ START A NEW GAME ]"
};

static const StringListProvider k1NewGameStringDOSEnglishProvider = { ARRAYSIZE(k1NewGameStringDOSEnglish), k1NewGameStringDOSEnglish };

static const char *const k1ConfigStringsDOSEnglish[9] = {
	"slowest",
	"slow",
	"normal",
	"fast",
	"fastest",
	"slowest",
	"normal",
	"fast",
	"clickable"
};

static const StringListProvider k1ConfigStringsDOSEnglishProvider = { ARRAYSIZE(k1ConfigStringsDOSEnglish), k1ConfigStringsDOSEnglish };

