static const char *const k3MainMenuStringsDOSCDEnglish[4] = {
	"Start a new game",
	"Introduction",
	"Load a game",
	"Exit the game"
};

static const StringListProvider k3MainMenuStringsDOSCDEnglishProvider = { ARRAYSIZE(k3MainMenuStringsDOSCDEnglish), k3MainMenuStringsDOSCDEnglish };

static const char *const k3VqaSubtitlesIntroDOSCDEnglish[] = {
	""// TODO
};

static const StringListProvider k3VqaSubtitlesIntroDOSCDEnglishProvider = { ARRAYSIZE(k3VqaSubtitlesIntroDOSCDEnglish), k3VqaSubtitlesIntroDOSCDEnglish };

static const char *const k3VqaSubtitlesBoatDOSCDEnglish[] = {
	""// TODO
};

static const StringListProvider k3VqaSubtitlesBoatDOSCDEnglishProvider = { ARRAYSIZE(k3VqaSubtitlesBoatDOSCDEnglish), k3VqaSubtitlesBoatDOSCDEnglish };
